/*
 *	errmes.c
 *		Copyright (C)2001-2003 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include "errmes.h"

static void veprintf(const char *format, va_list argptr);

static char str_progname[1024] = "";

const char *err_getprogname(void)
{
	return str_progname;
}

void err_setprogname(const char *ptr_progname)
{
	if (ptr_progname == NULL)
		return;

	strncpy(str_progname, ptr_progname, sizeof(str_progname));
	str_progname[sizeof(str_progname) - 1] = '\0';
}

void err_exit(int status)
{
	fflush(stdout);

	if (*str_progname != '\0')
		fprintf(stderr, "%s: ", err_getprogname());

	fprintf(stderr, "%s\n", strerror(errno));

	exit(status);
}

void err_printf(const char *format, ...)
{
	va_list argptr;

	fflush(stdout);
	if (format != NULL)
	{
		va_start(argptr, format);
		veprintf(format, argptr);
		va_end(argptr);
	}

	exit(EXIT_FAILURE);
}

void wrn_printf(const char *format, ...)
{
	va_list argptr;

	if (format != NULL)
	{
		va_start(argptr, format);
		veprintf(format, argptr);
		va_end(argptr);
	}
}

static void veprintf(const char *format, va_list argptr)
{
	if (format == NULL)
		return;

	if (*str_progname != '\0')
		fprintf(stderr, "%s: ", err_getprogname());

	vfprintf(stderr, format, argptr);

	if (*format != '\0' && format[strlen(format) - 1] == ':')
		fprintf(stderr, " %s", strerror(errno));
	fprintf(stderr, "\n");
}
