/*
 *	ff4character.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "ff4file.h"
#include "ff4define.h"
#include "ff4decode.h"
#include "ff4character.h"

#define FILEPOS_CHRNAME			(0x036710 + 0x044000)	/* = 0x07a710 */
#define FILEPOS_CHRJOBNAME		(FILEPOS_CHRNAME + SIZEOF_CHRNAME * NUM_CHR)	/* = 0x07a764 */
#define FILEPOS_CHRDATA			(0x036900 + 0x044000)	/* = 0x07a900 */
#define FILEPOS_CHRLVUP			(0x037528 + 0x044000)	/* = 0x07b528 */
#define NUM_CHR					14
#define NUM_CHRNAME				14
#define NUM_CHRJOBNAME			14
#define NUM_CHRGROWPTN			8
#define SIZEOF_CHRNAME			6
#define SIZEOF_CHRJOBNAME		6
#define SIZEOF_CHRDATA			32
#define SIZEOF_CHRLVUP			5
#define SIZEOF_CHRGROWPTN		8

static void chr_setname(void);
static void chr_setjobname(void);

static char tbl_chrname[NUM_CHRNAME][256];			/* LN^̖Oi[e[u */
static char tbl_chrjobname[NUM_CHRJOBNAME][256];	/* Wu̖Oi[e[u */
static int flag_setchrname    = 0;					/* tbl_chrnameɃf[^i[Ă邩ǂ */
static int flag_setchrjobname = 0;					/* tbl_chrjobnameɃf[^i[Ă邩ǂ */

const char *chr_getname(int index)
{
	/* `FbN */
	if (index < 0 || NUM_CHR <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setchrname)
		chr_setname();

	return tbl_chrname[index];
}

const char *chr_getjobname(int index)
{
	/* `FbN */
	if (index < 0 || NUM_CHRJOBNAME <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setchrjobname)
		chr_setjobname();

	return tbl_chrjobname[index];
}

const char *chr_getname_id(int index)
{
	static const int tbl_id_to_chrname[] = {0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 3, 0, 3, 6, 9, 1, 5, 2, 10, 11, 1, 12, 13};

	assert(sizeof(tbl_id_to_chrname) / sizeof(tbl_id_to_chrname[0]) == 23);

	/* `FbN */
	if (index < 0 || 23 <= index)
	{
		assert(0);
		return "";
	}

	return chr_getname(tbl_id_to_chrname[index]);
}

const char *chr_getjobname_id(int index)
{
	static const int tbl_id_to_chrjobname[] = {0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 3, 9, 3, 6, 10, 1, 5, 2, 12, 13, 1, 0, 0};

	assert(sizeof(tbl_id_to_chrjobname) / sizeof(tbl_id_to_chrjobname[0]) == 23);

	/* `FbN */
	if (index < 0 || 23 <= index)
	{
		assert(0);
		return "";
	}

	return chr_getjobname(tbl_id_to_chrjobname[index]);
}

const char *chr_getname_no(int index)
{
	static const int tbl_no_to_chrname[] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 0, 9, 2, 10, 11, 13, 12};

	assert(sizeof(tbl_no_to_chrname) / sizeof(tbl_no_to_chrname[0]) == 16);

	/* `FbN */
	if (index < 0 || 16 <= index)
	{
		assert(0);
		return "";
	}

	return chr_getname(tbl_no_to_chrname[index]);
}

const char *chr_getjobname_no(int index)
{
	/* `FbN */
	if (index < 0 || 16 <= index)
	{
		assert(0);
		return "";
	}

	if (NUM_CHRJOBNAME < index)
		return "------------";

	return chr_getjobname(index);
}

static void chr_setname(void)
{
	int i;

	/* LN^̖Of[^ */
	for (i = 0; i < NUM_CHRNAME; i++)
		*tbl_chrname[i] = '\0';

	/* LN^̖Of[^ǂݍ */
	for (i = 0; i < NUM_CHRNAME; i++)
	{
		unsigned char buf[SIZEOF_CHRNAME];

		file_read(buf, FILEPOS_CHRNAME + SIZEOF_CHRNAME * i, SIZEOF_CHRNAME);
		dec_string(tbl_chrname[i], buf, sizeof(tbl_chrname[i]), SIZEOF_CHRNAME, 0);
	}

	flag_setchrname = 1;
}

static void chr_setjobname(void)
{
	int i;

	/* LN^̖Of[^ */
	for (i = 0; i < NUM_CHRJOBNAME; i++)
		*tbl_chrjobname[i] = '\0';

	/* LN^̃Wuf[^ǂݍ */
	for (i = 0; i < NUM_CHRJOBNAME; i++)
	{
		unsigned char buf[SIZEOF_CHRJOBNAME];

		file_read(buf, FILEPOS_CHRJOBNAME + SIZEOF_CHRJOBNAME * i, SIZEOF_CHRJOBNAME);
		dec_string(tbl_chrjobname[i], buf, sizeof(tbl_chrjobname[i]), SIZEOF_CHRJOBNAME, 0);
	}

	flag_setchrjobname = 1;
}

chrdata_t *chr_readdata(void)
{
	int i, j;
	long filepos;
	chrdata_t *buf_chrdata = NULL;

	/* LN^f[^i[obt@mۂ */
	if ((buf_chrdata = malloc(sizeof(chrdata_t) * NUM_CHR)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* LN^f[^ǂݍ */
	for (i = 0; i < NUM_CHR; i++)
	{
		int index = 0;
		unsigned char buf[SIZEOF_CHRDATA];

		file_read(buf, FILEPOS_CHRDATA + SIZEOF_CHRDATA * i, SIZEOF_CHRDATA);

		/* lnf[^ */
		buf_chrdata[i].chr_id                = buf[0] & 0x1f;
		buf_chrdata[i].chr_no                = buf[1] & 0x0f;
		buf_chrdata[i].initial_level         = buf[2];
		buf_chrdata[i].critical_effect_power = buf[21];
		buf_chrdata[i].critical_hit_rate     = buf[20];
		buf_chrdata[i].flag_right_hand       = ((buf[0] & 0x80) != 0x00);
		buf_chrdata[i].flag_left_hand        = ((buf[0] & 0x40) != 0x00);

		/* sv̈NA */
		index = buf_chrdata[i].initial_level - 1;
		for (j = 0; j < index; j++)
		{
			buf_chrdata[i].level[j].experience_point     = 0;
			buf_chrdata[i].level[j].hit_point_min        = 0;
			buf_chrdata[i].level[j].hit_point_max        = 0;
			buf_chrdata[i].level[j].magic_point_min      = 0;
			buf_chrdata[i].level[j].magic_point_max      = 0;
			buf_chrdata[i].level[j].increase_hit_point   = 0;
			buf_chrdata[i].level[j].increase_magic_point = 0;
			buf_chrdata[i].level[j].strength             = 0;
			buf_chrdata[i].level[j].speed                = 0;
			buf_chrdata[i].level[j].stamina              = 0;
			buf_chrdata[i].level[j].wisdom               = 0;
			buf_chrdata[i].level[j].will_power           = 0;
		}

		/* olExɂŒol */
		buf_chrdata[i].initial_experience_point          = (buf[25] << 16) | (buf[24] << 8) | buf[23];
		buf_chrdata[i].level[index].experience_point     = (buf[31] << 16) | (buf[30] << 8) | buf[29];
		if (buf_chrdata[i].level[index].experience_point > buf_chrdata[i].initial_experience_point)
			buf_chrdata[i].level[index].experience_point = buf_chrdata[i].initial_experience_point;

		/* xɂp[^ */
		buf_chrdata[i].level[index].hit_point_min        = (buf[10] << 8) | buf[ 9];
		buf_chrdata[i].level[index].hit_point_max        = buf_chrdata[i].level[index].hit_point_min;
		buf_chrdata[i].level[index].magic_point_min      = (buf[14] << 8) | buf[13];
		buf_chrdata[i].level[index].magic_point_max      = buf_chrdata[i].level[index].magic_point_min;
		buf_chrdata[i].level[index].increase_hit_point   = 0;
		buf_chrdata[i].level[index].increase_magic_point = 0;
		buf_chrdata[i].level[index].strength             = buf[15];
		buf_chrdata[i].level[index].speed                = buf[16];
		buf_chrdata[i].level[index].stamina              = buf[17];
		buf_chrdata[i].level[index].wisdom               = buf[18];
		buf_chrdata[i].level[index].will_power           = buf[19];
	}

	/* LN^f[^ǂݍ */
	filepos = FILEPOS_CHRLVUP;
	for (i = 0; i < NUM_CHR - 1; i++)
	{
		long diff_experience_point = 0;
		unsigned char buf_grow[SIZEOF_CHRGROWPTN];

		/* f[^ */
		for (j = buf_chrdata[i].initial_level; j < 70; j++)	/* x70܂ */
		{
			int increase_value;
			unsigned char buf[SIZEOF_CHRLVUP];

			filepos += file_read(buf, filepos, SIZEOF_CHRLVUP);

			increase_value = buf[0] & 0x07;
			if (increase_value > 0x04)
				increase_value -= 0x08;
			diff_experience_point = ((buf[2] & 0xe0) << (16 - 5)) | (buf[4] << 8) | buf[3];
			buf_chrdata[i].level[j].increase_hit_point   = buf[1];
			buf_chrdata[i].level[j].increase_magic_point = buf[2] & 0x1f;

			buf_chrdata[i].level[j].hit_point_min    = buf_chrdata[i].level[j - 1].hit_point_min +
														buf_chrdata[i].level[j].increase_hit_point;
			buf_chrdata[i].level[j].hit_point_max    = buf_chrdata[i].level[j - 1].hit_point_max +
														(buf_chrdata[i].level[j].increase_hit_point * 9 / 8);
			buf_chrdata[i].level[j].magic_point_min  = buf_chrdata[i].level[j - 1].magic_point_min +
														buf_chrdata[i].level[j].increase_magic_point;
			buf_chrdata[i].level[j].magic_point_max  = buf_chrdata[i].level[j - 1].magic_point_max +
														(buf_chrdata[i].level[j].increase_magic_point * 9 / 8);
			buf_chrdata[i].level[j].experience_point = buf_chrdata[i].level[j - 1].experience_point + diff_experience_point;

			if (buf_chrdata[i].level[j].hit_point_min > 9999)
				buf_chrdata[i].level[j].hit_point_min = 9999;
			if (buf_chrdata[i].level[j].hit_point_max > 9999)
				buf_chrdata[i].level[j].hit_point_max = 9999;
			if (buf_chrdata[i].level[j].magic_point_min > 9999)
				buf_chrdata[i].level[j].magic_point_min = 9999;
			if (buf_chrdata[i].level[j].magic_point_max > 9999)
				buf_chrdata[i].level[j].magic_point_max = 9999;

			buf_chrdata[i].level[j].strength   = buf_chrdata[i].level[j - 1].strength +
												(((buf[0] & 0x80) != 0x00) ? increase_value : 0);
			buf_chrdata[i].level[j].speed      = buf_chrdata[i].level[j - 1].speed +
												(((buf[0] & 0x40) != 0x00) ? increase_value : 0);
			buf_chrdata[i].level[j].stamina    = buf_chrdata[i].level[j - 1].stamina +
												(((buf[0] & 0x20) != 0x00) ? increase_value : 0);
			buf_chrdata[i].level[j].wisdom     = buf_chrdata[i].level[j - 1].wisdom +
												(((buf[0] & 0x10) != 0x00) ? increase_value : 0);
			buf_chrdata[i].level[j].will_power = buf_chrdata[i].level[j - 1].will_power +
												(((buf[0] & 0x08) != 0x00) ? increase_value : 0);
		}
		for (; j < NUM_CHRLV; j++)	/* x71ȏ */
		{
			buf_chrdata[i].level[j].increase_hit_point   = buf_chrdata[i].level[69].increase_hit_point;
			buf_chrdata[i].level[j].increase_magic_point = buf_chrdata[i].level[69].increase_magic_point;

			buf_chrdata[i].level[j].hit_point_min    = buf_chrdata[i].level[j - 1].hit_point_min +
														buf_chrdata[i].level[69].increase_hit_point;
			buf_chrdata[i].level[j].hit_point_max    = buf_chrdata[i].level[j - 1].hit_point_max +
														(buf_chrdata[i].level[69].increase_hit_point * 9 / 8);
			buf_chrdata[i].level[j].magic_point_min  = buf_chrdata[i].level[j - 1].magic_point_min +
														buf_chrdata[i].level[69].increase_magic_point;
			buf_chrdata[i].level[j].magic_point_max  = buf_chrdata[i].level[j - 1].magic_point_max +
														(buf_chrdata[i].level[69].increase_magic_point * 9 / 8);
			buf_chrdata[i].level[j].experience_point = buf_chrdata[i].level[j - 1].experience_point + diff_experience_point;

			if (buf_chrdata[i].level[j].hit_point_min > 9999)
				buf_chrdata[i].level[j].hit_point_min = 9999;
			if (buf_chrdata[i].level[j].hit_point_max > 9999)
				buf_chrdata[i].level[j].hit_point_max = 9999;
			if (buf_chrdata[i].level[j].magic_point_min > 9999)
				buf_chrdata[i].level[j].magic_point_min = 9999;
			if (buf_chrdata[i].level[j].magic_point_max > 9999)
				buf_chrdata[i].level[j].magic_point_max = 9999;

			buf_chrdata[i].level[j].strength   = 0;
			buf_chrdata[i].level[j].speed      = 0;
			buf_chrdata[i].level[j].stamina    = 0;
			buf_chrdata[i].level[j].wisdom     = 0;
			buf_chrdata[i].level[j].will_power = 0;
		}

		/* x71ȏł̐p^[ */
		filepos += file_read(buf_grow, filepos, SIZEOF_CHRGROWPTN);
		for (j = 0; j < NUM_CHRGROWPTN; j++)
		{
			int increase_value;

			increase_value = buf_grow[j] & 0x07;
			if (increase_value > 0x04)
				increase_value -= 0x08;

			buf_chrdata[i].grow_pattern[j].increase_strength   = ((buf_grow[j] & 0x80) != 0x00) ? increase_value : 0;
			buf_chrdata[i].grow_pattern[j].increase_speed      = ((buf_grow[j] & 0x40) != 0x00) ? increase_value : 0;
			buf_chrdata[i].grow_pattern[j].increase_stamina    = ((buf_grow[j] & 0x20) != 0x00) ? increase_value : 0;
			buf_chrdata[i].grow_pattern[j].increase_wisdom     = ((buf_grow[j] & 0x10) != 0x00) ? increase_value : 0;
			buf_chrdata[i].grow_pattern[j].increase_will_power = ((buf_grow[j] & 0x08) != 0x00) ? increase_value : 0;
		}
	}

	return buf_chrdata;
}

int chr_writeparam(const char *ptr_filename, const chrdata_t *ptr_chrdata)
{
	int i;
	FILE *fp = NULL;
	const char *str_hands[2][2] =
	{
		{"sp",	""},
		{"E",	""},
	};

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* LN^f[^(ln)o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* LN^f[^(ln)o͂ */
	fprintf(fp, "No.      O           "
				"Wu        "
				"  LV HP MP   EXP "
				"      ̗ m _ "
				"CH   CH  "
				"\n");
	for (i = 0; i < NUM_CHR; i++)
	{
		int index = ptr_chrdata[i].initial_level - 1;

		fprintf(fp, "%03d(%02X): %-12s : %-12s,  %3d, %5d, %5d, %8ld, %3d, %3d, %3d, %3d, %3d, %5d, %5d, %-6s\n", i, i,
				chr_getname_no(ptr_chrdata[i].chr_no),
				chr_getjobname_no(ptr_chrdata[i].chr_no),
				ptr_chrdata[i].initial_level,
				ptr_chrdata[i].level[index].hit_point_min,
				ptr_chrdata[i].level[index].magic_point_min,
				ptr_chrdata[i].initial_experience_point,
				ptr_chrdata[i].level[index].strength,
				ptr_chrdata[i].level[index].speed,
				ptr_chrdata[i].level[index].stamina,
				ptr_chrdata[i].level[index].wisdom,
				ptr_chrdata[i].level[index].will_power,
				ptr_chrdata[i].critical_effect_power,
				ptr_chrdata[i].critical_hit_rate,
				str_hands[(ptr_chrdata[i].flag_right_hand != 0) ? 1 : 0][(ptr_chrdata[i].flag_left_hand != 0) ? 1 : 0]);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int chr_writelvup(const char *ptr_filename, const chrdata_t *ptr_chrdata)
{
	int i, j;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* LN^f[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* LN^f[^o͂ */
	for (i = 0; i < NUM_CHR - 1; i++)
	{
		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[No.%03d(%02X): %s(%s)]\n", i, i, chr_getname_no(ptr_chrdata[i].chr_no), chr_getjobname_no(ptr_chrdata[i].chr_no));

		/* x71ȏł̐p^[ */
		fprintf(fp, "No.       ̗ m _\n");
		for (j = 0; j < NUM_CHRGROWPTN; j++)
		{
			fprintf(fp, "%02d:  %3d, %3d, %3d, %3d, %3d\n", j + 1,
					ptr_chrdata[i].grow_pattern[j].increase_strength,
					ptr_chrdata[i].grow_pattern[j].increase_speed,
					ptr_chrdata[i].grow_pattern[j].increase_stamina,
					ptr_chrdata[i].grow_pattern[j].increase_wisdom,
					ptr_chrdata[i].grow_pattern[j].increase_will_power);
		}

		/* f[^ */
		fprintf(fp, "LV        őHP (͈)     ㏸HP    őMP (͈)   ㏸MP    ol "
					"      ̗ m _\n");
		for (j = ptr_chrdata[i].initial_level - 1; j < 70; j++)
		{
			fprintf(fp, "%2d:  %4d (%4d - %4d), %3d - %3d, %3d (%3d - %3d), %2d - %2d, %8ld, %3d, %3d, %3d, %3d, %3d\n", j + 1,
					(ptr_chrdata[i].level[j].hit_point_min + ptr_chrdata[i].level[j].hit_point_max) / 2,
					ptr_chrdata[i].level[j].hit_point_min,
					ptr_chrdata[i].level[j].hit_point_max,
					ptr_chrdata[i].level[j].increase_hit_point,
					ptr_chrdata[i].level[j].increase_hit_point * 9 / 8,
					(ptr_chrdata[i].level[j].magic_point_min + ptr_chrdata[i].level[j].magic_point_max) / 2,
					ptr_chrdata[i].level[j].magic_point_min,
					ptr_chrdata[i].level[j].magic_point_max,
					ptr_chrdata[i].level[j].increase_magic_point,
					ptr_chrdata[i].level[j].increase_magic_point * 9 / 8,
					ptr_chrdata[i].level[j].experience_point,
					ptr_chrdata[i].level[j].strength,
					ptr_chrdata[i].level[j].speed,
					ptr_chrdata[i].level[j].stamina,
					ptr_chrdata[i].level[j].wisdom,
					ptr_chrdata[i].level[j].will_power);
		}
		for (; j < NUM_CHRLV; j++)
		{
			fprintf(fp, "%2d:  %4d (%4d - %4d), %3d - %3d, %3d (%3d - %3d), %2d - %2d, %8ld\n", j + 1,
					(ptr_chrdata[i].level[j].hit_point_min + ptr_chrdata[i].level[j].hit_point_max) / 2,
					ptr_chrdata[i].level[j].hit_point_min,
					ptr_chrdata[i].level[j].hit_point_max,
					ptr_chrdata[i].level[j].increase_hit_point,
					ptr_chrdata[i].level[j].increase_hit_point * 9 / 8,
					(ptr_chrdata[i].level[j].magic_point_min + ptr_chrdata[i].level[j].magic_point_max) / 2,
					ptr_chrdata[i].level[j].magic_point_min,
					ptr_chrdata[i].level[j].magic_point_max,
					ptr_chrdata[i].level[j].increase_magic_point,
					ptr_chrdata[i].level[j].increase_magic_point * 9 / 8,
					ptr_chrdata[i].level[j].experience_point);
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}
