/*
 *	ff4decode.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "ff4fileps.h"
#include "ff4decode.h"

/* t@Cit@^W[IV̕R[h\ */
static const char *tbl_charcode[] =
{
	"",   "\n", "", "", "", "", "", "", "", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "", "", "K", "M", "O", "Q", "S", "U", "W",
	"Y", "[", "]", "_", "a", "d", "f", "h", "o", "r", "u", "x", "{", "p", "s", "v",
	"y", "|", "A",  "B",  "C",  "D",  "E",  "F",  "G",  "H",  "I",  "J",  "K",  "L",  "M",  "N",
	"O",  "P",  "Q",  "R",  "S",  "T",  "U",  "V",  "W",  "X",  "Y",  "Z",  "a",  "b",  "c",  "d",
	"e",  "f",  "g",  "h",  "i",  "j",  "k",  "l",  "m",  "n",  "o",  "p",  "q",  "r",  "s",  "t",
	"u",  "v",  "w",  "x",  "y",  "z",  "M",  "H",  "P",  "w", "x", "", "", "", "", "",
	"0",  "1",  "2",  "3",  "4",  "5",  "6",  "7",  "8",  "9",  "", "", "", "", "", "",
	"", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "@", "B", "u", "F", "H", "b", "", "", "",
	"A", "B", "[", "d", "!",  "?",  "%",  "/",  ":",  "v", "A", "C", "E", "G", "I", "J",
	"L", "N", "P", "R", "T", "V", "X", "Z", "\", "^", "`", "c", "e", "g", "i", "j",
	"k", "l", "m", "n", "q", "t", "w", "z", "}", "~", "", "", "", "", "", "",
	"", "", "", "", "", "", "", "", "", "", "", "", "", "", "", " ",
};

static const char *str_chrname[] =
{
	"ZV",
	"JC",
	"fBA",
	"e",
	"Mo[g",
	"[U",
	"",
	"p",
	"|",
	"Vh",
	"GbW",
	"t[X[",
	"Sx[U",
	"Ai",
};

char *dec_string(char *ptr_dstbuf, const unsigned char *ptr_srcbuf, size_t dstbufsize, size_t srcbufsize, int flag_message)
{
	int i;
	char *buf_dstbuf = NULL;
	char *ptr = NULL;
	const size_t len_buf = srcbufsize * 8;
	const char *str_dummy = "";

	assert(sizeof(tbl_charcode) / sizeof(tbl_charcode[0]) == 256);
	assert(sizeof(str_chrname) / sizeof(str_chrname[0]) == 14);

	/* `FbN */
	if (ptr_dstbuf == NULL)
	{
		assert(0);
		return NULL;
	}
	if (ptr_srcbuf == NULL || srcbufsize == 0 || dstbufsize == 0)
	{
		assert(0);
		return ptr_dstbuf;
	}

	/* ꎞobt@mۂ */
	if ((buf_dstbuf = malloc(len_buf)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* fR[h */
	memset(buf_dstbuf, '\0', sizeof(len_buf));
	if (!flag_message)
	{
		for (i = 0; i < (int)srcbufsize; i++)
			strcat(buf_dstbuf, tbl_charcode[ptr_srcbuf[i]]);
	}
	else
	{
		int control_code = 0x00;

		for (i = 0; i < (int)srcbufsize; i++)
		{
			if (control_code > 0x00)	/* R[h2oCgڂ̏ */
			{
				switch (control_code)
				{
				case 0x04:	/* LN^̖O */
					if (ptr_srcbuf[i] < 14)
						strcat(buf_dstbuf, str_chrname[ptr_srcbuf[i]]);
					break;
				case 0x06:
					strcat(buf_dstbuf, str_dummy);
					break;
				default:
					break;
				}

				control_code = 0x00;
			}
			else if (ptr_srcbuf[i] >= 0x02 && ptr_srcbuf[i] <= 0x06 && control_code == 0x00)	/* R[h */
				control_code = ptr_srcbuf[i];
			else
				strcat(buf_dstbuf, tbl_charcode[ptr_srcbuf[i]]);
		}
	}

	/* ̋󔒂 */
	while ((ptr = strrchr(buf_dstbuf, ' ')) != NULL && *(ptr + 1) == '\0')
		*ptr = '\0';

	/* ꎞobt@̓eo͂ */
	strncpy(ptr_dstbuf, buf_dstbuf, dstbufsize - 1);
	ptr_dstbuf[dstbufsize - 1] = '\0';

	free(buf_dstbuf);
	buf_dstbuf = NULL;

	return ptr_dstbuf;
}

int dec_file(const char *ptr_readfilename, const char *ptr_writefilename, int flag_message, int flag_interleave)
{
	int i, j;
	FILE *fp_read  = NULL;
	FILE *fp_write = NULL;

	assert(sizeof(tbl_charcode) / sizeof(tbl_charcode[0]) == 256);
	assert(sizeof(str_chrname) / sizeof(str_chrname[0]) == 14);

	/* `FbN */
	if (ptr_readfilename == NULL || *ptr_readfilename == '\0' || ptr_writefilename == NULL || *ptr_writefilename == '\0')
	{
		assert(0);
		return 0;
	}

	/* fR[ht@CJ */
	if ((fp_read = fopen(ptr_readfilename, "rb")) == NULL)
		return 0;

	/* fR[hʂo͂t@CJ */
	if ((fp_write = fopen(ptr_writefilename, "w")) == NULL)
	{
		fclose(fp_read);
		fp_read = NULL;
		return 0;
	}

	/* fR[h */
	if (!flag_message)
	{
		while (!feof(fp_read))
		{
			int readnum;
			unsigned char buf[8];

			memset(buf, 0x00, sizeof(buf));
			if ((readnum = fread(buf, 1, sizeof(buf), fp_read)) < sizeof(buf) && ferror(fp_read))
				err_printf(STR_FATAL_ERROR_IN_FREAD);
			if (flag_interleave)
				file_alignword(buf, sizeof(buf));

			for (i = 0; i < readnum; i++)
			{
				if (buf[i] == 0x00)	/* Kxɉs邽 */
					fputc('\n', fp_write);
				else
					fputs(tbl_charcode[buf[i]], fp_write);
			}
		}
	}
	else
	{
		int control_code = 0x00;

		while (!feof(fp_read))
		{
			int readnum;
			unsigned char buf[8];

			memset(buf, 0x00, sizeof(buf));
			if ((readnum = fread(buf, 1, sizeof(buf), fp_read)) < sizeof(buf) && ferror(fp_read))
				err_printf(STR_FATAL_ERROR_IN_FREAD);
			if (flag_interleave)
				file_alignword(buf, sizeof(buf));

			for (i = 0; i < readnum; i++)
			{
				if (control_code > 0x00)	/* R[h2oCgڂ̏ */
				{
					switch (control_code)
					{
					case 0x02:	/* A󔒂 */
						for (j = 0; j < buf[i]; j++)
							fputs(tbl_charcode[0xff], fp_write);
						break;
					case 0x04:	/* LN^̖O */
						if (buf[i] < 14)
							fputs(str_chrname[buf[i]], fp_write);
						break;
					case 0x03:	/* y炷 */
					case 0x05:	/* s */
					default:
						break;
					}

					control_code = 0x00;
				}
				else if (buf[i] >= 0x02 && buf[i] <= 0x05 && control_code == 0x00)	/* R[h */
					control_code = buf[i];
				else if (buf[i] == 0x06)	/* R[h */
					;	/* o͂Ȃ */
				else if (buf[i] == 0x00)	/* Kxɉs邽 */
					fputc('\n', fp_write);
				else
					fputs(tbl_charcode[buf[i]], fp_write);
			}
		}
	}

	fclose(fp_read);
	fclose(fp_write);
	fp_read  = NULL;
	fp_write = NULL;

	return 1;
}
