/*
 *	ff4define.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "ff4file.h"
#include "ff4define.h"

#define FILEPOS_ELEMSTAT	(0x036590 + 0x044000)	/* = 0x07a590 */
#define NUM_ELEMSTAT		128
#define SIZEOF_ELEMSTAT		3

#define elem_maskbit(elem_no)		(1U << (elem_no))
#define state_maskbit(state_no)		(1U << (state_no))
#define montype_maskbit(montype_no)	(1U << (montype_no))

#define STR_NOTHING		"E"
#define STR_FLAG_TRUE	""
#define STR_FLAG_FALSE	"~"

static void elemstate_settbl(void);

static const char *str_element[/* NUM_ELEMENT */] =
{
	"",	/*  */
	"X",	/* C */
	"",	/*  */
	"",	/* ł̗ */
	"",	/* Ȃ */
	"",	/* ѓ */
	"z",	/* z */
	"",	/* (ʑ)*/
};

static const char *str_state[/* NUM_STATE */] =
{
	"",	/*  */
	"",	/* È */
	"",	/*  */
	"",	/*  */
	"",	/* l */
	"^",	/* ^ */
	"",	/* Ή */
	"",	/* 퓬s\ */

	"",	/* XɐΉ */
	"",	/* XɐΉ(_~[?) */
	"",	/* m */
	"",	/*  */
	"",	/*  */
	"",	/*  */
	"",	/* reg */
	"",	/*  */

	"",	/* ̐鍐 */
	"??",	/* s */
	"??",	/* s */
	"??",	/* s */
	"??",	/* s */
	"~",	/* Xgbv */
	"??",	/* s */
	"??",	/* s */

	"m",	/* m */
	"??",	/* s */
	"",	/* g(1) */
	"",	/* g(2) */
	"o",	/* oA */
	"",	/* tN */
	"??",	/* s */
	"??",	/* s(E? G?) */
};

static const char *str_montype[/* NUM_MONTYPE */] =
{
	"hS",
	"@B",
	"U[h",
	"썰",
	"l",
	"v",
	"@g",
	"Afbh",
};

static elem_t tbl_element[NUM_ELEMSTAT];	/* p^[i[e[u */
static state_t tbl_state[NUM_ELEMSTAT];		/* ԕωp^[i[e[u */
static int flag_setelemstat = 0;			/* tbl_elementEtbl_stateɃf[^i[Ă邩ǂ */

elem_t elem_gettbl(int index)
{
	/* `FbN */
	if (index < 0 || NUM_ELEMSTAT <= index)
	{
		assert(0);
		return 0x00;
	}

	if (!flag_setelemstat)
		elemstate_settbl();

	return tbl_element[index];
}

state_t state_gettbl(int index)
{
	/* `FbN */
	if (index < 0 || NUM_ELEMSTAT <= index)
	{
		assert(0);
		return 0x00;
	}

	if (!flag_setelemstat)
		elemstate_settbl();

	return tbl_state[index];
}

static void elemstate_settbl(void)
{
	int i;

	/* Eԕωf[^ǂݍ */
	for (i = 0; i < NUM_ELEMSTAT; i++)
	{
		unsigned char buf[SIZEOF_ELEMSTAT];

		file_read(buf, FILEPOS_ELEMSTAT + SIZEOF_ELEMSTAT * i, SIZEOF_ELEMSTAT);
		tbl_element[i] = buf[0];
		tbl_state[i]   = (buf[2] << 8) | buf[1];
	}

	flag_setelemstat = 1;
}

const char *str_getflag(int flag)
{
	return (flag != 0x00) ? STR_FLAG_TRUE : STR_FLAG_FALSE;
}

const char *str_getelem(elem_t elem, int index)
{
	assert(sizeof(str_element) / sizeof(str_element[0]) == NUM_ELEMENT);

	/* `FbN */
	if (index < 0 || NUM_ELEMENT <= index)
	{
		assert(0);
		return "";
	}

	return ((elem & elem_maskbit(index)) != 0x00) ? str_element[index] : "";
}

void str_printelem(FILE *stream, elem_t elem, elem_t elem_mask)
{
	int i;

	assert(sizeof(str_element) / sizeof(str_element[0]) == NUM_ELEMENT);

	/* `FbN */
	if (stream == NULL)
	{
		assert(0);
		return;
	}

	for (i = 0; i < NUM_ELEMENT; i++)
	{
		if ((elem_mask & elem_maskbit(i)) != 0x00)
			fprintf(stream, "%-2s", ((elem & elem_maskbit(i)) != 0x00) ? str_element[i] : STR_NOTHING);
	}
}

const char *str_getstate(state_t state, int index)
{
	assert(sizeof(str_state) / sizeof(str_state[0]) == NUM_STATE);

	/* `FbN */
	if (index < 0 || NUM_STATE <= index)
	{
		assert(0);
		return "";
	}

	return ((state & state_maskbit(index)) != 0x00) ? str_state[index] : "";
}

void str_printstate(FILE *stream, state_t state, state_t state_mask)
{
	int i, count = 0;

	assert(sizeof(str_state) / sizeof(str_state[0]) == NUM_STATE);

	/* `FbN */
	if (stream == NULL)
	{
		assert(0);
		return;
	}

	for (i = 0; i < NUM_STATE; i++)
	{
		if ((i % 8) == 0)
		{
			if (count > 0 && state_mask > state_maskbit(i))
				fprintf(stream, " ");
			count = 0;
		}

		if ((state_mask & state_maskbit(i)) != 0x00)
		{
			fprintf(stream, "%-2s", ((state & state_maskbit(i)) != 0x00) ? str_state[i] : STR_NOTHING);
			count++;
		}
	}
}

const char *str_getmontype(montype_t montype, int index)
{
	assert(sizeof(str_montype) / sizeof(str_montype[0]) == NUM_MONTYPE);

	/* `FbN */
	if (index < 0 || NUM_MONTYPE <= index)
	{
		assert(0);
		return "";
	}

	return ((montype & montype_maskbit(index)) != 0x00) ? str_montype[index] : "";
}

void str_printmontype(FILE *stream, montype_t montype, montype_t montype_mask)
{
	int i, count = 0;

	assert(sizeof(str_montype) / sizeof(str_montype[0]) == NUM_MONTYPE);

	/* `FbN */
	if (stream == NULL)
	{
		assert(0);
		return;
	}

	for (i = 0; i < NUM_MONTYPE; i++)
	{
		if ((montype_mask & montype_maskbit(i)) != 0x00)
		{
			if ((montype & montype_maskbit(i)) != 0x00)
			{
				if (count > 0)
					fprintf(stream, " ");
				count++;
			}

			fprintf(stream, "%s", str_getmontype(montype, i));
		}
	}
}
