/*
 *	ff4fileps.c
 *		Copyright (C)2001, 2002 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "ff4fileps.h"

#define FILENAME_FF4		"ff4.bin"
#define FILEPOS_ORIGIN_PS	0x044000

static size_t file_alignfread(void *ptr_buf, size_t size, size_t count, FILE *stream);

static FILE *fp_ps = NULL;					/* t@C|C^(vCXe[Vŗp) */
static char str_dirname_ps[1024] = "";		/* CD-ROM̃[gpX */

int file_open_ps(const char *ptr_dirname)
{
	char path[1024] = "";

	/* `FbN */
	if (ptr_dirname == NULL || *ptr_dirname == '\0')
	{
		assert(0);
		return 0;
	}

	strncpy(str_dirname_ps, ptr_dirname, sizeof(str_dirname_ps));
	str_dirname_ps[sizeof(str_dirname_ps) - 1] = '\0';

	/* CD-ROM̃t@CJ */
	strncpy(path, str_dirname_ps, sizeof(path));
	str_dirname_ps[sizeof(path) - 1] = '\0';
	strcat(path, FILENAME_FF4);
	if ((fp_ps = fopen(path, "rb")) == NULL)
		return 0;

	return 1;
}

int file_close_ps(void)
{
	int ret = 0;

	if (fp_ps != NULL)
		fclose(fp_ps);

	/* f[^ */
	fp_ps = NULL;
	*str_dirname_ps = '\0';

	return ret;
}

size_t file_read_ps(void *ptr_buf, long filepos, size_t size)
{
	size_t ret = 0;

	/* `FbN */
	if (ptr_buf == NULL || filepos < 0)
	{
		assert(0);
		return 0;
	}

	/* V[Nʒu␳ă[h */
	if (ftell(fp_ps) != filepos - FILEPOS_ORIGIN_PS)
	{
		if (fseek(fp_ps, filepos - FILEPOS_ORIGIN_PS, SEEK_SET))
			err_printf(STR_FATAL_ERROR_IN_FSEEK);
	}
	ret = file_alignfread(ptr_buf, 1, size, fp_ps);

	return ret;
}

void file_alignword(unsigned char *ptr_buf, size_t bufsize)
{
	int i, j;
	unsigned char *ptr = ptr_buf;

	/* `FbN */
	if (ptr_buf == NULL)
	{
		assert(0);
		return;
	}

	for (i = 0; i < (int)bufsize / 8; i++)
	{
		for (j = 0; j < 4; j++)
		{
			unsigned char tmp_ch;

			tmp_ch = ptr[j + 4];
			ptr[j + 4] = ptr[j];
			ptr[j] = tmp_ch;
		}
		ptr += 8;
	}
}

static size_t file_alignfread(void *ptr_buf, size_t size, size_t count, FILE *stream)
{
	long offset;
	long read_begin;
	long read_end;
	size_t bufsize;
	unsigned char *buf = NULL;

	/* `FbN */
	if (ptr_buf == NULL || stream == NULL)
	{
		assert(0);
		return 0;
	}

	offset = ftell(stream);
	read_begin = (offset / 8) * 8;
	read_end = ((offset + size * count + 7) / 8) * 8;
	bufsize = read_end - read_begin;

	if ((buf = malloc(bufsize)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);
	memset(buf, 0x00, bufsize);

	if (offset != read_begin)
	{
		if (fseek(stream, read_begin, SEEK_SET))
			err_printf(STR_FATAL_ERROR_IN_FSEEK);
	}

	if (fread(buf, 1, bufsize, stream) < bufsize)
		err_printf(STR_FATAL_ERROR_IN_FREAD);

	file_alignword(buf, bufsize);

	if (ftell(stream) != offset + (long)(size * count))
	{
		if (fseek(stream, offset + size * count, SEEK_SET))
			err_printf(STR_FATAL_ERROR_IN_FSEEK);
	}

	memmove(ptr_buf, &buf[offset - read_begin], size * count);

	free(buf);
	buf = NULL;

	return size * count;
}
