/*
 *	ff4item.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "ff4file.h"
#include "ff4decode.h"
#include "ff4item.h"

#define FILEPOS_ITEMNAME		(0x034000 + 0x044000)	/* = 0x078000 */
#define FILEPOS_ITEMDATA		(0x035100 + 0x044000)	/* = 0x079100 */
#define FILEPOS_ITEMPRICE		(0x036450 + 0x044000)	/* = 0x07a450 */
#define FILEPOS_ITEMEQUIP		(0x036550 + 0x044000)	/* = 0x07a550 */
#define NUM_ITEM_WEAPON			96
#define NUM_ITEM_PROTECTOR		80
#define NUM_ITEM_TOOL			46
#define NUM_ITEM_OTHER			34
#define NUM_ITEM_EQUIP			(NUM_ITEM_WEAPON + NUM_ITEM_PROTECTOR)	/* = 176 */
#define NUM_ITEM				(NUM_ITEM_WEAPON + NUM_ITEM_PROTECTOR + NUM_ITEM_TOOL + NUM_ITEM_OTHER)	/* = 256 */
#define NUM_ITEMEQUIP			32
#define SIZEOF_ITEMNAME			9
#define SIZEOF_ITEMDATA_EQUIP	8
#define SIZEOF_ITEMDATA_TOOL	6
#define SIZEOF_ITEMDATA_OTHER	0
#define SIZEOF_ITEMPRICE		1
#define SIZEOF_ITEMEQUIP		2
#define REGION_WEAPON_BEGIN		0
#define REGION_WEAPON_END		(REGION_WEAPON_BEGIN + NUM_ITEM_WEAPON - 1)			/* =  95 */
#define REGION_PROTECTOR_BEGIN	(REGION_WEAPON_END + 1)								/* =  96 */
#define REGION_PROTECTOR_END	(REGION_PROTECTOR_BEGIN + NUM_ITEM_PROTECTOR - 1)	/* = 175 */
#define REGION_TOOL_BEGIN		(REGION_PROTECTOR_END + 1)							/* = 176 */
#define REGION_TOOL_END			(REGION_TOOL_BEGIN + NUM_ITEM_TOOL - 1)				/* = 221 */
#define REGION_OTHER_BEGIN		(REGION_TOOL_END + 1)								/* = 222 */
#define REGION_OTHER_END		(REGION_OTHER_BEGIN + NUM_ITEM_OTHER - 1)			/* = 255 */

static void item_setname(void);

static char tbl_itemname[NUM_ITEM][256];	/* ACe̖Oi[e[u */
static int flag_setitemname = 0;			/* tbl_itemnameɃf[^i[Ă邩ǂ */

const char *item_getname(int index)
{
	/* `FbN */
	if (index < 0 || NUM_ITEM <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setitemname)
		item_setname();

	return tbl_itemname[index];
}

static void item_setname(void)
{
	int i;

	/* ACe̖Of[^ */
	for (i = 0; i < NUM_ITEM; i++)
		*tbl_itemname[i] = '\0';

	/* ACe̖Of[^ǂݍ */
	for (i = 0; i < NUM_ITEM; i++)
	{
		unsigned char buf[SIZEOF_ITEMNAME];

		file_read(buf, FILEPOS_ITEMNAME + SIZEOF_ITEMNAME * i, SIZEOF_ITEMNAME);
		dec_string(tbl_itemname[i], &buf[1], sizeof(tbl_itemname[i]), SIZEOF_ITEMNAME - 1, 0);
	}

	flag_setitemname = 1;
}

itemdata_t *item_readdata(void)
{
	int i;
	long filepos;
	itemdata_t *buf_itemdata = NULL;
	unsigned long tbl_equip_chr[NUM_ITEMEQUIP];
	const int tbl_inclease[8][2] =
	{
		{  0,   3},
		{  0,   5},
		{  0,  10},
		{  0,  15},
		{ -5,   5},
		{-10,  10},
		{-15,  15},
		{ -5,   5},
	};

	/* ACe̊ef[^i[obt@mۂ */
	if ((buf_itemdata = malloc(sizeof(itemdata_t) * NUM_ITEM)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* ACẻif[^ǂݍ */
	for (i = 0; i < NUM_ITEM; i++)
	{
		unsigned char buf[SIZEOF_ITEMPRICE];

		file_read(buf, FILEPOS_ITEMPRICE + SIZEOF_ITEMPRICE * i, SIZEOF_ITEMPRICE);
		buf_itemdata[i].price = (long)(buf[0] & 0x7f) * (((buf[0] & 0x80) != 0x00) ? 1000 : 10);
	}

	/* LN^f[^ǂݍ */
	for (i = 0; i < NUM_ITEMEQUIP; i++)
	{
		unsigned char buf[SIZEOF_ITEMEQUIP];

		file_read(buf, FILEPOS_ITEMEQUIP + SIZEOF_ITEMEQUIP * i, SIZEOF_ITEMEQUIP);
		tbl_equip_chr[i] = (buf[1] << 8) | buf[0];
	}

	/* ACe̊ef[^ǂݍ */
	filepos = FILEPOS_ITEMDATA;
	for (i = 0; i < NUM_ITEM; i++)
	{
		/* ACe̎ނ𔻕ʂ */
		if (REGION_WEAPON_BEGIN <= i && i <= REGION_WEAPON_END)
		{
			unsigned char buf[SIZEOF_ITEMDATA_EQUIP];

			filepos += file_read(buf, filepos, SIZEOF_ITEMDATA_EQUIP);

			/* lnf[^ */
			buf_itemdata[i].param.weapon.attack_power = buf[1];
			buf_itemdata[i].param.weapon.hit_rate     = buf[2] & 0x7f;
			buf_itemdata[i].param.weapon.effect_no    = buf[3];

			/* Eԕωnf[^ */
			buf_itemdata[i].param.weapon.elem_attack    = elem_gettbl(buf[4] & 0x7f);
			buf_itemdata[i].param.weapon.state_attack   = state_gettbl(buf[4] & 0x7f);
			buf_itemdata[i].param.weapon.montype_attack = buf[5];
			buf_itemdata[i].param.weapon.bits_equip_chr = tbl_equip_chr[buf[6] & 0x1f];

			/* tOnf[^ */
			buf_itemdata[i].param.weapon.flag_magnetic  = ((buf[0] & 0x80) != 0x00);
			buf_itemdata[i].param.weapon.flag_throw     = ((buf[0] & 0x40) != 0x00);
			buf_itemdata[i].param.weapon.flag_backrow   = ((buf[0] & 0x20) != 0x00);
			buf_itemdata[i].param.weapon.flag_use       = ((buf[2] & 0x80) != 0x00);
			buf_itemdata[i].param.weapon.flag_good_hand = ((buf[6] & 0x40) != 0x00);
			buf_itemdata[i].param.weapon.flag_weak_hand = ((buf[6] & 0x80) != 0x00);
			buf_itemdata[i].param.weapon.flag_2handed   = ((buf[6] & 0x20) != 0x00);

			/* p[^ωf[^ */
			buf_itemdata[i].param.weapon.increase_strength =
				tbl_inclease[buf[7] & 0x07][((buf[7] & 0x80) != 0x00) ? 1 : 0];
			buf_itemdata[i].param.weapon.increase_speed =
				tbl_inclease[buf[7] & 0x07][((buf[7] & 0x40) != 0x00) ? 1 : 0];
			buf_itemdata[i].param.weapon.increase_stamina =
				tbl_inclease[buf[7] & 0x07][((buf[7] & 0x20) != 0x00) ? 1 : 0];
			buf_itemdata[i].param.weapon.increase_wisdom =
				tbl_inclease[buf[7] & 0x07][((buf[7] & 0x10) != 0x00) ? 1 : 0];
			buf_itemdata[i].param.weapon.increase_will_power =
				tbl_inclease[buf[7] & 0x07][((buf[7] & 0x08) != 0x00) ? 1 : 0];

			buf_itemdata[i].type = ITEM_TYPE_WEAPON;
		}
		else if (REGION_PROTECTOR_BEGIN <= i && i <= REGION_PROTECTOR_END)
		{
			unsigned char buf[SIZEOF_ITEMDATA_EQUIP];

			filepos += file_read(buf, filepos, SIZEOF_ITEMDATA_EQUIP);

			/* lnf[^ */
			buf_itemdata[i].param.protector.defense_power       = buf[1];
			buf_itemdata[i].param.protector.evade_rate          = buf[2] & 0x7f;
			buf_itemdata[i].param.protector.magic_defense_power = buf[3];
			buf_itemdata[i].param.protector.magic_evade_rate    = buf[0] & 0x7f;

			/* Eԕωnf[^ */
			buf_itemdata[i].param.protector.elem_defense    = elem_gettbl(buf[4] & 0x7f);
			buf_itemdata[i].param.protector.state_defense   = state_gettbl(buf[4] & 0x7f);
			buf_itemdata[i].param.protector.montype_defense = buf[5];
			buf_itemdata[i].param.protector.bits_equip_chr  = tbl_equip_chr[buf[6] & 0x1f];

			/* tOnf[^ */
			buf_itemdata[i].param.protector.flag_magnetic = ((buf[0] & 0x80) != 0x00);

			/* p[^ωf[^ */
			buf_itemdata[i].param.protector.increase_strength =
				tbl_inclease[buf[7] & 0x07][((buf[7] & 0x80) != 0x00) ? 1 : 0];
			buf_itemdata[i].param.protector.increase_speed =
				tbl_inclease[buf[7] & 0x07][((buf[7] & 0x40) != 0x00) ? 1 : 0];
			buf_itemdata[i].param.protector.increase_stamina =
				tbl_inclease[buf[7] & 0x07][((buf[7] & 0x20) != 0x00) ? 1 : 0];
			buf_itemdata[i].param.protector.increase_wisdom =
				tbl_inclease[buf[7] & 0x07][((buf[7] & 0x10) != 0x00) ? 1 : 0];
			buf_itemdata[i].param.protector.increase_will_power =
				tbl_inclease[buf[7] & 0x07][((buf[7] & 0x08) != 0x00) ? 1 : 0];

			buf_itemdata[i].type = ITEM_TYPE_PROTECTOR;
		}
		else if (REGION_TOOL_BEGIN <= i && i <= REGION_TOOL_END)
		{
			unsigned char buf[SIZEOF_ITEMDATA_TOOL];

			filepos += file_read(buf, filepos, SIZEOF_ITEMDATA_TOOL);
			mag_setparam(&buf_itemdata[i].param.tool, buf, sizeof(buf));

			buf_itemdata[i].type = ITEM_TYPE_TOOL;
		}
		else if (REGION_OTHER_BEGIN <= i && i <= REGION_OTHER_END)
		{
			buf_itemdata[i].type = ITEM_TYPE_OTHER;
		}
		else
			assert(0);
	}

	return buf_itemdata;
}

int item_writeparam(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ACe̐lnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ACe̐lnf[^o͂() */
	fprintf(fp, "No.      O               "
				"  i "
				"U                "
				"      ̗ m _\n");
	for (i = REGION_WEAPON_BEGIN; i <= REGION_WEAPON_END; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s : %6ld, %5d, %5d,               %3d, %3d, %3d, %3d, %3d\n", i, i,
				item_getname(i),
				ptr_itemdata[i].price,
				ptr_itemdata[i].param.weapon.attack_power,
				ptr_itemdata[i].param.weapon.hit_rate,
				ptr_itemdata[i].param.weapon.increase_strength,
				ptr_itemdata[i].param.weapon.increase_speed,
				ptr_itemdata[i].param.weapon.increase_stamina,
				ptr_itemdata[i].param.weapon.increase_wisdom,
				ptr_itemdata[i].param.weapon.increase_will_power);
	}

	/* ACe̐lnf[^o͂(h) */
	fprintf(fp, "No.      O               "
				"  i "
				"h  h  "
				"      ̗ m _\n");
	for (i = REGION_PROTECTOR_BEGIN; i <= REGION_PROTECTOR_END; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s : %6ld, %5d, %5d, %5d, %5d, %3d, %3d, %3d, %3d, %3d\n", i, i,
				item_getname(i),
				ptr_itemdata[i].price,
				ptr_itemdata[i].param.protector.defense_power,
				ptr_itemdata[i].param.protector.evade_rate,
				ptr_itemdata[i].param.protector.magic_defense_power,
				ptr_itemdata[i].param.protector.magic_evade_rate,
				ptr_itemdata[i].param.protector.increase_strength,
				ptr_itemdata[i].param.protector.increase_speed,
				ptr_itemdata[i].param.protector.increase_stamina,
				ptr_itemdata[i].param.protector.increase_wisdom,
				ptr_itemdata[i].param.protector.increase_will_power);
	}

	/* ACe̐lnf[^o͂() */
	fprintf(fp, "No.      O               "
				"  i "
				"ʒl   䗦 \n");
	for (i = REGION_TOOL_BEGIN; i <= REGION_TOOL_END; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s : %6ld, %5d, %2s%3d, %5d\n", i, i,
				item_getname(i),
				ptr_itemdata[i].price,
				ptr_itemdata[i].param.tool.effect_power,
				(ptr_itemdata[i].param.tool.effect_rate != 0) ? "1/" : "  ",
				ptr_itemdata[i].param.tool.effect_rate,
				ptr_itemdata[i].param.tool.hit_rate);
	}
	for (i = REGION_OTHER_BEGIN; i <= REGION_OTHER_END; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s : %6ld\n", i, i,
				item_getname(i),
				ptr_itemdata[i].price);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writeelem(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ACȇnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ACȇnf[^o͂ */
	fprintf(fp, "No.      O               "
				"\n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :", i, i, item_getname(i));

		switch (ptr_itemdata[i].type)
		{
		case ITEM_TYPE_WEAPON:
			fprintf(fp, " ");

			/* U */
			str_printelem(fp, ptr_itemdata[i].param.weapon.elem_attack, ELEM_ALL);
			break;

		case ITEM_TYPE_PROTECTOR:
			fprintf(fp, " ");

			/* Uւ̑ϐ */
			str_printelem(fp, ptr_itemdata[i].param.protector.elem_defense, ELEM_ALL);
			break;

		case ITEM_TYPE_TOOL:
			fprintf(fp, " ");

			/*  */
			str_printelem(fp, ptr_itemdata[i].param.tool.elem_effect, ELEM_ALL);
			break;
		}

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writestate(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ACȅԕωnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ACȅԕωnf[^o͂ */
	fprintf(fp, "No.      O               "
				"ԕω1        2                3    4\n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :", i, i, item_getname(i));

		switch (ptr_itemdata[i].type)
		{
		case ITEM_TYPE_WEAPON:
			fprintf(fp, " ");

			/* ʏȔԕω */
			str_printstate(fp, ptr_itemdata[i].param.weapon.state_attack, STATE_ALL & ~STATE_OTHER);
			break;

		case ITEM_TYPE_PROTECTOR:
			fprintf(fp, " ");

			/* ԕωւ̑ϐ */
			str_printstate(fp, ptr_itemdata[i].param.protector.state_defense, STATE_ALL & ~STATE_OTHER);
			break;

		case ITEM_TYPE_TOOL:
			fprintf(fp, " ");

			/* ԕω */
			str_printstate(fp, ptr_itemdata[i].param.tool.state_effect, STATE_ALL);
			break;
		}

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writeflag(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_nothing = "..";
	const char *str_chrmain[] =
	{
		"CD",	/* ZV(ÍRm) */
		"Ca",	/* JC */
		"RC",	/* fBA(q) */
		"Te",	/* e */
		"Gi",	/* Mo[g */
		"Ro",	/* [U */
		"Ya",	/*  */
		"Pa",	/* p */
		"Po",	/* | */
		"CP",	/* ZV(pfB) */
		"Ci",	/* Vh */
		"RA",	/* fBA(l) */
		"Ed",	/* GbW */
		"Fu",	/* t[X[ */
		"??",	/* s */
		"Go",	/* Sx[U */
	};
	const char *str_direction[2][2] =
	{
		{"",	"/G"},
		{"G",		"G/"},
	};
	const char *str_range[2][2] =
	{
		{"P",	"P"},
		{"S",	"P/S"},
	};

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ACẽtOnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ACẽtOnf[^o͂(i) */
	fprintf(fp, "No.      O               "
				"     "
				"\LN^                               "
				"L푰\n");
	for (i = REGION_WEAPON_BEGIN; i <= REGION_PROTECTOR_END; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :", i, i, item_getname(i));

		switch (ptr_itemdata[i].type)
		{
		case ITEM_TYPE_WEAPON:
			/* etO */
			fprintf(fp, "  %2s   %2s   %2s   %2s  ",
					str_getflag(ptr_itemdata[i].param.weapon.flag_magnetic),
					str_getflag(ptr_itemdata[i].param.weapon.flag_throw),
					str_getflag(ptr_itemdata[i].param.weapon.flag_backrow),
					str_getflag(ptr_itemdata[i].param.weapon.flag_2handed));

			/* \LN^ */
			for (j = 0; j < 16; j++)
				fprintf(fp, " %-2s", ((ptr_itemdata[i].param.weapon.bits_equip_chr & (1U << j)) != 0x00) ? str_chrmain[j] : str_nothing);

			/* LX^[^Cv */
			if (ptr_itemdata[i].param.weapon.montype_attack != 0x00)
			{
				fprintf(fp, "  ");
				str_printmontype(fp, ptr_itemdata[i].param.weapon.montype_attack, MONTYPE_ALL);
			}
			break;

		case ITEM_TYPE_PROTECTOR:
			/* etO */
			fprintf(fp, "  %2s                 ",
					str_getflag(ptr_itemdata[i].param.protector.flag_magnetic));

			/* \LN^ */
			for (j = 0; j < 16; j++)
				fprintf(fp, " %-2s", ((ptr_itemdata[i].param.protector.bits_equip_chr & (1U << j)) != 0x00) ? str_chrmain[j] : str_nothing);

			/* LX^[^Cv */
			if (ptr_itemdata[i].param.protector.montype_defense != 0x00)
			{
				fprintf(fp, "  ");
				str_printmontype(fp, ptr_itemdata[i].param.protector.montype_defense, MONTYPE_ALL);
			}
			break;
		}

		fputc('\n', fp);
	}

	/* ACẽtOnf[^o͂(iȊO) */
	fprintf(fp, "No.      O               "
				"{X 1    "
				"ʔ͈\n");
	for (i = REGION_TOOL_BEGIN; i <= REGION_OTHER_END; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :", i, i, item_getname(i));

		switch (ptr_itemdata[i].type)
		{
		case ITEM_TYPE_TOOL:
			/* ʔ͈ */
			fprintf(fp, "  %2s   %2s   %2s   %-7s %-9s",
					str_getflag(!ptr_itemdata[i].param.tool.flag_boss_noeffect),
					str_getflag(!ptr_itemdata[i].param.tool.flag_once),
					str_getflag(!ptr_itemdata[i].param.tool.flag_reflect),
					str_direction[ptr_itemdata[i].param.tool.direction][ptr_itemdata[i].param.tool.change],
					str_range[ptr_itemdata[i].param.tool.range][ptr_itemdata[i].param.tool.change]);
			break;

		case ITEM_TYPE_OTHER:
			break;
		}

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writeeffect(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;
	const char *str_nothing = "----------------";

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ACěʃf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ACěʃf[^o͂ */
	fprintf(fp, "No.      O               "
				"gp\n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :", i, i, item_getname(i));
		if (ptr_itemdata[i].type == ITEM_TYPE_WEAPON)
		{
			fprintf(fp, " %-16s",
					(ptr_itemdata[i].param.weapon.effect_no != 0x00) ?
					mag_getname(ptr_itemdata[i].param.weapon.effect_no) : str_nothing);
		}
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writeparambug(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ACe̊ef[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ACe̊ef[^́Eo͂ */
	fprintf(fp, "No.      O               U\n");
	for (i = REGION_TOOL_BEGIN; i < NUM_ITEM; i++)
	{
		weapon_t weapon;
		unsigned char buf[SIZEOF_ITEMDATA_EQUIP];

		file_read(buf, FILEPOS_ITEMDATA + SIZEOF_ITEMDATA_EQUIP * NUM_ITEM_EQUIP + SIZEOF_ITEMDATA_EQUIP * (i - NUM_ITEM_EQUIP), SIZEOF_ITEMDATA_EQUIP);

		weapon.attack_power = buf[1];
		weapon.hit_rate     = buf[2] & 0x7f;
		weapon.effect_no    = buf[3];

		fprintf(fp, "%03d(%02X): %-16s :    %3d\n", i, i,
				item_getname(i),
				weapon.attack_power);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}
