/*
 *	ff4magic.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "ff4file.h"
#include "ff4decode.h"
#include "ff4magic.h"

#define FILEPOS_MAGNAME			(0x034900 + 0x044000)	/* = 0x078900 */
#define FILEPOS_MAGDATA			(0x0357a0 + 0x044000)	/* = 0x0797a0 */
#define NUM_MAG_SHORT			72
#define NUM_MAG_LONG			184
#define NUM_MAG					(NUM_MAG_SHORT + NUM_MAG_LONG)	/* = 256 */
#define SIZEOF_MAGNAME_SHORT	6
#define SIZEOF_MAGNAME_LONG		8
#define SIZEOF_MAGDATA			6
#define REGION_SHORT_BEGIN		0
#define REGION_SHORT_END		(REGION_SHORT_BEGIN + NUM_MAG_SHORT - 1)	/* =  71 */
#define REGION_LONG_BEGIN		(REGION_SHORT_END + 1)						/* =  72 */
#define REGION_LONG_END			(REGION_LONG_BEGIN + NUM_MAG_LONG - 1)		/* = 255 */

static void mag_setname(void);

static char tbl_magname[NUM_MAG][256];	/* @̖Oi[e[u */
static int flag_setmagname = 0;			/* tbl_magnameɃf[^i[Ă邩ǂ */

const char *mag_getname(int index)
{
	/* `FbN */
	if (index < 0 || NUM_MAG <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setmagname)
		mag_setname();

	return tbl_magname[index];
}

static void mag_setname(void)
{
	int i;
	long filepos;

	/* @̖Of[^ */
	for (i = 0; i < NUM_MAG; i++)
		*tbl_magname[i] = '\0';

	/* @̖Of[^ǂݍ */
	filepos = FILEPOS_MAGNAME;
	for (i = 0; i < NUM_MAG; i++)
	{
		int origin = 0;
		size_t len_read = 0;
		unsigned char buf[SIZEOF_MAGNAME_LONG];

		if (i < REGION_LONG_BEGIN)
		{
			origin = 1;
			len_read = SIZEOF_MAGNAME_SHORT;
		}
		else
		{
			origin = 0;
			len_read = SIZEOF_MAGNAME_LONG;
		}

		filepos += file_read(buf, filepos, len_read);
		dec_string(tbl_magname[i], &buf[origin], sizeof(tbl_magname[i]), len_read - origin, 0);
	}

	flag_setmagname = 1;
}

magdata_t *mag_readdata(void)
{
	int i;
	magdata_t *buf_magdata = NULL;

	/* @̊ef[^i[obt@mۂ */
	if ((buf_magdata = malloc(sizeof(magdata_t) * NUM_MAG)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* @̊ef[^ǂݍ */
	for (i = 0; i < NUM_MAG; i++)
	{
		unsigned char buf[SIZEOF_MAGDATA];

		file_read(buf, FILEPOS_MAGDATA + SIZEOF_MAGDATA * i, SIZEOF_MAGDATA);
		mag_setparam(&buf_magdata[i], buf, sizeof(buf));
	}

	return buf_magdata;
}

int mag_setparam(magdata_t *ptr_magdata, unsigned char *ptr_buf, size_t bufsize)
{
	/* `FbN */
	if (ptr_magdata == NULL || ptr_buf == NULL || bufsize < SIZEOF_MAGDATA)
		return 0;

	/* lnf[^ */
	ptr_magdata->magic_point  = ptr_buf[5] & 0x7f;
	ptr_magdata->effect_type  = ptr_buf[3] & 0x7f;
	ptr_magdata->hit_rate     = ptr_buf[2] & 0x7f;
	ptr_magdata->delay_time   = ptr_buf[0] & 0x0f;

	/* tOnf[^ */
	ptr_magdata->flag_boss_noeffect = ((ptr_buf[2] & 0x80) != 0x00);
	ptr_magdata->flag_once          = ((ptr_buf[3] & 0x80) == 0x00);
	ptr_magdata->flag_reflect       = ((ptr_buf[5] & 0x80) != 0x00);
	ptr_magdata->direction          = ((ptr_buf[0] & 0x80) != 0x00) ? 1 : 0;
	ptr_magdata->range              = ((ptr_buf[0] & 0x40) != 0x00) ? 1 : 0;
	ptr_magdata->change             = ((ptr_buf[0] & 0x20) != 0x00) ? 1 : 0;

	/* σf[^() */
	ptr_magdata->effect_power = 0;
	ptr_magdata->effect_rate  = 0;
	ptr_magdata->elem_effect  = elem_gettbl(ptr_buf[4] & 0x7f);		/*  */
	ptr_magdata->state_effect = state_gettbl(ptr_buf[4] & 0x7f);	/*  */

	switch (ptr_magdata->effect_type)
	{
	case 0x00:	/* U@() */
	case 0x01:	/* ԕω(Xbv)ƂȂU@(?) */
	case 0x04:	/* qbg|CgzU@() */
	case 0x05:	/* }WbN|CgzU@() */
		ptr_magdata->effect_power = ptr_buf[1];
		ptr_magdata->elem_effect = elem_gettbl(ptr_buf[4] & 0x7f);
		break;

	case 0x14:	/* 䗦_[WU@(?) */
	case 0x1e:	/* 䗦_[WU@(?) */
	case 0x25:	/* 䗦_[WU@(0x00) */
		ptr_magdata->effect_rate = ptr_buf[1];
		ptr_magdata->elem_effect = elem_gettbl(ptr_buf[4] & 0x7f);
		break;

	case 0x06:	/* ԕω@(ԕω1E2) */
	case 0x07:	/* ԕω@(ԕω1E2) */
	case 0x0b:	/* ԕωÖ@(ԕω1E2) */
	case 0x0c:	/* ԕω(؁ElE^)@(ԕω1E2) */
	case 0x17:	/* ԕω()@(ԕω1E2) */
	case 0x18:	/* ԕω(Ή)@(ԕω1E2) */
	case 0x19:	/* ԕω()@(ԕω1E2) */
	case 0x1f:	/* ԕω(XɐΉ)@(ԕω1E2) */
	case 0x23:	/* ԕω()@(ԕω1E2) */
	case 0x24:	/* ԕω(Ή)@(ԕω1E2) */
		ptr_magdata->effect_power = ptr_buf[1];
		ptr_magdata->state_effect = state_gettbl(ptr_buf[4] & 0x7f);
		break;

	case 0x08:	/* ԕω(g)@(ԕω3E4?) */
	case 0x09:	/* ԕω(tN)@(ԕω3E4?) */
	case 0x11:	/* ԕω(Xgbv)@(ԕω3E4?) */
		ptr_magdata->effect_power = ptr_buf[1];
		ptr_magdata->state_effect = state_gettbl(ptr_buf[4] & 0x7f) << 16;
		break;

	case 0x02:	/* 񕜖@(0x00) */
	case 0x03:	/* mU@(0x00) */
	case 0x0a:	/* 퓬s\Ö@(0x00) */
	case 0x0d:	/* h͕ω@(0x00) */
	case 0x0e:	/* @h͕ω@(0x00) */
	case 0x0f:	/* Xs[hω@(0x00) */
	case 0x10:	/* fBXy(0x00) */
	case 0x12:	/* Cun@(0x00) */
	case 0x13:	/* e|n@(0x00) */
	case 0x15:	/* }WbN|Cg(0x00) */
	case 0x16:	/* qbg|CgE}WbN|CgS(0x00) */
	case 0x1a:	/* ̂₫(0x00) */
	case 0x1b:	/* Ă(0x00) */
	case 0x1d:	/* ̂񂱂(0x00) */
	case 0x20:	/* (0x00) */
	case 0x21:	/* uX^[(0x00) */
	case 0x22:	/* ЂĂ(0x00) */
		ptr_magdata->effect_power = ptr_buf[1];
		break;
	}

	return 1;
}

int mag_writeparam(const char *ptr_filename, const magdata_t *ptr_magdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_magdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* @̐lnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* @̐lnf[^o͂ */
	fprintf(fp, "No.      O               "
				"MP ʒl   䗦    \n");
	for (i = 0; i < NUM_MAG; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :  %5d, %5d, %2s%3d, %5d, %5d\n", i, i,
				mag_getname(i),
				ptr_magdata[i].magic_point,
				ptr_magdata[i].effect_power,
				(ptr_magdata[i].effect_rate != 0) ? "1/" : "  ",
				ptr_magdata[i].effect_rate,
				ptr_magdata[i].hit_rate,
				ptr_magdata[i].delay_time);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mag_writeelem(const char *ptr_filename, const magdata_t *ptr_magdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_magdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* @̑nf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* @̑nf[^o͂ */
	fprintf(fp, "No.      O               "
				"\n");
	for (i = 0; i < NUM_MAG; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s : ", i, i, mag_getname(i));

		/*  */
		str_printelem(fp, ptr_magdata[i].elem_effect, ELEM_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mag_writestate(const char *ptr_filename, const magdata_t *ptr_magdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_magdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* @̏ԕωnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* @̏ԕωnf[^o͂ */
	fprintf(fp, "No.      O               "
				"ԕω1        2                3    4\n");
	for (i = 0; i < NUM_MAG; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s : ", i, i, mag_getname(i));

		/* ԕω */
		str_printstate(fp, ptr_magdata[i].state_effect, STATE_ALL /* & ~STATE_OTHER */);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mag_writeflag(const char *ptr_filename, const magdata_t *ptr_magdata)
{
	int i;
	FILE *fp = NULL;
	const char *str_direction[2][2] =
	{
		{"",	"/G"},
		{"G",		"G/"},
	};
	const char *str_range[2][2] =
	{
		{"P",	"P"},
		{"S",	"P/S"},
	};

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_magdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* @̃tOnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* @̃tOnf[^o͂ */
	fprintf(fp, "No.      O               "
				"{X 1    "
				"ʔ͈\n");
	for (i = 0; i < NUM_MAG; i++)
	{
		/* ʔ͈ */
		fprintf(fp, "%03d(%02X): %-16s :  %2s   %2s   %2s   %-7s %-9s\n", i, i, mag_getname(i),
				str_getflag(!ptr_magdata[i].flag_boss_noeffect),
				str_getflag(!ptr_magdata[i].flag_once),
				str_getflag(!ptr_magdata[i].flag_reflect),
				str_direction[ptr_magdata[i].direction][ptr_magdata[i].change],
				str_range[ptr_magdata[i].range][ptr_magdata[i].change]);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}
