/*
 *	ff4monster.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include "errmes.h"
#include "ff4file.h"
#include "ff4decode.h"
#include "ff4character.h"
#include "ff4monster.h"

#define FILEPOS_MONNAME			(0x02d800 + 0x044000)	/* = 0x071800 */
#define FILEPOS_MONEXP			(0x02e1c0 + 0x044000)	/* = 0x0721c0 */
#define FILEPOS_MONGIL			(0x02e000 + 0x044000)	/* = 0x072000 */
#define FILEPOS_MONITEM			(0x02df00 + 0x044000)	/* = 0x071f00 */
#define FILEPOS_MONTRIPLET		(0x02e380 + 0x044000)	/* = 0x072380 */
#define FILEPOS_MONDATA_ADDR	(0x02e6a0 + 0x044000)	/* = 0x0726a0 */
#define FILEPOS_MONDATA			(0x02e860 + 0x044000)	/* = 0x072860 */
#define FILEPOS_MONACT1			(0x032900 + 0x044000)	/* = 0x076900 */
#define FILEPOS_MONACT2			(0x02f6c0 + 0x044000)	/* = 0x0736c0 */
#define FILEPOS_MONACTLIST		(0x032030 + 0x044000)	/* = 0x076030 */
#define FILEPOS_MONACTCND		(0x032700 + 0x044000)	/* = 0x076700 */
#define FILEPOS_MONACTCNDHP		(0x032000 + 0x044000)	/* = 0x076000 */
#define FILEPOS_MONACTCNDLIST	(0x032600 + 0x044000)	/* = 0x076600 */
#define FILEPOS_MONMES_ADDR		(0x033200 + 0x044000)	/* = 0x077200 */
#define FILEPOS_MONMES			(0x033374 + 0x044000)	/* = 0x077374 */
#define FILEPOS_MONPARTY		(0x02c000 + 0x044000)	/* = 0x070000 */
#define FILEPOS_MONENC			(0x030796 + 0x044000)	/* = 0x074796 */
#define FILEPOS_MONMAP			(0x030542 + 0x044000)	/* = 0x074542 */
#define NUM_MON					224
#define NUM_MONITEM				64
#define NUM_MONTRIPLET			224
#define NUM_MONACT1				254
#define NUM_MONACT2				91
#define NUM_MONACTLIST			254
#define NUM_MONACTCND			80
#define NUM_MONACTCNDHP			16
#define NUM_MONACTCNDLIST		98
#define NUM_MONMES				186
#define NUM_MONPARTY			512
#define NUM_MONENC				96
#define NUM_MONMAP				3
#define SIZEOF_MONNAME			8
#define SIZEOF_MONEXP			2
#define SIZEOF_MONGIL			2
#define SIZEOF_MONTRIPLET		3
#define SIZEOF_MONITEM			4
#define SIZEOF_MONACTCND		4
#define SIZEOF_MONACTCNDHP		2
#define SIZEOF_MONPARTY			8
#define SIZEOF_MONENC			8
#define SIZEOF_OFFSETADDR		2

static void mon_setname(void);
static void mon_settriplet(void);
static void mon_setmes(void);
static void mon_setactcnd(void);
static const char *mon_getmes(int index);
static triplet_t mon_gettriplet(int index);
static const char *mon_getactcnd(int index);
static const char *mon_gettarget(int index);
static unsigned char *mon_printact(FILE *stream, unsigned char *buf_action);

static char tbl_monname[NUM_MON][256];			/* X^[̖Oi[e[u */
static char tbl_monmes[NUM_MONMES][256];		/* X^[̃bZ[Wi[e[u */
static triplet_t tbl_montriplet[NUM_MONTRIPLET];/* ep[^f[^(񐔁ElEm)i[e[u */
static char tbl_monactcnd[NUM_MONACTCND][256];	/* X^[sp^[̓Kp()i[e[u */
static int flag_setmonname    = 0;				/* tbl_monnameɃf[^i[Ă邩ǂ */
static int flag_setmonmes     = 0;				/* tbl_monmesɃf[^i[Ă邩ǂ */
static int flag_setmontriplet = 0;				/* tbl_montripletɃf[^i[Ă邩ǂ */
static int flag_setmonactcnd  = 0;				/* tbl_monactcndɃf[^i[Ă邩ǂ */

static const int tbl_sizeof_map[NUM_MONMAP] = {8, 4, 2};

const char *mon_getname(int index)
{
	/* `FbN */
	if (index < 0 || NUM_MON <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setmonname)
		mon_setname();

	return tbl_monname[index];
}

static const char *mon_getmes(int index)
{
	/* `FbN */
	if (index < 0 || NUM_MONMES <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setmonmes)
		mon_setmes();

	return tbl_monmes[index];
}

static triplet_t mon_gettriplet(int index)
{
	const triplet_t tmp_param = {0, 0, 0};

	/* `FbN */
	if (index < 0 || NUM_MONTRIPLET <= index)
	{
		assert(0);
		return tmp_param;
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setmontriplet)
		mon_settriplet();

	return tbl_montriplet[index];
}

static const char *mon_getactcnd(int index)
{
	/* `FbN */
	if (index < 0 || NUM_MONACTCND <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setmonactcnd)
		mon_setactcnd();

	return tbl_monactcnd[index];
}

static const char *mon_gettarget(int index)
{
	static char str_buf[256];	/* ēs */
	const int target_no = index & 0x7f;
	const int flag_anyone = ((index & 0x80) != 0x00);
	static const char *str_target[] =
	{
		/* 0x16`0x2f */
		"ƓO[ṽX^[", "X^[g",
		"ȊÕX^[S", "GS", "O[v1̃X^[", "O[v2̃X^[",
		"O[v3̃X^[", "1Dh", "1Eh", "1Fh",
		"20h", "21h", "G", "ȊO̓G",
		"X^[", "ȊÕX^[", "26h", "27h",
		"LN^S", "퓬s\̃X^[S", "2Ah", "2Bh",
		"2Ch", "2Dh", "2Eh", "2Fh",
	};

	assert(sizeof(str_target) / sizeof(str_target[0]) == 0x30 - 0x16);

	if (target_no < 0x16)
		sprintf(str_buf, "%.12s(%.12s)", chr_getname_id(target_no), chr_getjobname_id(target_no));
	else if (index == 0x98)		/* 0x98ʈ */
		sprintf(str_buf, str_target[18]);
	else if (target_no < 0x30)
		sprintf(str_buf, "%s%s", str_target[target_no - 0x16], (flag_anyone) ? "̒N" : "");
	else
		sprintf(str_buf, "%02Xh", target_no);

	return str_buf;
}

static void mon_setname(void)
{
	int i;

	/* X^[̖Of[^ */
	for (i = 0; i < NUM_MON; i++)
		*tbl_monname[i] = '\0';

	/* X^[̖Of[^ǂݍ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONNAME];

		file_read(buf, FILEPOS_MONNAME + SIZEOF_MONNAME * i, SIZEOF_MONNAME);
		dec_string(tbl_monname[i], buf, sizeof(tbl_monname[i]), SIZEOF_MONNAME, 0);
	}

	flag_setmonname = 1;
}

static void mon_setmes(void)
{
	int i;
	long filepos;
	size_t bufsize;
	int tbl_offset[NUM_MONMES + 1];

	/* X^[̃bZ[Wf[^ */
	for (i = 0; i < NUM_MONMES; i++)
		*tbl_monmes[i] = '\0';

	/* X^[̃bZ[Wf[^i[ĂAhX(ւ̃ItZbg)ǂݍ */
	for (i = 0; i < NUM_MONMES; i++)
	{
		unsigned char buf[SIZEOF_OFFSETADDR];

		file_read(buf, FILEPOS_MONMES_ADDR + SIZEOF_OFFSETADDR * i, SIZEOF_OFFSETADDR);
		tbl_offset[i] = (buf[1] << 8) | buf[0];
	}
	tbl_offset[NUM_MONMES] = tbl_offset[NUM_MONMES - 1] + 0x0a;	/*  */

	/* X^[̃bZ[Wf[^ǂݍ */
	filepos = FILEPOS_MONMES;
	for (i = 0; i < NUM_MONMES; i++)
	{
		unsigned char *buf = NULL;

		if (tbl_offset[i + 1] - tbl_offset[i] <= 0)
			break;

		bufsize = tbl_offset[i + 1] - tbl_offset[i];
		if ((buf = malloc(bufsize)) == NULL)
			err_printf(STR_FATAL_ERROR_IN_MALLOC);

		filepos += file_read(buf, filepos, bufsize);
		if (bufsize > 1)
			dec_string(tbl_monmes[i], buf, sizeof(tbl_monmes[i]), bufsize - 1, 1);

		free(buf);
		buf = NULL;
	}

	flag_setmonmes = 1;
}

static void mon_settriplet(void)
{
	int i;

	/* X^[̊ep[^f[^(񐔁ElEm)ǂݍ */
	for (i = 0; i < NUM_MONTRIPLET; i++)
	{
		unsigned char buf[SIZEOF_MONTRIPLET];

		file_read(buf, FILEPOS_MONTRIPLET + SIZEOF_MONTRIPLET * i, SIZEOF_MONTRIPLET);
		tbl_montriplet[i].times = buf[0];
		tbl_montriplet[i].power = buf[2];
		tbl_montriplet[i].rate  = buf[1];
	}

	flag_setmontriplet = 1;
}

static void mon_setactcnd(void)
{
	int i;
	long tbl_hit_point[NUM_MONACTCNDHP];

	/* X^[̍sp^[̓Kp */
	for (i = 0; i < NUM_MONACTCND; i++)
		*tbl_monactcnd[i] = '\0';

	/* qbg|CgElǂݍ */
	for (i = 0; i < NUM_MONACTCNDHP; i++)
	{
		unsigned char buf[SIZEOF_MONACTCNDHP];

		file_read(buf, FILEPOS_MONACTCNDHP + SIZEOF_MONACTCNDHP * i, SIZEOF_MONACTCNDHP);
		tbl_hit_point[i] = (buf[1] << 8) | buf[0];
	}

	/* X^[̍sp^[̓Kpǂݍ */
	for (i = 0; i < NUM_MONACTCND; i++)
	{
		unsigned char buf[SIZEOF_MONACTCND];

		file_read(buf, FILEPOS_MONACTCND + SIZEOF_MONACTCND * i, SIZEOF_MONACTCND);

		if (buf[0] == 0x00)
		{
			/* Ԃ̏Ԃƈv(00 xx yy zz) */
			/*   xx: Ώ */
			/*   yy: ԕωrbg̉oCgڂgp邩(0`3) */
			/*   zz: ԕωrbg */
			int index;
			char str_state[256] = "";
			char *ptr = str_state;
			state_t state = 0x00;

			state = bits_make32((buf[2] == 0x00) ? buf[3] : 0x00,
								(buf[2] == 0x01) ? buf[3] : 0x00,
								(buf[2] == 0x02) ? buf[3] : 0x00,
								(buf[2] == 0x03) ? buf[3] : 0x00);

			for (index = 0; index < NUM_STATE; index++)
				ptr += sprintf(ptr, "%s", str_getstate(state, index));

			sprintf(tbl_monactcnd[i], "%s̏Ԃ \"%s\" ł", mon_gettarget(buf[1]), str_state);
		}
		else if (buf[0] == 0x01)
		{
			/* qbg|Cglȉ(01 xx 00 zz) */
			/*   xx: Ώ */
			/*   zz: Eqbg|Cge[ũCfbNX */
			sprintf(tbl_monactcnd[i], "%sHP%ldȉł", mon_gettarget(buf[1]), tbl_hit_point[buf[3]]);
		}
		else if (buf[0] == 0x02)
		{
			/* ϐ̒lƈv(02 00 yy zz) */
			/*   yy: ǂ̓ϐlgp邩 */
			/*   zz: ϐ̒l */
			sprintf(tbl_monactcnd[i], "ϐ%d \"%d\" ł", buf[2], buf[3]);
		}
		else if (buf[0] == 0x03)
		{
			/* Ώۂ݂(03 xx 01 00) */
			/*   xx: Ώ */
			sprintf(tbl_monactcnd[i], "%s݂", mon_gettarget(buf[1]));
		}
		else if (buf[0] == 0x04)
		{
			/* Ώ(X^[)݂Ȃ(04 00 01 zz) */
			/*   zz: ΏۃX^[̔ԍ */
			sprintf(tbl_monactcnd[i], "%s݂Ȃ", mon_getname(buf[3]));
		}
		else if (buf[0] == 0x05)
		{
			/* X^[p[eBԍ̒lƈv(05 00 yy zz) */
			/*   yy: X^[p[eBԍ(ʃoCg) */
			/*   zz: X^[p[eBԍ(ʃoCg) */
			int party_no = (buf[2] << 8) | buf[3];

			sprintf(tbl_monactcnd[i], "X^[p[eBԍ%d(%03Xh)ł", party_no, party_no);
		}
		else if (buf[0] == 0x06)
		{
			/* X^[g̃O[v݂Ȃ(06 00 00 00) */
			sprintf(tbl_monactcnd[i], "X^[g̃O[v݂Ȃ");
		}
		else if (buf[0] == 0x07 || buf[0] == 0x08)
		{
			/* ̑Ώۂ̍sɑ΂JE^[(07 xx yy zz) */
			/* GS̍sɑ΂JE^[(08 19 yy zz) */
			/*   xx: Ώ */
			/*   yy: @ER}h̔ԍ */
			/*   zz:  */
			int index;
			char str_attack[256] = "";
			char *ptr = str_attack;

			switch (buf[2])
			{
			case 0xc0:
				ptr += sprintf(ptr, "U");
				break;
			case 0xc2:
				ptr += sprintf(ptr, "@");
				break;
			case 0xc4:
				ptr += sprintf(ptr, "@");
				break;
			case 0xde:
				ptr += sprintf(ptr, "Wv");
				break;
			case 0xe0:
				ptr += sprintf(ptr, "ӂ肪");
				break;
			}

			if (buf[3] != 0x00)
			{
				elem_t element = buf[3];

				ptr += sprintf(ptr, "(");
				for (index = 0; index < NUM_ELEMENT; index++)
					ptr += sprintf(ptr, "%s", str_getelem(element, index));
				ptr += sprintf(ptr, ")");
			}

			sprintf(tbl_monactcnd[i], "%s%s%s󂯂",
					mon_gettarget(buf[1]),
					((buf[1] & 0x7f) >= 0x16 && (buf[1] & 0x80) == 0x00) ? "̒N" : "",
					str_attack);
		}
		else if (buf[0] == 0x09)
		{
			/* {XƂ̐퓬ł͂Ȃ(09 00 00 00) */
			sprintf(tbl_monactcnd[i], "{XƂ̐퓬ł͂Ȃ");
		}
		else if (buf[0] == 0x0a)
		{
			/* _[WJE^[(0A 00 00 00) */
			sprintf(tbl_monactcnd[i], "_[W󂯂");
		}
		else if (buf[0] == 0x0b)
		{
			/* ȊÕX^[Ȃ(0B 00 00 00) */
			sprintf(tbl_monactcnd[i], "ȊÕX^[Ȃ");
		}
		else
		{
			/* ʏ펞(FF FF FF FF) */
			sprintf(tbl_monactcnd[i], "ʏ펞");
		}
	}

	flag_setmonactcnd = 1;
}

mondata_t *mon_readdata(void)
{
	int i, j;
	long filepos;
	mondata_t *buf_mondata = NULL;
	int tbl_offset[NUM_MON + 1];
	int tbl_item[NUM_MONITEM][SIZEOF_MONITEM];

	/* X^[̊ef[^i[obt@mۂ */
	if ((buf_mondata = malloc(sizeof(mondata_t) * NUM_MON)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* X^[̊lolf[^ǂݍ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONEXP];

		file_read(buf, FILEPOS_MONEXP + SIZEOF_MONEXP * i, SIZEOF_MONEXP);
		buf_mondata[i].experience_point = (buf[1] << 8) | buf[0];
	}

	/* X^[̏Mf[^ǂݍ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONGIL];

		file_read(buf, FILEPOS_MONGIL + SIZEOF_MONGIL * i, SIZEOF_MONGIL);
		buf_mondata[i].gil = (buf[1] << 8) | buf[0];
	}

	/* X^[̏ACef[^ǂݍ */
	for (i = 0; i < NUM_MONITEM; i++)
	{
		unsigned char buf[SIZEOF_MONITEM];

		file_read(buf, FILEPOS_MONITEM + SIZEOF_MONITEM * i, SIZEOF_MONITEM);
		for (j = 0; j < SIZEOF_MONITEM; j++)
			tbl_item[i][j] = buf[j];
	}

	/* X^[̊ef[^i[ĂAhX(ւ̃ItZbg)ǂݍ */
	for (i = 0; i < NUM_MON - 1; i++)
	{
		unsigned char buf[SIZEOF_OFFSETADDR];

		file_read(buf, FILEPOS_MONDATA_ADDR + SIZEOF_OFFSETADDR * i, SIZEOF_OFFSETADDR);
		tbl_offset[i] = (buf[1] << 8) | buf[0];
	}
	tbl_offset[NUM_MON] = tbl_offset[NUM_MON - 1] = tbl_offset[NUM_MON - 2] + 0x0a;	/*  */

	/* X^[̊ef[^ǂݍ */
	filepos = FILEPOS_MONDATA;
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[256];
		unsigned char *ptr_buf = NULL;
		triplet_t attack;
		triplet_t defense;
		triplet_t magic_defense;

		if (tbl_offset[i + 1] - tbl_offset[i] < 0)
			break;

		memset(buf, 0x00, sizeof(buf));
		filepos += file_read(buf, filepos, tbl_offset[i + 1] - tbl_offset[i]);

		/* Œ蒷 */
		attack        = mon_gettriplet(buf[3]);
		defense       = mon_gettriplet(buf[4]);
		magic_defense = mon_gettriplet(buf[5]);
		buf_mondata[i].hit_point           = (buf[2] << 8) | buf[1];
		buf_mondata[i].level               = buf[0] & 0x7f;
		buf_mondata[i].attack_times        = attack.times;
		buf_mondata[i].attack_power        = attack.power;
		buf_mondata[i].hit_rate            = attack.rate;
		buf_mondata[i].defense_times       = defense.times;
		buf_mondata[i].defense_power       = defense.power;
		buf_mondata[i].evade_rate          = defense.rate;
		buf_mondata[i].magic_defense_times = magic_defense.times;
		buf_mondata[i].magic_defense_power = magic_defense.power;
		buf_mondata[i].magic_evade_rate    = magic_defense.rate;
		buf_mondata[i].speed               = buf[6];
		buf_mondata[i].item_get            = (buf[7] & 0xc0) >> 6;
		buf_mondata[i].flag_boss           = ((buf[0] & 0x80) != 0x00);
		buf_mondata[i].actptn_no[0]        = buf[8];

		/* ACef[^ */
		for (j = 0; j < 4; j++)
			buf_mondata[i].item_no[j] = tbl_item[buf[7] & 0x3f][j];

		/* ϒ */
		buf_mondata[i].elem_defense  = 0x00;
		buf_mondata[i].elem_weakness = 0x00;
		buf_mondata[i].state_attack  = 0x00;
		buf_mondata[i].state_defense = 0x00;
		buf_mondata[i].montype       = 0x00;
		buf_mondata[i].magic_power   = 0;
		buf_mondata[i].actptn_no[1]  = -1;
		ptr_buf = &buf[10];

		if ((buf[9] & 0x80) != 0x00)	/* ʏȔԕω */
		{
			ptr_buf++;	/* buf[10]Ɋi[Ăl͕K0x00Ȃ̂œǂݔ΂ */
			buf_mondata[i].state_attack = *ptr_buf++;
			buf_mondata[i].state_attack |= (state_t)*ptr_buf++ << 8;
		}

		if ((buf[9] & 0x40) != 0x00)	/* Uւ̑ϐ & ԕωւ̑ϐ */
		{
			buf_mondata[i].elem_defense = *ptr_buf++;
			buf_mondata[i].state_defense = *ptr_buf++;
			buf_mondata[i].state_defense |= (state_t)*ptr_buf++ << 8;
		}

		if ((buf[9] & 0x20) != 0x00)	/* _̑ */
			buf_mondata[i].elem_weakness = *ptr_buf++;

		if ((buf[9] & 0x10) != 0x00)	/*  */
			buf_mondata[i].magic_power = *ptr_buf++;

		if ((buf[9] & 0x08) != 0x00)	/* X^[^Cv */
			buf_mondata[i].montype = *ptr_buf++;

		if ((buf[9] & 0x04) != 0x00)	/* sp^[̔ԍ2 */
			buf_mondata[i].actptn_no[1] = *ptr_buf++;

		/* p[^Zołf[^ */
		buf_mondata[i].magic_point        = buf_mondata[i].hit_point / 16;
		buf_mondata[i].magic_attack_times = 1 + buf_mondata[i].magic_power / 4;
	}

	return buf_mondata;
}

monact_t *mon_readact(void)
{
	int i, j, k;
	unsigned char *buf = NULL;
	unsigned char *ptr_buf = NULL;
	monact_t *buf_monact = NULL;
	int tbl_condition_no[NUM_MONACTCNDLIST][3];
	struct acttbl_tag
	{
		size_t action_size;		/* sp^[f[^̃oCg */
		unsigned char *action;	/* sp^[f[^ */
	};
	struct acttbl_tag tbl_action1[NUM_MONACT1], tbl_action2[NUM_MONACT2];
	const size_t bufsize = 2048;
	const int tbl_actselect[/* NUM_MONACTLIST */] =
	{
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 1, 0, 0, 0, 0,
		0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0,
		0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1,
		1, 1, 1, 1, 1, 0, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1,
		1, 1, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0,
	};

	assert(sizeof(tbl_actselect) / sizeof(tbl_actselect[0]) == NUM_MONACTLIST);

	/* X^[̍sp^[f[^i[obt@mۂ */
	if ((buf_monact = malloc(sizeof(monact_t) * NUM_MONACTLIST)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);
	for (i = 0; i < NUM_MONACTLIST; i++)
	{
		for (j = 0; j < sizeof(buf_monact[i].ptn) / sizeof(buf_monact[i].ptn[0]); j++)
		{
			buf_monact[i].ptn[j].action_size = 0;
			buf_monact[i].ptn[j].action = NULL;

			for (k = 0; k < 3; k++)
				buf_monact[i].ptn[j].condition_no[k] = -1;
		}
	}

	/* ee[ȕs */
	for (i = 0; i < NUM_MONACTCNDLIST; i++)
	{
		for (j = 0; j < 3; j++)
			tbl_condition_no[i][j] = -1;
	}
	for (i = 0; i < NUM_MONACT1; i++)
	{
		tbl_action1[i].action_size = 0;
		tbl_action1[i].action = NULL;
	}
	for (i = 0; i < NUM_MONACT2; i++)
	{
		tbl_action2[i].action_size = 0;
		tbl_action2[i].action = NULL;
	}

	/* X^[sXgǂݍ */
	if ((buf = malloc(bufsize)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);
	memset(buf, 0x00, bufsize);
	file_read(buf, FILEPOS_MONACTCNDLIST, bufsize);

	ptr_buf = buf;
	for (i = 0; i < NUM_MONACTCNDLIST; i++)
	{
		unsigned char *ptr_next = ptr_buf;

		if (ptr_buf - buf > (ptrdiff_t)bufsize)
			break;
		if ((ptr_next = memchr(ptr_buf, 0xff, bufsize - (ptr_buf - buf))) == NULL)
			break;

		for (j = 0; j < ptr_next - ptr_buf; j++)
			tbl_condition_no[i][j] = ptr_buf[j];

		ptr_buf = ptr_next + 1;
	}

	free(buf);
	buf = NULL;

	/* X^[sp^[1ǂݍ */
	if ((buf = malloc(bufsize)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);
	memset(buf, 0x00, bufsize);
	file_read(buf, FILEPOS_MONACT1, bufsize);

	ptr_buf = buf;
	for (i = 0; i < NUM_MONACT1; i++)
	{
		unsigned char *ptr_next = ptr_buf;

		if (ptr_buf - buf > (ptrdiff_t)bufsize)
			break;
		if ((ptr_next = memchr(ptr_buf, 0xff, bufsize - (ptr_buf - buf))) == NULL)
			break;

		tbl_action1[i].action_size = ptr_next - ptr_buf;
		if ((tbl_action1[i].action = malloc(tbl_action1[i].action_size)) == NULL)
			err_printf(STR_FATAL_ERROR_IN_MALLOC);
		memmove(tbl_action1[i].action, ptr_buf, tbl_action1[i].action_size);

		ptr_buf = ptr_next + 1;
	}

	free(buf);
	buf = NULL;

	/* X^[sp^[2ǂݍ */
	if ((buf = malloc(bufsize)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);
	memset(buf, 0x00, bufsize);
	file_read(buf, FILEPOS_MONACT2, bufsize);

	ptr_buf = buf;
	for (i = 0; i < NUM_MONACT2; i++)
	{
		unsigned char *ptr_next = ptr_buf;

		if (ptr_buf - buf > (ptrdiff_t)bufsize)
			break;
		if ((ptr_next = memchr(ptr_buf, 0xff, bufsize - (ptr_buf - buf))) == NULL)
			break;

		tbl_action2[i].action_size = ptr_next - ptr_buf;
		if ((tbl_action2[i].action = malloc(tbl_action2[i].action_size)) == NULL)
			err_printf(STR_FATAL_ERROR_IN_MALLOC);
		memmove(tbl_action2[i].action, ptr_buf, tbl_action2[i].action_size);

		ptr_buf = ptr_next + 1;
	}

	free(buf);
	buf = NULL;

	/* X^[sp^[f[^ݒ肷 */
	if ((buf = malloc(bufsize)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);
	memset(buf, 0x00, bufsize);
	file_read(buf, FILEPOS_MONACTLIST, bufsize);

	ptr_buf = buf;
	for (i = 0; i < NUM_MONACTLIST; i++)
	{
		unsigned char *ptr_next = ptr_buf;
		const struct acttbl_tag *ptr_action = (tbl_actselect[i] == 0) ? tbl_action1 : tbl_action2;

		if (ptr_buf - buf > (ptrdiff_t)bufsize)
			break;
		if ((ptr_next = memchr(ptr_buf, 0xff, bufsize - (ptr_buf - buf))) == NULL)
			break;

		if (ptr_next - ptr_buf >= 2)
		{
			for (j = 0; j < ptr_next - ptr_buf; j += 2)
			{
				if (j / 2 >= sizeof(buf_monact[i].ptn) / sizeof(buf_monact[i].ptn[0]))
					break;

				for (k = 0; k < 3; k++)
				{
					if (ptr_buf[j] < NUM_MONACTCNDLIST)
						buf_monact[i].ptn[j / 2].condition_no[k] = tbl_condition_no[ptr_buf[j]][k];
				}

				if (ptr_buf[j + 1] >= ((tbl_actselect[i] == 0) ? NUM_MONACT1 : NUM_MONACT2))
					break;

				buf_monact[i].ptn[j / 2].action_size = ptr_action[ptr_buf[j + 1]].action_size;
				if ((buf_monact[i].ptn[j / 2].action = malloc(buf_monact[i].ptn[j / 2].action_size)) == NULL)
					err_printf(STR_FATAL_ERROR_IN_MALLOC);
				memmove(buf_monact[i].ptn[j / 2].action, ptr_action[ptr_buf[j + 1]].action, buf_monact[i].ptn[j / 2].action_size);
			}
		}

		ptr_buf = ptr_next + 1;
	}

	/* ee[ũ */
	for (i = 0; i < NUM_MONACT1; i++)
	{
		free(tbl_action1[i].action);
		tbl_action1[i].action = NULL;
	}
	for (i = 0; i < NUM_MONACT2; i++)
	{
		free(tbl_action2[i].action);
		tbl_action2[i].action = NULL;
	}

	return buf_monact;
}

monparty_t *mon_readparty(void)
{
	int i, j;
	monparty_t *buf_monparty = NULL;

	/* X^[̃p[eBҐf[^i[obt@mۂ */
	if ((buf_monparty = malloc(sizeof(monparty_t) * NUM_MONPARTY)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* X^[̃p[eBҐ̊ef[^ǂݍ */
	for (i = 0; i < NUM_MONPARTY; i++)
	{
		unsigned char buf[SIZEOF_MONPARTY];

		file_read(buf, FILEPOS_MONPARTY + SIZEOF_MONPARTY * i, SIZEOF_MONPARTY);

		buf_monparty[i].flag_eggy[0]   = ((buf[0] & 0x80) != 0x00);
		buf_monparty[i].flag_eggy[1]   = ((buf[0] & 0x40) != 0x00);
		buf_monparty[i].flag_eggy[2]   = ((buf[0] & 0x20) != 0x00);
		buf_monparty[i].flag_exist[0]  = 0x00;
		buf_monparty[i].flag_exist[1]  = ((buf[4] & 0x01) != 0x00) ^ ((buf[4] & 0x02) != 0x00);
		buf_monparty[i].flag_exist[2]  = ((buf[4] & 0x02) != 0x00);
		buf_monparty[i].flag_escape    = ((buf[6] & 0x01) == 0x00);
		buf_monparty[i].flag_event     = ((buf[6] & 0x20) != 0x00);

		for (j = 0; j < 3; j++)
		{
			buf_monparty[i].num[j] = (buf[4] >> (6 - j * 2)) & 0x03;
			buf_monparty[i].monster_no[j] = buf[j + 1];
		}
	}

	return buf_monparty;
}

monenc_t *mon_readenc(void)
{
	int i, j;
	monenc_t *buf_monenc = NULL;
	const int tbl_encselect[/* NUM_MONENC */] =
	{
		0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1,
		1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,
		1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1,
		1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0,
	};

	assert(sizeof(tbl_encselect) / sizeof(tbl_encselect[0]) == NUM_MONENC);

	/* X^[e[uf[^i[obt@mۂ */
	if ((buf_monenc = malloc(sizeof(monenc_t) * NUM_MONENC)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* X^[e[uf[^ǂݍ */
	for (i = 0; i < NUM_MONENC; i++)
	{
		unsigned char buf[SIZEOF_MONENC];

		file_read(buf, FILEPOS_MONENC + SIZEOF_MONENC * i, SIZEOF_MONENC);

		for (j = 0; j < SIZEOF_MONENC; j++)
		{
			buf_monenc[i].party_no[j] = buf[j];
			if (tbl_encselect[i] != 0x00)
				buf_monenc[i].party_no[j] += 0x100;
		}
	}

	return buf_monenc;
}

monmap_t *mon_readmap(void)
{
	int i, x, y;
	unsigned char *buf = NULL;
	unsigned char *ptr_buf = NULL;
	monmap_t *buf_monmap = NULL;
	const size_t bufsize = 256;

	/* X^[}bvf[^i[obt@mۂ */
	if ((buf_monmap = malloc(sizeof(monmap_t) * NUM_MONMAP)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);
	for (i = 0; i < NUM_MONMAP; i++)
	{
		for (y = 0; y < 8; y++)
		{
			for (x = 0; x < 8; x++)
				buf_monmap[i].area[x][y] = 0;
		}
	}

	/* X^[}bvf[^ǂݍ */
	if ((buf = malloc(bufsize)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);
	memset(buf, 0x00, bufsize);
	file_read(buf, FILEPOS_MONMAP, bufsize);

	ptr_buf = buf;
	for (i = 0; i < NUM_MONMAP; i++)
	{
		for (y = 0; y < tbl_sizeof_map[i]; y++)
		{
			for (x = 0; x < tbl_sizeof_map[i]; x++)
				buf_monmap[i].area[x][y] = *ptr_buf++;
		}
	}

	free(buf);
	buf = NULL;

	return buf_monmap;
}

int mon_writeparam(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̐lnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̐lnf[^o͂ */
	fprintf(fp, "No.      O                "
				" LV     HP     MP    EXP   M "
				"U U  "
				"h h  "
				"h h  "
				"U    f\n");

	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :  %3d, %5ld, %5ld, %5ld, %5ld, %5d, %5d, %5d, %5d, %5d, %5d, %5d, %5d, %5d, %5d, %5d, %5d\n", i, i,
				mon_getname(i),
				ptr_mondata[i].level,
				ptr_mondata[i].hit_point,
				ptr_mondata[i].magic_point,
				ptr_mondata[i].experience_point,
				ptr_mondata[i].gil,
				ptr_mondata[i].attack_times,
				ptr_mondata[i].attack_power,
				ptr_mondata[i].hit_rate,
				ptr_mondata[i].defense_times,
				ptr_mondata[i].defense_power,
				ptr_mondata[i].evade_rate,
				ptr_mondata[i].magic_defense_times,
				ptr_mondata[i].magic_defense_power,
				ptr_mondata[i].magic_evade_rate,
				ptr_mondata[i].magic_attack_times,
				ptr_mondata[i].magic_power,
				ptr_mondata[i].speed);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeelem(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̑nf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̑nf[^o͂ */
	fprintf(fp, "No.      O               "
				"ϐ              "
				"_\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s : ", i, i, mon_getname(i));

		/* Uւ̑ϐ */
		str_printelem(fp, ptr_mondata[i].elem_defense, ELEM_ALL);
		fprintf(fp, "  ");

		/* _̑ */
		str_printelem(fp, ptr_mondata[i].elem_weakness, ELEM_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writestate(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̏ԕωnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̏ԕωnf[^o͂ */
	fprintf(fp, "No.      O               "
				"{X  "
				"ԕωU1    2                 "
				"ԕωϐ1    2\n");
	for (i = 0; i < NUM_MON; i++)
	{
		/* {Xł邩ǂ */
		fprintf(fp, "%03d(%02X): %-16s :  %2s   ", i, i, mon_getname(i),
				str_getflag(ptr_mondata[i].flag_boss));

		/* ʏȔԕω */
		str_printstate(fp, ptr_mondata[i].state_attack, STATE_ALL & ~STATE_OTHER);
		fprintf(fp, "  ");

		/* ԕωւ̑ϐ */
		str_printstate(fp, ptr_mondata[i].state_defense, STATE_ALL & ~STATE_OTHER);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeflag(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̃tOnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̃tOnf[^o͂ */
	fprintf(fp, "No.      O               "
				"{X  푰\n");
	for (i = 0; i < NUM_MON; i++)
	{
		/* etO */
		fprintf(fp, "%03d(%02X): %-16s :  %2s", i, i, mon_getname(i),
				str_getflag(ptr_mondata[i].flag_boss));

		/* X^[^Cv */
		if ((ptr_mondata[i].montype & MONTYPE_ALL) != 0x00)
		{
			fprintf(fp, "   ");
			str_printmontype(fp, ptr_mondata[i].montype, MONTYPE_ALL);
		}

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeitem(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i, j;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̏ACef[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̏ACef[^o͂ */
	fprintf(fp, "No.      O               "
				"m   "
				"ACe1         "
				"ACe2         "
				"ACe3         "
				"ACe4\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :  %3d,  ", i, i, mon_getname(i), ptr_mondata[i].item_get);
		for (j = 0; j < 4; j++)
		{
			if (j != 0)
				fprintf(fp, ", ");
			fprintf(fp, "%-16s", item_getname(ptr_mondata[i].item_no[j]));
		}
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeact(const char *ptr_filename, const monact_t *ptr_monact, const mondata_t *ptr_mondata)
{
	int i, j, k;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monact == NULL || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̍sp^[f[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̍sp^[f[^o͂ */
	for (i = 0; i < NUM_MON; i++)
	{
		int act;

		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[No.%03d(%02X): %s]\n", i, i, mon_getname(i));
		for (act = 0; act < 2; act++)
		{
			const int actptn_no = ptr_mondata[i].actptn_no[act];

			if (actptn_no < 0)
				continue;

			for (j = 0; j < 16 && ptr_monact[actptn_no].ptn[j].action_size != 0 && ptr_monact[actptn_no].ptn[j].action != NULL; j++)
			{
				int turn = 0;
				int turn_old = 0;
				int letter_num = 0;
				unsigned char *ptr_action = NULL;
				unsigned char *buf_action = NULL;
				const size_t bufsize = ptr_monact[i].ptn[j].action_size + 256;

				/* ͗p̈ꎞobt@mۂ */
				if ((buf_action = malloc(bufsize)) == NULL)
					err_printf(STR_FATAL_ERROR_IN_MALLOC);
				memset(buf_action, 0x00, bufsize);
				memmove(buf_action, ptr_monact[actptn_no].ptn[j].action, ptr_monact[actptn_no].ptn[j].action_size);

				/* Kp */
				fprintf(fp, ": ");
				for (k = 0; k < 3; k++)
				{
					if (ptr_monact[actptn_no].ptn[j].condition_no[k] < 0)
						break;

					if (k > 0)
						fprintf(fp, " & ");
					fprintf(fp, "[%s]", mon_getactcnd(ptr_monact[actptn_no].ptn[j].condition_no[k]));
				}
				fputc('\n', fp);

				/* sp^[ */
				turn = 1;
				ptr_action = buf_action;
				while (ptr_action - buf_action < (ptrdiff_t)ptr_monact[actptn_no].ptn[j].action_size)
				{
					/* ^[̕\ */
					if (turn != turn_old)
						letter_num = fprintf(fp, "%d^[: ", turn);
					else
						fprintf(fp, "%*s", letter_num, "");
					turn_old = turn;

					/* sp^[f[^̉ */
					if (*ptr_action != 0xfd)
					{
						ptr_action = mon_printact(fp, ptr_action);
						fputc('\n', fp);
					}
					else
					{
						int flag_first = 1;

						fprintf(fp, "[s]\n%*s", letter_num, "");

						ptr_action++;
						while (*ptr_action != 0xfc)
						{
							if (*ptr_action == 0xfb)
							{
								fprintf(fp, "\n%*s", letter_num, "");
								flag_first = 1;
								ptr_action++;
							}
							else
							{
								if (flag_first)
									flag_first = 0;
								else
									fprintf(fp, " & ");

								ptr_action = mon_printact(fp, ptr_action);
							}
						}

						ptr_action++;
						fputc('\n', fp);
					}

					/* ̃oCg0xfe(^[̋؂)Ȃ΁Aǂݔ΂ă^[𑝉 */
					if (*ptr_action == 0xfe)
					{
						turn++;
						ptr_action++;
					}
				}

				free(buf_action);
				buf_action = NULL;
			}
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writemes(const char *ptr_filename)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0')
	{
		assert(0);
		return 0;
	}

	/* X^[̃bZ[Wf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̃bZ[Wf[^o͂ */
	fprintf(fp, "No.      bZ[W\n");
	for (i = 0; i < NUM_MONMES; i++)
	{
		fprintf(fp, "%03d(%02X): ", i, i);
		fputs(mon_getmes(i), fp);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeparty(const char *ptr_filename, const monparty_t *ptr_monparty)
{
	int i, j;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monparty == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̃p[eBҐf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̃p[eBҐf[^o͂ */
	fprintf(fp, "No.       "
				"  G2 G3 G1 G2  "
				"O[v1             "
				"O[v2             "
				"O[v3\n");
	for (i = 0; i < NUM_MONPARTY; i++)
	{
		const monparty_t tmp_party = ptr_monparty[i];

		fprintf(fp, "%03d(%03X):  %2s   %2s   %2s   %2s   %2s   %2s  ", i, i,
				str_getflag(tmp_party.flag_escape),
				str_getflag(tmp_party.flag_event),
				str_getflag(tmp_party.flag_exist[1]),
				str_getflag(tmp_party.flag_exist[2]),
				str_getflag(tmp_party.flag_eggy[0]),
				str_getflag(tmp_party.flag_eggy[1]));
		for (j = 0; j < 3; j++)
		{
			if (tmp_party.monster_no[j] != 0xff)
			{
				int flag_absent = (tmp_party.num[j] == 0 || tmp_party.flag_eggy[j] || tmp_party.flag_exist[j]);

				if (j != 0)
					fprintf(fp, ", ");
				fprintf(fp, "%s%-16s%s %1d",
						(flag_absent) ? "(" : " ",
						mon_getname(tmp_party.monster_no[j]),
						(flag_absent) ? ")" : " ",
						tmp_party.num[j]);
			}
		}
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeenc(const char *ptr_filename, const monenc_t *ptr_monenc, const monparty_t *ptr_monparty)
{
	int i, j, k;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monenc == NULL || ptr_monparty == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[e[uf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[e[uo͂ */
	for (i = 0; i < NUM_MONENC; i++)
	{
		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[No.%03d(%02X)]\n", i, i);
		for (j = 0; j < SIZEOF_MONENC; j++)
		{
			const monparty_t tmp_party = ptr_monparty[ptr_monenc[i].party_no[j]];

			fprintf(fp, "%02d: %03d(%03X): ", j + 1, ptr_monenc[i].party_no[j], ptr_monenc[i].party_no[j]);
			for (k = 0; k < 3; k++)
			{
				if (tmp_party.monster_no[k] != 0xff)
				{
					const int flag_absent = (tmp_party.num[k] == 0 || tmp_party.flag_eggy[k] || tmp_party.flag_exist[k]);

					if (k != 0)
						fprintf(fp, ", ");
					fprintf(fp, "%s%-16s%s %1d",
							(flag_absent) ? "(" : " ",
							mon_getname(tmp_party.monster_no[k]),
							(flag_absent) ? ")" : " ",
							tmp_party.num[k]);
				}
			}
			fputc('\n', fp);
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writemap(const char *ptr_filename, const monmap_t *ptr_monmap)
{
	int i, x, y;
	FILE *fp = NULL;
	const char *str_name[NUM_MONMAP] =
	{
		"n㐢E",
		"nꐢE",
		"E",
	};

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monmap == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[}bvf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[}bvo͂ */
	for (i = 0; i < NUM_MONMAP; i++)
	{
		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[No.%1d: %s]\n", i, str_name[i]);
		for (y = 0; y < tbl_sizeof_map[i]; y++)
		{
			for (x = 0; x < tbl_sizeof_map[i]; x++)
			{
				if (x != 0)
					fprintf(fp, " ");

				fprintf(fp, "%03d", ptr_monmap[i].area[x][y]);
			}
			fputc('\n', fp);
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

static unsigned char *mon_printact(FILE *stream, unsigned char *buf_action)
{
	int i;
	int act1, act2;
	unsigned char *ptr_action = buf_action;

	/* `FbN */
	if (stream == NULL || buf_action == NULL)
	{
		assert(0);
		return buf_action;
	}

	act1 = buf_action[0];
	act2 = buf_action[1];
	if (act1 < 0xc0)
	{
		/* @n */
		if (0x4d <= act1 && act1 <= 0x5e)
		{
			/* S̖@ */
			fprintf(stream, "%s(S)", mag_getname(act1 - 0x30));
		}
		else
		{
			/* ꖂ@ */
			switch (act1)
			{
			case 0xa9:
				fprintf(stream, "(GԐݒ)");
				break;
			case 0xaa:
				fprintf(stream, "(Gԉ)");
				break;
			case 0xad:
				fprintf(stream, "(X^[oԕύX)");
				break;
			case 0xae:
				fprintf(stream, "(퓬I)");
				break;
			case 0xbb:
				fprintf(stream, "(U)");
				break;
			case 0xbe:
				fprintf(stream, "(HP/MPS)");
				break;
			default:
				if (act1 < 0xaf)
					fprintf(stream, "%s", mag_getname(act1));
				else
					fprintf(stream, "(o)");
				break;
			}
		}

		ptr_action++;
	}
	else if (0xc0 <= act1 && act1 < 0xe8)
	{
		/* R}hn */
		switch (act1)
		{
		case 0xc0:
			fprintf(stream, "");
			break;
		case 0xc5:
			fprintf(stream, "񂱂");
			break;
		case 0xc6:
			fprintf(stream, "Wv");
			break;
		case 0xce:
			fprintf(stream, "");
			break;
		case 0xe1:
			fprintf(stream, "(Ȃ)");
			break;
		default:
			fprintf(stream, "(R}h: %02Xh)", act1);
			break;
		}

		ptr_action++;
	}
	else
	{
		/* R[hn */
		if (act1 == 0xe8)
		{
			char str_montype[256] = "";
			char *ptr = str_montype;
			const montype_t montype = act2;

			for (i = 0; i < NUM_MONTYPE; i++)
				ptr += sprintf(ptr, "%s", str_getmontype(montype, i));
			if (*str_montype == '\0')
				strcpy(str_montype, "(Ȃ)");

			fprintf(stream, "[X^[^Cvݒ: %s]", str_montype);
			ptr_action += 2;
		}
		else if (0xe9 <= act1 && act1 <= 0xeb)
		{
			const triplet_t triplet = mon_gettriplet(act2);

			switch (act1)
			{
			case 0xe9:
				fprintf(stream, "[Up[^ݒ: U%d, U%d, %d%%]",
						triplet.times, triplet.power, triplet.rate);
				break;
			case 0xea:
				fprintf(stream, "[hp[^ݒ: h%d, h%d, %d%%]",
						triplet.times, triplet.power, triplet.rate);
				break;
			case 0xeb:
				fprintf(stream, "[@hp[^ݒ: @h%d, @h%d, @%d%%]",
						triplet.times, triplet.power, triplet.rate);
				break;
			}

			ptr_action += 2;
		}
		else if (act1 == 0xec)
		{
			fprintf(stream, "[Xs[hݒ: %d]", act2);
			ptr_action += 2;
		}
		else if (act1 == 0xed || act1 == 0xef)
		{
			char str_element[256] = "";
			char *ptr = str_element;
			const elem_t element = act2;

			for (i = 0; i < NUM_ELEMENT; i++)
				ptr += sprintf(ptr, "%s", str_getelem(element, i));
			if (*str_element == '\0')
				strcpy(str_element, "(Ȃ)");

			fprintf(stream, "[%sݒ: %s]", (act1 == 0xed) ? "ϐ" : "_", str_element);
			ptr_action += 2;
		}
		else if (act1 == 0xee)
		{
			fprintf(stream, "[(@U)ݒ: %d(%d)]", act2, 1 + act2 / 4);
			ptr_action += 2;
		}
		else if (act1 == 0xf0)
		{
			fprintf(stream, "[OtBbNݒ: %02Xh]", act2);
			ptr_action += 2;
		}
		else if (act1 == 0xf1 || act1 == 0xf2)
		{
			fprintf(stream, "[bZ[W\: %s]", mon_getmes(act2));
			ptr_action += 2;
		}
		else if (act1 == 0xf3)
		{
			fprintf(stream, "[o: %02Xh]", act2);
			ptr_action += 2;
		}
		else if (0xf4 <= act1 && act1 <= 0xf8)
		{
			fprintf(stream, "[ϐ%dݒ: %s%d]", act1 - 0xf4, ((act2 & 0x80) == 0x00) ? "+" : "", act2 & 0x7f);
			ptr_action += 2;
		}
		else if (act1 == 0xf9)
		{
			fprintf(stream, "[UڕWݒ: %s]", mon_gettarget(act2));
			ptr_action += 2;
		}
		else if (0xe8 <= act1 && act1 <= 0xf9)
		{
			fprintf(stream, "%02Xh (%02Xh)", act1, act2);
			ptr_action += 2;
		}
		else
		{
			fprintf(stream, "%02Xh", act1);
			ptr_action++;
		}
	}

	return ptr_action;
}

void mon_freeact(monact_t *ptr_monact)
{
	int i, j;

	/* `FbN */
	if (ptr_monact == NULL)
	{
		assert(0);
		return;
	}

	for (i = 0; i < NUM_MONACTLIST; i++)
	{
		for (j = 0; j < sizeof(ptr_monact[i].ptn) / sizeof(ptr_monact[i].ptn[0]); j++)
		{
			if (ptr_monact[i].ptn[j].action != NULL)
			{
				free(ptr_monact[i].ptn[j].action);
				ptr_monact[i].ptn[j].action = NULL;
			}
		}
	}
}
