/*
 *	ff5analyzer.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "errmes.h"
#include "utility.h"
#include "ff5file.h"
#include "ff5monster.h"
#include "ff5character.h"
#include "ff5item.h"
#include "ff5magic.h"
#include "ff5special.h"
#include "ff5misc.h"

#define FILENAME_MON_PARAM		"ff5monparam.txt"
#define FILENAME_MON_ELEM		"ff5monelem.txt"
#define FILENAME_MON_STATE		"ff5monstate.txt"
#define FILENAME_MON_FLAG		"ff5monflag.txt"
#define FILENAME_MON_ITEM		"ff5monitem.txt"
#define FILENAME_MON_FREE		"ff5monfree.txt"
#define FILENAME_MON_CTRL		"ff5monctrl.txt"
#define FILENAME_MON_TECH		"ff5montech.txt"
#define FILENAME_MON_MISC		"ff5monmisc.txt"
#define FILENAME_MON_ACT		"ff5monact.txt"
#define FILENAME_MON_MES		"ff5monmes.txt"
#define FILENAME_MON_PARTY		"ff5monparty.txt"
#define FILENAME_MON_ENC		"ff5monenc.txt"
#define FILENAME_MON_MAP		"ff5monmap.txt"
#define FILENAME_CHR_LVUP		"ff5chrlvup.txt"
#define FILENAME_CHR_JOB		"ff5chrjob.txt"
#define FILENAME_CHR_CMD_PARAM	"ff5chrcmdparam.txt"
#define FILENAME_CHR_CMD_ELEM	"ff5chrcmdelem.txt"
#define FILENAME_CHR_CMD_STATE	"ff5chrcmdstate.txt"
#define FILENAME_CHR_CMD_FLAG	"ff5chrcmdflag.txt"
#define FILENAME_CHR_ABT		"ff5chrabt.txt"
#define FILENAME_ITEM_PARAM		"ff5itemparam.txt"
#define FILENAME_ITEM_ELEM		"ff5itemelem.txt"
#define FILENAME_ITEM_STATE		"ff5itemstate.txt"
#define FILENAME_ITEM_FLAG		"ff5itemflag.txt"
#define FILENAME_ITEM_EFFECT	"ff5itemeffect.txt"
#define FILENAME_MAG_PARAM		"ff5magparam.txt"
#define FILENAME_MAG_ELEM		"ff5magelem.txt"
#define FILENAME_MAG_STATE		"ff5magstate.txt"
#define FILENAME_MAG_FLAG		"ff5magflag.txt"
#define FILENAME_SP_PARAM		"ff5spparam.txt"
#define FILENAME_SP_ELEM		"ff5spelem.txt"
#define FILENAME_SP_STATE		"ff5spstate.txt"
#define FILENAME_SP_FLAG		"ff5spflag.txt"
#define FILENAME_TREASURE		"ff5treasure.txt"
#define FILENAME_ITEM_PARAM_BUG	"ff5itemparambug.txt"
#define FILENAME_ITEM_ELEM_BUG	"ff5itemelembug.txt"
#define FILENAME_ITEM_STATE_BUG	"ff5itemstatebug.txt"

#define STR_TOO_FEW_ARGUMENTS		"Too few arguments"
#define STR_INVALID_OPTION			"Invalid option"
#define STR_INVALID_ROMIMAGE_FILE	"Invalid ROM-Image file"
#define STR_INVALID_CDROM_FILES		"Invalid CD-ROM files"
#define STR_CANNOT_CREATE_FILE		"Cannot create file"

/* 쌠\L */
static const char *str_copyright = "ff5analyzer - Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>";

int main(int argc, char *argv[])
{
	int i;
	int ret = EXIT_SUCCESS;
	int mode = FILEMODE_SFC;
	int flag_invalid = 0;
	int flag_usage   = 0;
	int index_srcpath = -1;
	int index_dstpath = -1;
	char str_srcpath[1024] = "";
	char str_dstpath[1024] = "";
	char str_dstfilename[1024] = "";
	mondata_t  *buf_mondata  = NULL;
	monact_t   *buf_monact   = NULL;
	monparty_t *buf_monparty = NULL;
	monenc_t   *buf_monenc   = NULL;
	monmap_t   *buf_monmap   = NULL;
	chrlvup_t  *buf_chrlvup  = NULL;
	chrjob_t   *buf_chrjob   = NULL;
	chrcmd_t   *buf_chrcmd   = NULL;
	chrabt_t   *buf_chrabt   = NULL;
	itemdata_t *buf_itemdata = NULL;
	magdata_t  *buf_magdata  = NULL;
	spdata_t   *buf_spdata   = NULL;
	treasure_t *buf_trdata   = NULL;

	errno = 0;
	err_setprogname(argv[0]);

	/* R}hC͂ */
	for (i = 1; i < argc; i++)
	{
		if (argv[i] == NULL)
			continue;

		if (*argv[i] == '-')
		{
			const char *ptr = argv[i];

			while (*++ptr != '\0')
			{
				switch (tolower(*ptr))
				{
				case 'f':	/* X[p[t@~RŃ[h */
					mode = FILEMODE_SFC;
					break;
				case 'p':	/* vCXe[VŃ[h */
					mode = FILEMODE_PS;
					break;
				case 'h':	/* gp@̕\ */
					flag_usage = 1;
					break;
				default:
					flag_invalid = 1;
					break;
				}
			}
		}
		else
		{
			if (index_srcpath < 0)
				index_srcpath = i;
			else if (index_dstpath < 0)
				index_dstpath = i;
		}
	}

	/* /o̓pXݒ肷 */
	if (index_srcpath >= 0)
	{
		if (mode == FILEMODE_PS)
			transpath(str_srcpath, argv[index_srcpath], sizeof(str_srcpath));
		else
			transfilename(str_srcpath, argv[index_srcpath], sizeof(str_srcpath));
	}
	if (index_dstpath >= 0)
		transpath(str_dstpath, argv[index_dstpath], sizeof(str_dstpath));

	/* gp@\ */
	if (flag_usage)
	{
		puts(str_copyright);
		printf("%s: [option] [src-pathname] [dst-pathname]\n", err_getprogname());
		printf("\t-f\tanalyze in \"SUPER FAMICOM Ver.\" mode\n");
		printf("\t-p\tanalyze in \"PlayStation Ver.\" mode\n");
		printf("\t-h\tdisplay this help and exit\n");
		printf("\n");

		exit(EXIT_SUCCESS);
	}

	/* R}hC`FbN */
	if (flag_invalid)
		err_printf(STR_INVALID_OPTION);
	if (mode == FILEMODE_SFC && *str_srcpath == '\0')
		err_printf(STR_TOO_FEW_ARGUMENTS);

	/* t@CI[v */
	if (!file_open(str_srcpath, mode))
	{
		if (mode == FILEMODE_PS)
			err_printf(STR_INVALID_CDROM_FILES " - %s*", str_srcpath);
		else
			err_printf(STR_INVALID_ROMIMAGE_FILE " - %s", str_srcpath);
	}

	/* ef[^́E͂ */
	buf_mondata  = mon_readdata();
	buf_monact   = mon_readact();
	buf_monparty = mon_readparty();
	buf_monenc   = mon_readenc();
	buf_monmap   = mon_readmap();
	buf_chrlvup  = chr_readlvup();
	buf_chrjob   = chr_readjob();
	buf_chrcmd   = chr_readcmd();
	buf_chrabt   = chr_readabt();
	buf_itemdata = item_readdata();
	buf_magdata  = mag_readdata();
	buf_spdata   = sp_readdata();
	buf_trdata   = tr_readdata();

	/* ef[^o͂ */
	if (buf_mondata  == NULL ||
		buf_monact   == NULL ||
		buf_monparty == NULL ||
		buf_monenc   == NULL ||
		buf_monmap   == NULL ||
		buf_chrlvup  == NULL ||
		buf_chrjob   == NULL ||
		buf_chrcmd   == NULL ||
		buf_chrabt   == NULL ||
		buf_itemdata == NULL ||
		buf_magdata  == NULL ||
		buf_spdata   == NULL ||
		buf_trdata   == NULL)
	{
		if (mode == FILEMODE_PS)
			wrn_printf(STR_INVALID_CDROM_FILES " - %s*", str_srcpath);
		else
			wrn_printf(STR_INVALID_ROMIMAGE_FILE " - %s", str_srcpath);

		ret = EXIT_FAILURE;
	}
	else
	{
		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_PARAM);
		if (!mon_writeparam(str_dstfilename, buf_mondata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_ELEM);
		if (!mon_writeelem(str_dstfilename, buf_mondata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_STATE);
		if (!mon_writestate(str_dstfilename, buf_mondata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_FLAG);
		if (!mon_writeflag(str_dstfilename, buf_mondata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_ITEM);
		if (!mon_writeitem(str_dstfilename, buf_mondata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_FREE);
		if (!mon_writefree(str_dstfilename, buf_mondata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_CTRL);
		if (!mon_writectrl(str_dstfilename, buf_mondata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_TECH);
		if (!mon_writetech(str_dstfilename, buf_mondata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_MISC);
		if (!mon_writemisc(str_dstfilename, buf_mondata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_ACT);
		if (!mon_writeact(str_dstfilename, buf_monact, buf_mondata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_MES);
		if (!mon_writemes(str_dstfilename))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_PARTY);
		if (!mon_writeparty(str_dstfilename, buf_monparty))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_ENC);
		if (!mon_writeenc(str_dstfilename, buf_monenc, buf_monparty))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_MAP);
		if (!mon_writemap(str_dstfilename, buf_monmap))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_CHR_LVUP);
		if (!chr_writelvup(str_dstfilename, buf_chrlvup))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_CHR_JOB);
		if (!chr_writejob(str_dstfilename, buf_chrjob))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_CHR_CMD_PARAM);
		if (!chr_writecmdparam(str_dstfilename, buf_chrcmd))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}
/*
		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_CHR_CMD_ELEM);
		if (!chr_writecmdelem(str_dstfilename, buf_chrcmd))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}
*/
		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_CHR_CMD_STATE);
		if (!chr_writecmdstate(str_dstfilename, buf_chrcmd))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_CHR_CMD_FLAG);
		if (!chr_writecmdflag(str_dstfilename, buf_chrcmd))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_CHR_ABT);
		if (!chr_writeabtparam(str_dstfilename, buf_chrabt))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_ITEM_PARAM);
		if (!item_writeparam(str_dstfilename, buf_itemdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_ITEM_ELEM);
		if (!item_writeelem(str_dstfilename, buf_itemdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_ITEM_STATE);
		if (!item_writestate(str_dstfilename, buf_itemdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_ITEM_FLAG);
		if (!item_writeflag(str_dstfilename, buf_itemdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_ITEM_EFFECT);
		if (!item_writeeffect(str_dstfilename, buf_itemdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MAG_PARAM);
		if (!mag_writeparam(str_dstfilename, buf_magdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MAG_ELEM);
		if (!mag_writeelem(str_dstfilename, buf_magdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MAG_STATE);
		if (!mag_writestate(str_dstfilename, buf_magdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MAG_FLAG);
		if (!mag_writeflag(str_dstfilename, buf_magdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_SP_PARAM);
		if (!sp_writeparam(str_dstfilename, buf_spdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_SP_ELEM);
		if (!sp_writeelem(str_dstfilename, buf_spdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_SP_STATE);
		if (!sp_writestate(str_dstfilename, buf_spdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}
/*
		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_SP_FLAG);
		if (!sp_writeflag(str_dstfilename, buf_spdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}
*/
		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_TREASURE);
		if (!tr_writedata(str_dstfilename, buf_trdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_ITEM_PARAM_BUG);
		if (!item_writeparambug(str_dstfilename, buf_itemdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_ITEM_ELEM_BUG);
		if (!item_writeelembug(str_dstfilename, buf_itemdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_ITEM_STATE_BUG);
		if (!item_writestatbug(str_dstfilename, buf_itemdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}
	}

	/* t@CN[Y */
	file_close();

	/* obt@ */
	mon_freeact(buf_monact);
	free(buf_mondata);
	free(buf_monact);
	free(buf_monparty);
	free(buf_monenc);
	free(buf_monmap);
	free(buf_chrlvup);
	free(buf_itemdata);
	free(buf_magdata);
	free(buf_spdata);
	free(buf_trdata);
	buf_mondata  = NULL;
	buf_monact   = NULL;
	buf_monparty = NULL;
	buf_monenc   = NULL;
	buf_monmap   = NULL;
	buf_chrlvup  = NULL;
	buf_chrjob   = NULL;
	buf_chrcmd   = NULL;
	buf_chrabt   = NULL;
	buf_itemdata = NULL;
	buf_magdata  = NULL;
	buf_spdata   = NULL;
	buf_trdata   = NULL;

	return ret;
}
