/*
 *	ff5character.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "ff5file.h"
#include "ff5define.h"
#include "ff5decode.h"
#include "ff5character.h"

#define FILEPOS_CHREXP			(0x019600 + 0x0fba00)	/* = 0x115000 */
#define FILEPOS_CHRHP			(0x019729 + 0x0fba00)	/* = 0x115129 */
#define FILEPOS_CHRMP			(0x0197ef + 0x0fba00)	/* = 0x1151ef */
#define FILEPOS_CHRJOBNAME		(0x019c00 + 0x0fba00)	/* = 0x115600 */
#define FILEPOS_CHRJOBDATA		(0x019cb0 + 0x0fba00)	/* = 0x1156b0 */
#define FILEPOS_CHRJOBLVUP_ADDR	(0x0198c0 + 0x0fba00)	/* = 0x1152c0 */
#define FILEPOS_CHRJOBLVUP		(0x019900 + 0x0fba00)	/* = 0x115300 */
#define FILEPOS_CHRJOBLVMAX		(0x0198ea + 0x0fba00)	/* = 0x1152ea */
#define FILEPOS_CHRJOBEQUIP		(0x019d08 + 0x0fba00)	/* = 0x115708 */
#define FILEPOS_CHRJOBCMD		(0x019d60 + 0x0fba00)	/* = 0x115760 */
#define FILEPOS_CHRCMDNAME		(0x019e00 + 0x0fba00)	/* = 0x115800 */
#define FILEPOS_CHRCMDDATA		(0x019fe0 + 0x0fba00)	/* = 0x1159e0 */
#define FILEPOS_CHRCMDPARAM		(0x01a400 + 0x0fba00)	/* = 0x115e00 */
#define FILEPOS_CHRCMDTIME		(0x01a3a0 + 0x0fba00)	/* = 0x115da0 */
#define FILEPOS_CHRABTNAME		(0x01a800 + 0x0fba00)	/* = 0x116200 */
#define FILEPOS_CHRABTPARAM		(0x01a908 + 0x0fba00)	/* = 0x116308 */
#define NUM_CHRLV				99
#define NUM_CHRJOB				22
#define NUM_CHRCMD				96
#define NUM_CHRABT				33
#define SIZEOF_CHREXP			3
#define SIZEOF_CHRHP			2
#define SIZEOF_CHRMP			2
#define SIZEOF_CHRJOBNAME		8
#define SIZEOF_CHRJOBDATA		4
#define SIZEOF_CHRJOBLVUP		3
#define SIZEOF_CHRJOBLVMAX		1
#define SIZEOF_CHRJOBEQUIP		4
#define SIZEOF_CHRJOBCMD		4
#define SIZEOF_CHRCMDNAME		5
#define SIZEOF_CHRCMDDATA		8
#define SIZEOF_CHRCMDTIME		1
#define SIZEOF_CHRABTNAME		8
#define SIZEOF_CHRPARAM			4
#define SIZEOF_OFFSETADDR		2

static void chr_setjobname(void);
static void chr_setcmdname(void);
static void chr_setabtname(void);

static char tbl_jobname[NUM_CHRJOB][256];	/* Wu̖Oi[e[u */
static char tbl_cmdname[NUM_CHRCMD][256];	/* R}hAreB̖Oi[e[u */
static char tbl_abtname[NUM_CHRABT][256];	/* AreB̖Oi[e[u */
static int flag_setjobname = 0;				/* tbl_jobnameɃf[^i[Ă邩ǂ */
static int flag_setcmdname = 0;				/* tbl_cmdnameɃf[^i[Ă邩ǂ */
static int flag_setabtname = 0;				/* tbl_abtnameɃf[^i[Ă邩ǂ */

const char *chr_getjobname(int index)
{
	/* `FbN */
	if (index < 0 || NUM_CHRJOB <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setjobname)
		chr_setjobname();

	return tbl_jobname[index];
}

const char *chr_getcmdname(int index)
{
	/* `FbN */
	if (index < 0 || NUM_CHRCMD <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setcmdname)
		chr_setcmdname();

	return tbl_cmdname[index];
}

const char *chr_getabtname(int index)
{
	/* `FbN */
	if (index < 0 || NUM_CHRABT <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setabtname)
		chr_setabtname();

	return tbl_abtname[index];
}

static void chr_setjobname(void)
{
	int i;

	/* Wu̖Of[^ */
	for (i = 0; i < NUM_CHRJOB; i++)
		*tbl_jobname[i] = '\0';

	/* Wu̖Of[^ǂݍ */
	for (i = 0; i < NUM_CHRJOB; i++)
	{
		unsigned char buf[SIZEOF_CHRJOBNAME];

		file_read(buf, FILEPOS_CHRJOBNAME + SIZEOF_CHRJOBNAME * i, SIZEOF_CHRJOBNAME);
		dec_string(tbl_jobname[i], buf, sizeof(tbl_jobname[i]), SIZEOF_CHRJOBNAME, 0);
	}

	flag_setjobname = 1;
}

static void chr_setcmdname(void)
{
	int i;

	/* R}hAreB̖Of[^ */
	for (i = 0; i < NUM_CHRCMD; i++)
		*tbl_cmdname[i] = '\0';

	/* R}hAreB̖Of[^ǂݍ */
	for (i = 0; i < NUM_CHRCMD; i++)
	{
		unsigned char buf[SIZEOF_CHRCMDNAME];

		file_read(buf, FILEPOS_CHRCMDNAME + SIZEOF_CHRCMDNAME * i, SIZEOF_CHRCMDNAME);
		dec_string(tbl_cmdname[i], buf, sizeof(tbl_cmdname[i]), SIZEOF_CHRCMDNAME, 0);
	}

	flag_setcmdname = 1;
}

static void chr_setabtname(void)
{
	int i;

	/* AreB̖Of[^ */
	for (i = 0; i < NUM_CHRABT; i++)
		*tbl_abtname[i] = '\0';

	/* AreB̖Of[^ǂݍ */
	for (i = 0; i < NUM_CHRABT; i++)
	{
		unsigned char buf[SIZEOF_CHRABTNAME];

		file_read(buf, FILEPOS_CHRABTNAME + SIZEOF_CHRABTNAME * i, SIZEOF_CHRABTNAME);
		dec_string(tbl_abtname[i], buf, sizeof(tbl_abtname[i]), SIZEOF_CHRABTNAME, 0);
	}

	flag_setabtname = 1;
}

chrlvup_t *chr_readlvup(void)
{
	int i;
	chrlvup_t *buf_chrlvup = NULL;

	/* LN^̐f[^i[obt@mۂ */
	if ((buf_chrlvup = malloc(sizeof(chrlvup_t) * NUM_CHRLV)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* xƂ̗݌volǂݍ */
	for (i = 0; i < NUM_CHRLV; i++)
	{
		unsigned char buf[SIZEOF_CHREXP];

		file_read(buf, FILEPOS_CHREXP + SIZEOF_CHREXP * i, SIZEOF_CHREXP);
		buf_chrlvup[i].experience_point = (buf[2] << 16) | (buf[1] << 8) | buf[0];
	}

	/* xƂ̊{qbg|Cgǂݍ */
	for (i = 0; i < NUM_CHRLV; i++)
	{
		unsigned char buf[SIZEOF_CHRHP];

		file_read(buf, FILEPOS_CHRHP + SIZEOF_CHRHP * i, SIZEOF_CHRHP);
		buf_chrlvup[i].hit_point = (buf[1] << 8) | buf[0];
	}

	/* xƂ̊{}WbN|Cgǂݍ */
	for (i = 0; i < NUM_CHRLV; i++)
	{
		unsigned char buf[SIZEOF_CHRMP];

		file_read(buf, FILEPOS_CHRMP + SIZEOF_CHRMP * i, SIZEOF_CHRMP);
		buf_chrlvup[i].magic_point = (buf[1] << 8) | buf[0];
	}

	return buf_chrlvup;
}

chrjob_t *chr_readjob(void)
{
	int i, j;
	int tbl_offset[NUM_CHRJOB + 1];
	chrjob_t *buf_chrjob = NULL;

	/* Wůef[^i[obt@mۂ */
	if ((buf_chrjob = malloc(sizeof(chrjob_t) * NUM_CHRJOB)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* Wů{p[^ǂݍ */
	for (i = 0; i < NUM_CHRJOB; i++)
	{
		unsigned char buf[SIZEOF_CHRJOBDATA];

		file_read(buf, FILEPOS_CHRJOBDATA + SIZEOF_CHRJOBDATA * i, SIZEOF_CHRJOBDATA);
		buf_chrjob[i].strength    = buf[0];
		buf_chrjob[i].speed       = buf[1];
		buf_chrjob[i].stamina     = buf[2];
		buf_chrjob[i].magic_power = buf[3];
	}

	/* WułACẽf[^ǂݍ */
	for (i = 0; i < NUM_CHRJOB; i++)
	{
		unsigned char buf[SIZEOF_CHRJOBEQUIP];

		file_read(buf, FILEPOS_CHRJOBEQUIP + SIZEOF_CHRJOBCMD * i, SIZEOF_CHRJOBEQUIP);
		buf_chrjob[i].bits_equip_item = bits_make32(buf[0], buf[1], buf[2], buf[3]);
	}

	/* Wů{R}hǂݍ */
	for (i = 0; i < NUM_CHRJOB; i++)
	{
		unsigned char buf[SIZEOF_CHRJOBCMD];

		file_read(buf, FILEPOS_CHRJOBCMD + SIZEOF_CHRJOBCMD * i, SIZEOF_CHRJOBCMD);
		for (j = 0; j < 4; j++)
			buf_chrjob[i].cmd_no[j] = buf[j];
	}

	/* WuƂ̍ōWuxǂݍ */
	for (i = 0; i < NUM_CHRJOB; i++)
	{
		unsigned char buf[SIZEOF_CHRJOBLVMAX];

		file_read(buf, FILEPOS_CHRJOBLVMAX + SIZEOF_CHRJOBLVMAX * i, SIZEOF_CHRJOBLVMAX);
		buf_chrjob[i].level_max = buf[0];
	}

	/* Wux̃f[^i[ĂAhX(ւ̃ItZbg)ǂݍ */
	for (i = 0; i < NUM_CHRJOB; i++)
	{
		unsigned char buf[SIZEOF_OFFSETADDR];

		file_read(buf, FILEPOS_CHRJOBLVUP_ADDR + SIZEOF_OFFSETADDR * i, SIZEOF_OFFSETADDR);
		tbl_offset[i] = (buf[1] << 8) | buf[0];
	}
	tbl_offset[NUM_CHRJOB] = tbl_offset[NUM_CHRJOB - 1] + 0x03;	/*  */

	/* Wux̃f[^ǂݍ */
	for (i = 0; i < NUM_CHRJOB; i++)
	{
		for (j = 0; j < buf_chrjob[i].level_max; j++)
		{
			unsigned char buf[SIZEOF_CHRJOBLVUP];

			if (tbl_offset[i] - tbl_offset[0] < 0)
				break;

			file_read(buf, FILEPOS_CHRJOBLVUP + (tbl_offset[i] - tbl_offset[0]) + SIZEOF_CHRJOBLVUP * j, SIZEOF_CHRJOBLVUP);
			buf_chrjob[i].level[j].ability_point = (buf[1] << 8) | buf[0];
			buf_chrjob[i].level[j].ability_no    = buf[2];
		}
		for (; j < NUM_CHRJOBLVUPMAX; j++)
		{
			buf_chrjob[i].level[j].ability_point = 0;
			buf_chrjob[i].level[j].ability_no    = 0;
		}
	}

	return buf_chrjob;
}

chrcmd_t *chr_readcmd(void)
{
	int i;
	chrcmd_t *buf_chrcmd = NULL;

	/* R}hAreB̊ef[^i[obt@mۂ */
	if ((buf_chrcmd = malloc(sizeof(chrcmd_t) * NUM_CHRCMD)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* R}hAreB̊ef[^ǂݍ */
	for (i = 0; i < NUM_CHRCMD; i++)
	{
		unsigned char buf[SIZEOF_CHRCMDDATA];

		file_read(buf, FILEPOS_CHRCMDDATA + SIZEOF_CHRCMDDATA * i, SIZEOF_CHRCMDDATA);
		mag_setparam(&buf_chrcmd[i].param, buf, sizeof(buf));
	}

	/* R}hAreB̌ʔ܂ł̎Ԃǂݍ */
	for (i = 0; i < NUM_CHRCMD; i++)
	{
		unsigned char buf[SIZEOF_CHRCMDTIME];

		file_read(buf, FILEPOS_CHRCMDTIME + SIZEOF_CHRCMDTIME * i, SIZEOF_CHRCMDTIME);
		buf_chrcmd[i].time_wait = buf[0] & 0x7f;
	}

	/* R}hAreB̃p[^㏸lǂݍ */
	for (i = 0; i < NUM_CHRCMD; i++)
	{
		unsigned char buf[SIZEOF_CHRPARAM];

		file_read(buf, FILEPOS_CHRCMDPARAM + SIZEOF_CHRPARAM * i, SIZEOF_CHRPARAM);
		buf_chrcmd[i].increase_strength    = (buf[0] == 0x00) ? 0 : buf[0] - 24;
		buf_chrcmd[i].increase_speed       = (buf[1] == 0x00) ? 0 : buf[1] - 24;
		buf_chrcmd[i].increase_stamina     = (buf[2] == 0x00) ? 0 : buf[2] - 24;
		buf_chrcmd[i].increase_magic_power = (buf[3] == 0x00) ? 0 : buf[3] - 24;
	}

	return buf_chrcmd;
}

chrabt_t *chr_readabt(void)
{
	int i;
	chrabt_t *buf_chrabt = NULL;

	/* AreB̊ef[^i[obt@mۂ */
	if ((buf_chrabt = malloc(sizeof(chrabt_t) * NUM_CHRABT)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* AreB̃p[^㏸lǂݍ */
	for (i = 0; i < NUM_CHRABT; i++)
	{
		unsigned char buf[SIZEOF_CHRPARAM];

		file_read(buf, FILEPOS_CHRABTPARAM + SIZEOF_CHRPARAM * i, SIZEOF_CHRPARAM);
		buf_chrabt[i].increase_strength    = (buf[0] == 0x00) ? 0 : buf[0] - 24;
		buf_chrabt[i].increase_speed       = (buf[1] == 0x00) ? 0 : buf[1] - 24;
		buf_chrabt[i].increase_stamina     = (buf[2] == 0x00) ? 0 : buf[2] - 24;
		buf_chrabt[i].increase_magic_power = (buf[3] == 0x00) ? 0 : buf[3] - 24;
	}

	return buf_chrabt;
}

int chr_writelvup(const char *ptr_filename, const chrlvup_t *ptr_chrlvup)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrlvup == NULL)
	{
		assert(0);
		return 0;
	}

	/* LN^̐f[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* LN^̐f[^o͂ */
	fprintf(fp, "LV       EXP {HP {MP\n");
	for (i = 0; i < NUM_CHRLV; i++)
	{
		fprintf(fp, "%2d: %8ld, %5d, %5d\n", i + 1,
				ptr_chrlvup[i].experience_point,
				ptr_chrlvup[i].hit_point,
				ptr_chrlvup[i].magic_point);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int chr_writejob(const char *ptr_filename, const chrjob_t *ptr_chrjob)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_nothing = "----------";

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrjob == NULL)
	{
		assert(0);
		return 0;
	}

	/* Wůef[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* Wůef[^o͂ */
	for (i = 0; i < NUM_CHRJOB; i++)
	{
		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[No.%03d(%02X): %s]\n", i, i, chr_getjobname(i));

		/* Xe[^X */
		fprintf(fp, "      ̗ \n");
		fprintf(fp, " %3d, %3d, %3d, %3d\n",
				ptr_chrjob[i].strength,
				ptr_chrjob[i].speed,
				ptr_chrjob[i].stamina,
				ptr_chrjob[i].magic_power);

		/* R}h */
		fprintf(fp, "No. R}h\n");
		for (j = 0; j < 4; j++)
		{
			fprintf(fp, "%2d: %s\n", j,
					(ptr_chrjob[i].cmd_no[j] != 0x00) ? chr_getcmdname(ptr_chrjob[i].cmd_no[j]) : str_nothing);
		}

		/* Wux */
		if (ptr_chrjob[i].level_max > 0)
		{
			fprintf(fp, "LV   ABP  AreB\n");
			for (j = 0; j < ptr_chrjob[i].level_max; j++)
			{
				int no = ptr_chrjob[i].level[j].ability_no & 0x7f;

				fprintf(fp, "%2d:  %3d, %s\n", j + 1,
						ptr_chrjob[i].level[j].ability_point,
						((ptr_chrjob[i].level[j].ability_no & 0x80) == 0x00) ? chr_getcmdname(no) : chr_getabtname(no));
			}
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int chr_writecmdparam(const char *ptr_filename, const chrcmd_t *ptr_chrcmd)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrcmd == NULL)
	{
		assert(0);
		return 0;
	}

	/* R}hAreB̐lnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* R}hAreB̐lnf[^o͂ */
	fprintf(fp, "No.      O         "
				"ʒl   䗦     Œl "
				"Ҏ "
				"      ̗ \n");
	for (i = 0; i < NUM_CHRCMD; i++)
	{
		fprintf(fp, (ptr_chrcmd[i].param.effect_rate != 0) ?
					"%03d(%02X): %-10s :  %5d,%3d/16, %5d, %5d, %5d, %5d, %3d, %3d, %3d, %3d\n" :
					"%03d(%02X): %-10s :  %5d, %5d, %5d, %5d, %5d, %5d, %3d, %3d, %3d, %3d\n",
				i, i, chr_getcmdname(i),
				ptr_chrcmd[i].param.effect_power,
				ptr_chrcmd[i].param.effect_rate,
				ptr_chrcmd[i].param.hit_rate,
				ptr_chrcmd[i].param.effect_time,
				ptr_chrcmd[i].param.fixed_value,
				ptr_chrcmd[i].time_wait,
				ptr_chrcmd[i].increase_strength,
				ptr_chrcmd[i].increase_speed,
				ptr_chrcmd[i].increase_stamina,
				ptr_chrcmd[i].increase_magic_power);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int chr_writecmdelem(const char *ptr_filename, const chrcmd_t *ptr_chrcmd)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrcmd == NULL)
	{
		assert(0);
		return 0;
	}

	/* R}hAreB̑nf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* R}hAreB̑nf[^o͂ */
	fprintf(fp, "No.      O         "
				"\n");
	for (i = 0; i < NUM_CHRCMD; i++)
	{
		fprintf(fp, "%03d(%02X): %-10s : ", i, i, chr_getcmdname(i));

		/*  */
		str_printelem(fp, ptr_chrcmd[i].param.elem_effect, ELEM_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int chr_writecmdstate(const char *ptr_filename, const chrcmd_t *ptr_chrcmd)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrcmd == NULL)
	{
		assert(0);
		return 0;
	}

	/* R}hAreB̏ԕωnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* R}hAreB̏ԕωnf[^o͂ */
	fprintf(fp, "No.      O         "
				"ԕω1        2                3                4\n");
	for (i = 0; i < NUM_CHRCMD; i++)
	{
		fprintf(fp, "%03d(%02X): %-10s : ", i, i, chr_getcmdname(i));

		/* ԕω */
		str_printstate(fp, ptr_chrcmd[i].param.state_effect, STATE_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int chr_writecmdflag(const char *ptr_filename, const chrcmd_t *ptr_chrcmd)
{
	int i;
	FILE *fp = NULL;
	const char *str_direction[2][2] =
	{
		{"",	"/G"},
		{"G",		"G/"},
	};
	const char *str_range[2][2] =
	{
		{"P",	"P/S"},
		{"S",	"S/P"},
	};
	const char *str_roulette = "[bg";

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrcmd == NULL)
	{
		assert(0);
		return 0;
	}

	/* R}hAreB̃tOnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* R}hAreB̃tOnf[^o͂ */
	fprintf(fp, "No.      O         "
				"ʔ͈           "
				"L푰\n");
	for (i = 0; i < NUM_CHRCMD; i++)
	{
		fprintf(fp, "%03d(%02X): %-10s : ", i, i, chr_getcmdname(i));

		/* ʔ͈ */
		if (ptr_chrcmd[i].param.flag_roulette)
			fprintf(fp, "%-17s", str_roulette);
		else
		{
			fprintf(fp, "%-7s %-9s",
					str_direction[ptr_chrcmd[i].param.direction][ptr_chrcmd[i].param.change_direction],
					str_range[ptr_chrcmd[i].param.range][ptr_chrcmd[i].param.change_range]);
		}

		/* LX^[^Cv */
		if (ptr_chrcmd[i].param.montype_effect != 0x00)
		{
			fprintf(fp, "  ");
			str_printmontype(fp, ptr_chrcmd[i].param.montype_effect, MONTYPE_ALL);
		}

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int chr_writeabtparam(const char *ptr_filename, const chrabt_t *ptr_chrabt)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrabt == NULL)
	{
		assert(0);
		return 0;
	}

	/* AreB̐lnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* AreB̐lnf[^o͂ */
	fprintf(fp, "No.      O               "
				"      ̗ \n");
	for (i = 0; i < NUM_CHRABT; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :  %3d, %3d, %3d, %3d\n", i, i,
				chr_getabtname(i),
				ptr_chrabt[i].increase_strength,
				ptr_chrabt[i].increase_speed,
				ptr_chrabt[i].increase_stamina,
				ptr_chrabt[i].increase_magic_power);
	}

	return 1;
}
