/*
 *	ff5fileps.c
 *		Copyright (C)2001, 2002 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "ff5fileps.h"

#define STR_INVALID_CDROM_FILES		"Invalid CD-ROM files"

struct filelayout_tag
{
	long filepos_sfc;
	long filepos_ps;
	const char *ptr_filename_ps;
};

static size_t unitread(void *ptr_buf, long filepos, size_t size);

static FILE *fp_ps = NULL;						/* t@C|C^(vCXe[Vŗp) */
static char str_dirname_ps[1024] = "";			/* CD-ROM̃[gpX */
static const char *ptr_cur_filename_ps = NULL;	/* ݃I[ṽt@C */

int file_open_ps(const char *ptr_dirname)
{
	/* `FbN */
	if (ptr_dirname == NULL)
	{
		assert(0);
		return 0;
	}

	strncpy(str_dirname_ps, ptr_dirname, sizeof(str_dirname_ps));
	str_dirname_ps[sizeof(str_dirname_ps) - 1] = '\0';

	return 1;
}

int file_close_ps(void)
{
	int ret = 0;

	if (fp_ps != NULL)
		fclose(fp_ps);

	/* f[^ */
	fp_ps = NULL;
	*str_dirname_ps = '\0';
	ptr_cur_filename_ps = NULL;

	return ret;
}

size_t file_read_ps(void *ptr_buf, long filepos, size_t size)
{
	unsigned char *ptr_buf_now = ptr_buf;
	long filepos_now = filepos;
	size_t len_rest = size;

	/* `FbN */
	if (ptr_buf == NULL || filepos < 0)
	{
		assert(0);
		return 0;
	}

	/* w肳ꂽTCYǂݏI܂ŌJԂ */
	while (len_rest > 0)
	{
		size_t ret;

		if ((ret = unitread(ptr_buf_now, filepos_now, len_rest)) == 0)
			break;

		ptr_buf_now += ret;
		filepos_now += ret;
		len_rest    -= ret;
	}

	return size - len_rest;
}

static size_t unitread(void *ptr_buf, long filepos, size_t size)
{
	int i;
	int index;
	size_t ret;
	size_t len_read;
	long filepos_ps;
	static const struct filelayout_tag tbl_filelayout[] =
	{
		{0x000000, 0x000000, NULL},
		{0x00cf00, 0x0702e0, "nar/ff5_binx.bin"},
		{0x00e500, 0x000000, NULL},
		{0x00fec0, 0x000000, "btl/ff5_btl.bin"},
		{0x00ffc0, 0x000000, NULL},
		{0x030200, 0x000000, "kok/k_data.bin"},
		{0x038340, 0x000000, NULL},
		{0x03bb00, 0x03bf80, "nar/ff5_binx.bin"},
		{0x03eb00, 0x000100, "btl/ff5_btl.bin"},
		{0x040000, 0x000000, NULL},
		{0x070000, 0x041600, "nar/ff5_binx.bin"},
		{0x082220, 0x0731e0, "nar/ff5_binx.bin"},
		{0x0b0000, 0x00c540, "nar/ff5_binx.bin"},
		{0x0dfa40, 0x0a0fc0, "nar/ff5_binx.bin"},
		{0x0e9c00, 0x053820, "nar/ff5_binx.bin"},
		{0x0ed000, 0x001600, "btl/ff5_btl.bin"},
		{0x0f0000, 0x000000, "nar/ff5_binx.bin"},
		{0x0fc540, 0x056c20, "nar/ff5_binx.bin"},
		{0x0fd800, 0x0ab180, "nar/ff5_binx.bin"},
		{0x0fdc00, 0x000000, NULL},
		{0x0fe000, 0x040c00, "nar/ff5_binx.bin"},
		{0x0fea00, 0x0ab580, "nar/ff5_binx.bin"},
		{0x0ff0c0, 0x040300, "nar/ff5_binx.bin"},
		{0x0fffc0, 0x000000, NULL},
		{0x100000, 0x004600, "btl/ff5_btl.bin"},
		{0x120000, 0x010000, "mnu/men_bin.bin"},
		{0x14b180, 0x028180, "btl/ff5_btl.bin"},
		{0x150000, 0x000000, "ff5_btl2.bin"},
		{0x190000, 0x02d1cc, "btl/ff5_btl.bin"},
		{0x197c50, 0x000000, NULL},
		{0x1a0000, 0x0582e0, "nar/ff5_binx.bin"},
		{0x1b8000, 0x039b00, "ff5_bin3.bin"},
		{0x1be000, 0x035e1c, "btl/ff5_btl.bin"},
		{0x1c0000, 0x036d7c, "ff5_bin3.bin"},
		{0x1c2d85, 0x000000, "ff5_bin3.bin"},
		{0x1f9b00, 0x0b4a0c, "nar/ff5_binx.bin"},
		{0x1ffc00, 0x057ee0, "nar/ff5_binx.bin"},
		{0x200000, 0x000000, NULL},
	};

	/* `FbN */
	if (ptr_buf == NULL || filepos < 0)
	{
		assert(0);
		return 0;
	}

	/* ǂ̃t@Ĉǂ̈ʒuɏ񂪊i[Ă邩 */
	index = 0;
	for (i = 0; i < sizeof(tbl_filelayout) / sizeof(struct filelayout_tag) - 1; i++)
	{
		if (filepos < tbl_filelayout[i + 1].filepos_sfc)
		{
			index = i;
			break;
		}
	}

	if (tbl_filelayout[index].ptr_filename_ps == NULL)
		err_printf(STR_INVALID_CDROM_FILES);

	/* t@CقȂĂꍇ̓I[v */
	if (ptr_cur_filename_ps == NULL ||
		(ptr_cur_filename_ps != tbl_filelayout[index].ptr_filename_ps && strcmp(ptr_cur_filename_ps, tbl_filelayout[index].ptr_filename_ps) != 0))
	{
		char path[1024] = "";

		if (fp_ps != NULL)
			fclose(fp_ps);

		strcpy(path, str_dirname_ps);
		strcat(path, tbl_filelayout[index].ptr_filename_ps);
		if ((fp_ps = fopen(path, "rb")) == NULL)
			err_printf(STR_INVALID_CDROM_FILES " - %s", path);

		ptr_cur_filename_ps = tbl_filelayout[index].ptr_filename_ps;
	}

	/* V[Nʒu␳ */
	filepos_ps = ftell(fp_ps);
	if (filepos - tbl_filelayout[index].filepos_sfc != filepos_ps - tbl_filelayout[index].filepos_ps)
	{
		if (fseek(fp_ps, tbl_filelayout[index].filepos_ps + (filepos - tbl_filelayout[index].filepos_sfc), SEEK_SET))
			err_printf(STR_FATAL_ERROR_IN_FSEEK);
		filepos_ps = ftell(fp_ps);
	}

	/* f[^ǂݍ */
	len_read = (size < (size_t)(tbl_filelayout[index + 1].filepos_sfc - filepos)) ?
				size : (size_t)(tbl_filelayout[index + 1].filepos_sfc - filepos);
	if ((ret = fread(ptr_buf, 1, len_read, fp_ps)) < len_read)
		err_printf(STR_FATAL_ERROR_IN_FREAD);

	return ret;
}
