/*
 *	ff5item.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "ff5file.h"
#include "ff5decode.h"
#include "ff5item.h"

#define FILEPOS_ITEMNAME		(0x015980 + 0x0fba00)	/* = 0x111380 */
#define FILEPOS_ITEMDATA		(0x014600 + 0x0fba00)	/* = 0x110000 */
#define FILEPOS_ITEMELEM		(0x016b80 + 0x0fba00)	/* = 0x112580 */
#define FILEPOS_ITEMSTATE		(0x016cc0 + 0x0fba00)	/* = 0x1126c0 */
#define FILEPOS_ITEMPRICE		(0x017000 + 0x0fba00)	/* = 0x112a00 */
#define FILEPOS_ITEMMES_ADDR	(0x018600 + 0x0fba00)	/* = 0x114000 */
#define FILEPOS_ITEMMES			(0x018700 + 0x0fba00)	/* = 0x114100 */
#define NUM_ITEM_WEAPON			128
#define NUM_ITEM_PROTECTOR		96
#define NUM_ITEM_TOOL			32
#define NUM_ITEM_EQUIP			(NUM_ITEM_WEAPON + NUM_ITEM_PROTECTOR)	/* = 224 */
#define NUM_ITEM				(NUM_ITEM_WEAPON + NUM_ITEM_PROTECTOR + NUM_ITEM_TOOL)	/* = 256 */
#define NUM_ITEMELEM			64
#define NUM_ITEMSTATE			64
#define NUM_ITEMMES				41
#define SIZEOF_ITEMNAME			9
#define SIZEOF_ITEMDATA_EQUIP	12
#define SIZEOF_ITEMDATA_TOOL	8
#define SIZEOF_ITEMELEM			5
#define SIZEOF_ITEMSTATE		7
#define SIZEOF_ITEMPRICE		2
#define SIZEOF_OFFSETADDR		2
#define REGION_WEAPON_BEGIN		0
#define REGION_WEAPON_END		(REGION_WEAPON_BEGIN + NUM_ITEM_WEAPON - 1)			/* = 127 */
#define REGION_PROTECTOR_BEGIN	(REGION_WEAPON_END + 1)								/* = 128 */
#define REGION_PROTECTOR_END	(REGION_PROTECTOR_BEGIN + NUM_ITEM_PROTECTOR - 1)	/* = 223 */
#define REGION_TOOL_BEGIN		(REGION_PROTECTOR_END + 1)							/* = 224 */
#define REGION_TOOL_END			(REGION_TOOL_BEGIN + NUM_ITEM_TOOL - 1)				/* = 255 */

struct element_tag
{
	elem_t elem_half;		/*  */
	elem_t elem_none;		/*  */
	elem_t elem_absorb;		/* z */
	elem_t elem_double;		/* _ */
};

struct state_tag
{
	state_t state_defense;	/* ԕωւ̑ϐ */
	state_t state_apriori;	/* ŏKpԕω */
};

static void item_setname(void);

static char tbl_itemname[NUM_ITEM][256];	/* ACe̖Oi[e[u */
static int flag_setitemname = 0;			/* tbl_itemnameɃf[^i[Ă邩ǂ */

static const int tbl_inclease[8][2] =
{
	{ 0,  1},
	{ 0,  2},
	{ 0,  3},
	{-1,  1},
	{-1,  0},
	{-5,  5},
	{-5,  0},
	{ 0,  5},
};

const char *item_getname(int index)
{
	/* `FbN */
	if (index < 0 || NUM_ITEM <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setitemname)
		item_setname();

	return tbl_itemname[index];
}

static void item_setname(void)
{
	int i;

	/* ACe̖Of[^ */
	for (i = 0; i < NUM_ITEM; i++)
		*tbl_itemname[i] = '\0';

	/* ACe̖Of[^ǂݍ */
	for (i = 0; i < NUM_ITEM; i++)
	{
		int origin = 0;
		unsigned char buf[SIZEOF_ITEMNAME];

		file_read(buf, FILEPOS_ITEMNAME + SIZEOF_ITEMNAME * i, SIZEOF_ITEMNAME);
		origin = ((0xe3 <= buf[0] && buf[0] <= 0xf4) || buf[0] == 0xff) ? 1 : 0;
		dec_string(tbl_itemname[i], &buf[origin], sizeof(tbl_itemname[i]), SIZEOF_ITEMNAME - origin, 0);
	}

	flag_setitemname = 1;
}

itemdata_t *item_readdata(void)
{
	int i, j;
	long filepos;
	itemdata_t *buf_itemdata = NULL;
	struct element_tag tbl_element[NUM_ITEMELEM];
	struct state_tag tbl_state[NUM_ITEMSTATE];

	/* ACe̊ef[^i[obt@mۂ */
	if ((buf_itemdata = malloc(sizeof(itemdata_t) * NUM_ITEM)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* ACe(h)̑f[^ǂݍ */
	for (i = 0; i < NUM_ITEMELEM; i++)
	{
		unsigned char buf[SIZEOF_ITEMELEM];

		file_read(buf, FILEPOS_ITEMELEM + SIZEOF_ITEMELEM * i, SIZEOF_ITEMELEM);

		tbl_element[i].elem_half   = buf[3];
		tbl_element[i].elem_none   = buf[2];
		tbl_element[i].elem_absorb = buf[0];
		tbl_element[i].elem_double = buf[4];
	}

	/* ACe(h)̏ԕωf[^ǂݍ */
	for (i = 0; i < NUM_ITEMSTATE; i++)
	{
		unsigned char buf[SIZEOF_ITEMSTATE];

		file_read(buf, FILEPOS_ITEMSTATE + SIZEOF_ITEMSTATE * i, SIZEOF_ITEMSTATE);

		tbl_state[i].state_defense = bits_make32(buf[4], buf[5], buf[6], 0x00);
		tbl_state[i].state_apriori = bits_make32(buf[0], buf[1], buf[2], buf[3]);
	}

	/* ACẻif[^ǂݍ */
	for (i = 0; i < NUM_ITEM; i++)
	{
		unsigned char buf[SIZEOF_ITEMPRICE];
		long price;

		file_read(buf, FILEPOS_ITEMPRICE + SIZEOF_ITEMPRICE * i, SIZEOF_ITEMPRICE);

		price = buf[1];
		for (j = 0; j < (buf[0] & 0x3f); j++)
			price *= 10;

		buf_itemdata[i].price_buy  = price;
		buf_itemdata[i].price_sell = ((buf[0] & 0x40) != 0) ? 5 : price / 2;
	}

	/* ACe̊ef[^ǂݍ */
	filepos = FILEPOS_ITEMDATA;
	for (i = 0; i < NUM_ITEM; i++)
	{
		/* ACe̎ނ𔻕ʂ */
		if (REGION_WEAPON_BEGIN <= i && i <= REGION_WEAPON_END)
		{
			unsigned char buf[SIZEOF_ITEMDATA_EQUIP];

			filepos += file_read(buf, filepos, SIZEOF_ITEMDATA_EQUIP);

			/* lnf[^ */
			buf_itemdata[i].message_no                   = buf[4] & 0x7f;
			buf_itemdata[i].param.weapon.attack_power    = buf[7];
			buf_itemdata[i].param.weapon.equip_type      = buf[2] & 0x1f;
			buf_itemdata[i].param.weapon.effect_type     = buf[8] & 0x7f;
			buf_itemdata[i].param.weapon.using_effect_no = buf[6] & 0x7f;

			/* tOnf[^ */
			buf_itemdata[i].param.weapon.flag_break               = ((buf[6] & 0x80) != 0x00);
			buf_itemdata[i].param.weapon.flag_throw               = ((buf[2] & 0x40) == 0x00);
			buf_itemdata[i].param.weapon.flag_enchant             = ((buf[5] & 0x04) != 0x00);
			buf_itemdata[i].param.weapon.flag_initiative          = ((buf[5] & 0x20) != 0x00);
			buf_itemdata[i].param.weapon.flag_evade               = ((buf[5] & 0x40) != 0x00);
			buf_itemdata[i].param.weapon.flag_evade_f             = ((buf[5] & 0x80) != 0x00);
			buf_itemdata[i].param.weapon.flag_additional_effect   = ((buf[5] & 0x08) != 0x00);
			buf_itemdata[i].param.weapon.flag_additional_effect_r = ((buf[5] & 0x02) != 0x00);
			buf_itemdata[i].param.weapon.flag_2handed             = ((buf[4] & 0x80) != 0x00);

			if ((buf[3] & 0x80) != 0x00)
			{
				/* p[^ωf[^ */
				buf_itemdata[i].param.weapon.increase_strength =
					tbl_inclease[buf[3] & 0x07][((buf[3] & 0x40) != 0x00) ? 1 : 0];
				buf_itemdata[i].param.weapon.increase_speed =
					tbl_inclease[buf[3] & 0x07][((buf[3] & 0x20) != 0x00) ? 1 : 0];
				buf_itemdata[i].param.weapon.increase_stamina =
					tbl_inclease[buf[3] & 0x07][((buf[3] & 0x10) != 0x00) ? 1 : 0];
				buf_itemdata[i].param.weapon.increase_magic_power =
					tbl_inclease[buf[3] & 0x07][((buf[3] & 0x08) != 0x00) ? 1 : 0];

				buf_itemdata[i].param.weapon.elem_powerup = 0x00;
			}
			else
			{
				/* p[Abvf[^ */
				buf_itemdata[i].param.weapon.increase_strength    = 0;
				buf_itemdata[i].param.weapon.increase_speed       = 0;
				buf_itemdata[i].param.weapon.increase_stamina     = 0;
				buf_itemdata[i].param.weapon.increase_magic_power = 0;

				buf_itemdata[i].param.weapon.elem_powerup = buf[3] & 0x7f;
			}

			/* σf[^() */
			buf_itemdata[i].param.weapon.hit_rate               =  100;
			buf_itemdata[i].param.weapon.critical_hit_rate      =    0;
			buf_itemdata[i].param.weapon.magic_point            =    0;
			buf_itemdata[i].param.weapon.additional_effect_rate =    0;
			buf_itemdata[i].param.weapon.additional_effect_no   =    0;
			buf_itemdata[i].param.weapon.elem_attack            = 0x00;
			buf_itemdata[i].param.weapon.state_attack           = 0x00;
			buf_itemdata[i].param.weapon.montype_attack         = 0x00;

			/* σf[^(ݒ) */
			switch (buf_itemdata[i].param.weapon.effect_type)
			{
			case 0x31:	/* UEǉʔEǉʂ̔ԍ */
			case 0x32:
			case 0x33:
				buf_itemdata[i].param.weapon.elem_attack            = buf[ 9];
				buf_itemdata[i].param.weapon.additional_effect_rate = buf[10];
				buf_itemdata[i].param.weapon.additional_effect_no   = buf[11];
				break;

			case 0x00:	/* EǉʔEǉʂ̔ԍ */
			case 0x34:
			case 0x38:
			case 0x3a:
			case 0x49:
				buf_itemdata[i].param.weapon.hit_rate               = buf[ 9];
				buf_itemdata[i].param.weapon.additional_effect_rate = buf[10];
				buf_itemdata[i].param.weapon.additional_effect_no   = buf[11];
				break;

			case 0x37:	/* NeBJqbgEǉʔEǉʂ̔ԍ */
				buf_itemdata[i].param.weapon.critical_hit_rate      = buf[ 9];
				buf_itemdata[i].param.weapon.additional_effect_rate = buf[10];
				buf_itemdata[i].param.weapon.additional_effect_no   = buf[11];
				break;

			case 0x72:	/* LX^[^CvEǉʔEǉʂ̔ԍ */
			case 0x73:
				buf_itemdata[i].param.weapon.montype_attack         = buf[ 9];
				buf_itemdata[i].param.weapon.additional_effect_rate = buf[10];
				buf_itemdata[i].param.weapon.additional_effect_no   = buf[11];
				break;

			case 0x64:	/* 0x00EǉʔEǉʂ̔ԍ */
			case 0x7f:	/* sEǉʔEǉʂ̔ԍ */
				buf_itemdata[i].param.weapon.additional_effect_rate = buf[10];
				buf_itemdata[i].param.weapon.additional_effect_no   = buf[11];
				break;

			case 0x36:	/* ENeBJqbgEU */
			case 0x3b:
				buf_itemdata[i].param.weapon.hit_rate          = buf[ 9];
				buf_itemdata[i].param.weapon.critical_hit_rate = buf[10];
				buf_itemdata[i].param.weapon.elem_attack       = buf[11];
				break;

			case 0x35:	/* EԕωEԕω */
				buf_itemdata[i].param.weapon.hit_rate               = buf[ 9];
				buf_itemdata[i].param.weapon.additional_effect_rate = buf[10] & 0x7f;
				buf_itemdata[i].param.weapon.state_attack           = (state_t)buf[11] << (((buf[10] & 0x80) != 0x00) ? 0 : 8);
				break;

			case 0x3c:	/* ENeBJqbg?E}WbN|Cg */
				buf_itemdata[i].param.weapon.hit_rate          = buf[ 9];
				buf_itemdata[i].param.weapon.critical_hit_rate = buf[10];
				buf_itemdata[i].param.weapon.magic_point       = buf[11];
				break;

			case 0x0d:	/* Es(IU?)E0x00 */
				buf_itemdata[i].param.weapon.hit_rate = buf[9];
				break;

			case 0x30:	/* NeBJqbgE0x00E0x00 */
				buf_itemdata[i].param.weapon.critical_hit_rate = buf[9];
				break;

			case 0x06:	/* 0x00Es(IU?)EU */
				buf_itemdata[i].param.weapon.elem_attack = buf[11];
				break;

			case 0x6c:	/* 0x00Es(IU?)ELX^[^Cv*/
				buf_itemdata[i].param.weapon.montype_attack = buf[11];
				break;

			case 0x28:	/* s(tOn?)E_[Wl(ʃoCg)E_[Wl(ʃoCg) */
			case 0x39:	/* 0x00E0x00E0x00 */
			case 0x6e:	/* 0x00E0x00E0x00 */
			default:
				break;
			}

			buf_itemdata[i].type = ITEM_TYPE_WEAPON;
		}
		else if (REGION_PROTECTOR_BEGIN <= i && i <= REGION_PROTECTOR_END)
		{
			unsigned char buf[SIZEOF_ITEMDATA_EQUIP];

			filepos += file_read(buf, filepos, SIZEOF_ITEMDATA_EQUIP);

			/* lnf[^ */
			buf_itemdata[i].message_no                          = buf[4] & 0x7f;
			buf_itemdata[i].param.protector.defense_power       = buf[7];
			buf_itemdata[i].param.protector.evade_rate          = buf[6];
			buf_itemdata[i].param.protector.magic_defense_power = buf[9];
			buf_itemdata[i].param.protector.magic_evade_rate    = buf[8];
			buf_itemdata[i].param.protector.weight              = buf[1];
			buf_itemdata[i].param.protector.equip_type          = buf[2] & 0x1f;

			/* nf[^ */
			buf_itemdata[i].param.protector.elem_half   = tbl_element[buf[10] & 0x3f].elem_half;
			buf_itemdata[i].param.protector.elem_none   = tbl_element[buf[10] & 0x3f].elem_none;
			buf_itemdata[i].param.protector.elem_absorb = tbl_element[buf[10] & 0x3f].elem_absorb;
			buf_itemdata[i].param.protector.elem_double = tbl_element[buf[10] & 0x3f].elem_double;

			/* ԕωnf[^ */
			buf_itemdata[i].param.protector.state_defense = tbl_state[buf[11] & 0x3f].state_defense;
			buf_itemdata[i].param.protector.state_apriori = tbl_state[buf[11] & 0x3f].state_apriori;

			/* tOnf[^ */
			buf_itemdata[i].param.protector.flag_evade       = ((buf[5] & 0x40) != 0x00);
			buf_itemdata[i].param.protector.flag_magic_evade = ((buf[5] & 0x80) != 0x00);
			buf_itemdata[i].param.protector.flag_steal       = ((buf[5] & 0x10) != 0x00);
			buf_itemdata[i].param.protector.flag_capture     = ((buf[5] & 0x01) != 0x00);
			buf_itemdata[i].param.protector.flag_monk_fight  = ((buf[5] & 0x20) != 0x00);
			buf_itemdata[i].param.protector.flag_mp_half     = ((buf[5] & 0x08) != 0x00);
			buf_itemdata[i].param.protector.flag_sword_dance = ((buf[5] & 0x04) != 0x00);
			buf_itemdata[i].param.protector.flag_undead      = ((buf[5] & 0x02) != 0x00);

			if ((buf[3] & 0x80) != 0x00)
			{
				/* p[^ωf[^ */
				buf_itemdata[i].param.protector.increase_strength =
					tbl_inclease[buf[3] & 0x07][((buf[3] & 0x40) != 0x00) ? 1 : 0];
				buf_itemdata[i].param.protector.increase_speed =
					tbl_inclease[buf[3] & 0x07][((buf[3] & 0x20) != 0x00) ? 1 : 0];
				buf_itemdata[i].param.protector.increase_stamina =
					tbl_inclease[buf[3] & 0x07][((buf[3] & 0x10) != 0x00) ? 1 : 0];
				buf_itemdata[i].param.protector.increase_magic_power =
					tbl_inclease[buf[3] & 0x07][((buf[3] & 0x08) != 0x00) ? 1 : 0];

				buf_itemdata[i].param.protector.elem_powerup = 0x00;
			}
			else
			{
				/* p[Abvf[^ */
				buf_itemdata[i].param.protector.increase_strength    = 0;
				buf_itemdata[i].param.protector.increase_speed       = 0;
				buf_itemdata[i].param.protector.increase_stamina     = 0;
				buf_itemdata[i].param.protector.increase_magic_power = 0;

				buf_itemdata[i].param.protector.elem_powerup = buf[3] & 0x7f;
			}

			buf_itemdata[i].type = ITEM_TYPE_PROTECTOR;
		}
		else if (REGION_TOOL_BEGIN <= i && i <= REGION_TOOL_END)
		{
			unsigned char buf[SIZEOF_ITEMDATA_TOOL];

			filepos += file_read(buf, filepos, SIZEOF_ITEMDATA_TOOL);

			mag_setparam(&buf_itemdata[i].param.tool, buf, sizeof(buf));

			buf_itemdata[i].message_no = buf[3] & 0x7f;
			buf_itemdata[i].type       = ITEM_TYPE_TOOL;
		}
		else
			assert(0);
	}

	return buf_itemdata;
}

int item_writeparam(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ACe̐lnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ACe̐lnf[^o͂() */
	fprintf(fp, "No.      O                 "
				"  l   l "
				"U    CH               "
				"      ̗ \n");
	for (i = REGION_WEAPON_BEGIN; i <= REGION_WEAPON_END; i++)
	{
		fprintf(fp, "%03d(%02X): %-18s : %6ld,%6ld, %5d, %5d, %5d,               %3d, %3d, %3d, %3d\n", i, i,
				item_getname(i),
				ptr_itemdata[i].price_buy,
				ptr_itemdata[i].price_sell,
				ptr_itemdata[i].param.weapon.attack_power,
				ptr_itemdata[i].param.weapon.hit_rate,
				ptr_itemdata[i].param.weapon.critical_hit_rate,
				ptr_itemdata[i].param.weapon.increase_strength,
				ptr_itemdata[i].param.weapon.increase_speed,
				ptr_itemdata[i].param.weapon.increase_stamina,
				ptr_itemdata[i].param.weapon.increase_magic_power);
	}

	/* ACe̐lnf[^o͂(h) */
	fprintf(fp, "No.      O                 "
				"  l   l "
				"h  h    d "
				"      ̗ \n");
	for (i = REGION_PROTECTOR_BEGIN; i <= REGION_PROTECTOR_END; i++)
	{
		fprintf(fp, "%03d(%02X): %-18s : %6ld,%6ld, %5d, %5d, %5d, %5d, %5d, %3d, %3d, %3d, %3d\n", i, i,
				item_getname(i),
				ptr_itemdata[i].price_buy,
				ptr_itemdata[i].price_sell,
				ptr_itemdata[i].param.protector.defense_power,
				ptr_itemdata[i].param.protector.evade_rate,
				ptr_itemdata[i].param.protector.magic_defense_power,
				ptr_itemdata[i].param.protector.magic_evade_rate,
				ptr_itemdata[i].param.protector.weight,
				ptr_itemdata[i].param.protector.increase_strength,
				ptr_itemdata[i].param.protector.increase_speed,
				ptr_itemdata[i].param.protector.increase_stamina,
				ptr_itemdata[i].param.protector.increase_magic_power);
	}

	/* ACe̐lnf[^o͂() */
	fprintf(fp, "No.      O                 "
				"  l   l "
				"ʒl   䗦  Œl\n");
	for (i = REGION_TOOL_BEGIN; i <= REGION_TOOL_END; i++)
	{
		fprintf(fp, (ptr_itemdata[i].param.tool.effect_rate != 0) ?
					"%03d(%02X): %-18s : %6ld,%6ld, %5d,%3d/16, %5d, %5d\n" :
					"%03d(%02X): %-18s : %6ld,%6ld, %5d, %5d, %5d, %5d\n",
				i, i, item_getname(i),
				ptr_itemdata[i].price_buy,
				ptr_itemdata[i].price_sell,
				ptr_itemdata[i].param.tool.effect_power,
				ptr_itemdata[i].param.tool.effect_rate,
				ptr_itemdata[i].param.tool.hit_rate,
				ptr_itemdata[i].param.tool.fixed_value);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writeelem(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ACȇnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ACȇnf[^o͂ */
	fprintf(fp, "No.      O                 "
				"/U     "
				"          "
				"z          "
				"_          "
				"p[Abv\n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		fprintf(fp, "%03d(%02X): %-18s :", i, i, item_getname(i));

		switch (ptr_itemdata[i].type)
		{
		case ITEM_TYPE_WEAPON:
			fprintf(fp, " ");

			/* U */
			str_printelem(fp, ptr_itemdata[i].param.weapon.elem_attack, ELEM_ALL);
			fprintf(fp, "                                                        ");

			/* p[Abv */
			str_printelem(fp, ptr_itemdata[i].param.weapon.elem_powerup, ELEM_ALL);
			break;

		case ITEM_TYPE_PROTECTOR:
			fprintf(fp, " ");

			/*  */
			str_printelem(fp, ptr_itemdata[i].param.protector.elem_half, ELEM_ALL);
			fprintf(fp, "  ");

			/*  */
			str_printelem(fp, ptr_itemdata[i].param.protector.elem_none, ELEM_ALL);
			fprintf(fp, "  ");

			/* z */
			str_printelem(fp, ptr_itemdata[i].param.protector.elem_absorb, ELEM_ALL);
			fprintf(fp, "  ");

			/* _ */
			str_printelem(fp, ptr_itemdata[i].param.protector.elem_double, ELEM_ALL);
			fprintf(fp, "  ");

			/* p[Abv */
			str_printelem(fp, ptr_itemdata[i].param.protector.elem_powerup, ELEM_ALL);
			break;

		case ITEM_TYPE_TOOL:
			fprintf(fp, " ");

			/*  */
			str_printelem(fp, ptr_itemdata[i].param.tool.elem_effect, ELEM_ALL);
		}

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writestate(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ACȅԕωnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ACȅԕωnf[^o͂ */
	fprintf(fp, "No.      O                 "
				"ԕωϐ1    2                3                 "
				"ԕω1    2                3                4\n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		fprintf(fp, "%03d(%02X): %-18s :", i, i, item_getname(i));

		switch (ptr_itemdata[i].type)
		{
		case ITEM_TYPE_PROTECTOR:
			fprintf(fp, " ");

			/* ԕωւ̑ϐ */
			str_printstate(fp, ptr_itemdata[i].param.protector.state_defense, STATE_ALL & ~STATE_OTHER);
			fprintf(fp, "  ");

			/* ŏKpԕω */
			str_printstate(fp, ptr_itemdata[i].param.protector.state_apriori, STATE_ALL);
			break;

		case ITEM_TYPE_TOOL:
			fprintf(fp, "                                                     ");

			/* ԕω */
			str_printstate(fp, ptr_itemdata[i].param.tool.state_effect, STATE_ALL);
			break;
		}

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writeflag(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;
	const char *str_direction[2][2] =
	{
		{"",	"/G"},
		{"G",		"G/"},
	};
	const char *str_range[2][2] =
	{
		{"P",	"P/S"},
		{"S",	"S/P"},
	};

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ACẽtOnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ACẽtOnf[^o͂(i) */
	fprintf(fp, "No.      O                 "
				"   "
				"L푰\n");
	for (i = REGION_WEAPON_BEGIN; i <= REGION_PROTECTOR_END; i++)
	{
		fprintf(fp, "%03d(%02X): %-18s :", i, i, item_getname(i));

		switch (ptr_itemdata[i].type)
		{
		case ITEM_TYPE_WEAPON:
			/* etO */
			fprintf(fp, "  %2s   %2s",
					str_getflag(ptr_itemdata[i].param.weapon.flag_throw),
					str_getflag(ptr_itemdata[i].param.weapon.flag_enchant));

			/* LX^[^Cv */
			if (ptr_itemdata[i].param.weapon.montype_attack != 0x00)
			{
				fprintf(fp, "   ");
				str_printmontype(fp, ptr_itemdata[i].param.weapon.montype_attack, MONTYPE_ALL);
			}
			break;

		case ITEM_TYPE_PROTECTOR:
			break;
		}

		fputc('\n', fp);
	}

	/* ACẽtOnf[^o͂(iȊO) */
	fprintf(fp, "No.      O                 "
				"     "
				"ʔ͈\n");
	for (i = REGION_TOOL_BEGIN; i <= REGION_TOOL_END; i++)
	{
		fprintf(fp, "%03d(%02X): %-18s :  %2s   %2s   %2s   %2s   %-7s %-9s\n", i, i,
				item_getname(i),
				str_getflag(ptr_itemdata[i].param.tool.flag_throw),
				str_getflag(ptr_itemdata[i].param.tool.flag_drink),
				str_getflag(ptr_itemdata[i].param.tool.flag_mix),
				str_getflag(ptr_itemdata[i].param.tool.flag_reflect),
				str_direction[ptr_itemdata[i].param.tool.direction][ptr_itemdata[i].param.tool.change_direction],
				str_range[ptr_itemdata[i].param.tool.range][ptr_itemdata[i].param.tool.change_range]);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writeeffect(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_nothing = "----------------";

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ACěʃf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ACěʃf[^o͂ */
	fprintf(fp, "No.      O                 "
				"j ǉ ω  "
				"gp          "
				"ǉ          "
				"\n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		char str_statecng[256] = "";
		const char *ptr_name_using      = str_nothing;
		const char *ptr_name_additional = str_nothing;

		fprintf(fp, "%03d(%02X): %-18s :", i, i, item_getname(i));

		if (ptr_itemdata[i].type == ITEM_TYPE_WEAPON)
		{
			/* gp */
			ptr_name_using = (ptr_itemdata[i].param.weapon.using_effect_no != 0) ?
							mag_getname(ptr_itemdata[i].param.weapon.using_effect_no) : str_nothing;

			/* ǉ */
			if (ptr_itemdata[i].param.weapon.state_attack != 0x00)
			{
				char *ptr = str_statecng;
				for (j = 0; j < NUM_STATE; j++)
					ptr += sprintf(ptr, "%s", str_getstate(ptr_itemdata[i].param.weapon.state_attack, j));
				ptr_name_additional = str_statecng;
			}
			else if (ptr_itemdata[i].param.weapon.flag_additional_effect_r)
				ptr_name_additional = chr_getcmdname(ptr_itemdata[i].param.weapon.additional_effect_no);
			else if (ptr_itemdata[i].param.weapon.flag_additional_effect)
				ptr_name_additional = mag_getname(ptr_itemdata[i].param.weapon.additional_effect_no);
			else
				ptr_name_additional = str_nothing;

			/* etO */
			fprintf(fp, "  %2s   %2s   %2s   %-16s, %-16s,  %5d",
					str_getflag(ptr_itemdata[i].param.weapon.flag_break),
					str_getflag(ptr_itemdata[i].param.weapon.flag_additional_effect),
					str_getflag(ptr_itemdata[i].param.weapon.flag_additional_effect_r),
					ptr_name_using,
					ptr_name_additional,
					ptr_itemdata[i].param.weapon.additional_effect_rate);
		}

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writeparambug(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ACe̐lnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ACe̐lnf[^́Eo͂ */
	fprintf(fp, "No.      O                 "
				"h  h    d "
				"      ̗ \n");
	for (i = REGION_TOOL_BEGIN; i <= REGION_TOOL_END; i++)
	{
		unsigned char buf[SIZEOF_ITEMDATA_EQUIP];
		protector_t tmp_protector;

		file_read(buf, FILEPOS_ITEMDATA + SIZEOF_ITEMDATA_EQUIP * NUM_ITEM_EQUIP + SIZEOF_ITEMDATA_TOOL * (i - NUM_ITEM_EQUIP), SIZEOF_ITEMDATA_EQUIP);

		tmp_protector.defense_power       = buf[7];
		tmp_protector.evade_rate          = buf[6];
		tmp_protector.magic_defense_power = buf[9];
		tmp_protector.magic_evade_rate    = buf[8];
		tmp_protector.weight              = buf[1];

		if (tmp_protector.defense_power > 0x80)
			tmp_protector.defense_power -= 0x100;
		if (tmp_protector.magic_defense_power > 0x80)
			tmp_protector.magic_defense_power -= 0x100;

		if ((buf[3] & 0x80) != 0x00)
		{
			tmp_protector.increase_strength =
				tbl_inclease[buf[3] & 0x07][((buf[3] & 0x40) != 0x00) ? 1 : 0];
			tmp_protector.increase_speed =
				tbl_inclease[buf[3] & 0x07][((buf[3] & 0x20) != 0x00) ? 1 : 0];
			tmp_protector.increase_stamina =
				tbl_inclease[buf[3] & 0x07][((buf[3] & 0x10) != 0x00) ? 1 : 0];
			tmp_protector.increase_magic_power =
				tbl_inclease[buf[3] & 0x07][((buf[3] & 0x08) != 0x00) ? 1 : 0];

			tmp_protector.elem_powerup = 0x00;
		}
		else
		{
			tmp_protector.increase_strength    = 0;
			tmp_protector.increase_speed       = 0;
			tmp_protector.increase_stamina     = 0;
			tmp_protector.increase_magic_power = 0;

			tmp_protector.elem_powerup = buf[3] & 0x7f;
		}

		fprintf(fp, "%03d(%02X): %-18s :  %5d, %5d, %5d, %5d, %5d, %3d, %3d, %3d, %3d\n", i, i,
				item_getname(i),
				tmp_protector.defense_power,
				tmp_protector.evade_rate,
				tmp_protector.magic_defense_power,
				tmp_protector.magic_evade_rate,
				tmp_protector.weight,
				tmp_protector.increase_strength,
				tmp_protector.increase_speed,
				tmp_protector.increase_stamina,
				tmp_protector.increase_magic_power);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writeelembug(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;
	struct element_tag tbl_element[256];

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ACȇnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ACe(h)̑f[^ǂݍ */
	for (i = 0; i < 256; i++)
	{
		unsigned char buf[SIZEOF_ITEMELEM];

		file_read(buf, FILEPOS_ITEMELEM + SIZEOF_ITEMELEM * i, SIZEOF_ITEMELEM);

		tbl_element[i].elem_half   = buf[3];
		tbl_element[i].elem_none   = buf[2];
		tbl_element[i].elem_absorb = buf[0];
		tbl_element[i].elem_double = buf[4];
	}

	/* ACȇnf[^́Eo͂ */
	fprintf(fp, "No.      O                 "
						"          "
						"          "
						"z          "
						"_          "
						"p[Abv\n");
	for (i = REGION_TOOL_BEGIN; i <= REGION_TOOL_END; i++)
	{
		unsigned char buf[SIZEOF_ITEMDATA_EQUIP];

		file_read(buf, FILEPOS_ITEMDATA + SIZEOF_ITEMDATA_EQUIP * i, SIZEOF_ITEMDATA_EQUIP);

		fprintf(fp, "%03d(%02X): %-18s : ", i, i, item_getname(i));

		/*  */
		str_printelem(fp, tbl_element[buf[10]].elem_half, ELEM_ALL);
		fprintf(fp, "  ");

		/*  */
		str_printelem(fp, tbl_element[buf[10]].elem_none, ELEM_ALL);
		fprintf(fp, "  ");

		/* z */
		str_printelem(fp, tbl_element[buf[10]].elem_absorb, ELEM_ALL);
		fprintf(fp, "  ");

		/* _ */
		str_printelem(fp, tbl_element[buf[10]].elem_double, ELEM_ALL);
		fprintf(fp, "  ");

		/* p[Abv */
		str_printelem(fp, ((buf[3] & 0x80) == 0x00) ? buf[3] : 0x00, ELEM_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writestatbug(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;
	struct state_tag tbl_state[256];

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ACȅԕωnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ACe(h)̏ԕωf[^ǂݍ */
	for (i = 0; i < 256; i++)
	{
		unsigned char buf[SIZEOF_ITEMSTATE];

		file_read(buf, FILEPOS_ITEMSTATE + SIZEOF_ITEMSTATE * i, SIZEOF_ITEMSTATE);

		tbl_state[i].state_defense = bits_make32(buf[4], buf[5], buf[6], 0x00);
		tbl_state[i].state_apriori = bits_make32(buf[0], buf[1], buf[2], buf[3]);
	}

	/* ACȅԕωnf[^́Eo͂ */
	fprintf(fp, "No.      O                 "
				"ԕωϐ1    2                3                "
				"ԕω1    2                3                4\n");
	for (i = REGION_TOOL_BEGIN; i <= REGION_TOOL_END; i++)
	{
		unsigned char buf[SIZEOF_ITEMDATA_EQUIP];

		file_read(buf, FILEPOS_ITEMDATA + SIZEOF_ITEMDATA_EQUIP * i, SIZEOF_ITEMDATA_EQUIP);

		fprintf(fp, "%03d(%02X): %-18s : ", i, i, item_getname(i));

		/* ԕωւ̑ϐ */
		str_printstate(fp, tbl_state[buf[11]].state_defense, STATE_ALL & ~STATE_OTHER);
		fprintf(fp, " ");

		/* ŏKpԕω */
		str_printstate(fp, tbl_state[buf[11]].state_apriori, STATE_ALL | STATE_INVINCIBLE2 | STATE_HANGUP);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}
