/*
 *	ff5magic.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "ff5file.h"
#include "ff5decode.h"
#include "ff5magic.h"

#define FILEPOS_MAGNAME			(0x016280 + 0x0fba00)	/* = 0x111c80 */
#define FILEPOS_MAGDATA			(0x015180 + 0x0fba00)	/* = 0x110b80 */
#define NUM_MAG_SHORT			72
#define NUM_MAG_MIDDLE			15
#define NUM_MAG_LONG			169
#define NUM_MAG					(NUM_MAG_SHORT + NUM_MAG_MIDDLE + NUM_MAG_LONG)	/* = 256 */
#define SIZEOF_MAGNAME_SHORT	6
#define SIZEOF_MAGNAME_MIDDLE	6
#define SIZEOF_MAGNAME_LONG		9
#define SIZEOF_MAGDATA			8
#define REGION_SHORT_BEGIN		0
#define REGION_SHORT_END		(REGION_SHORT_BEGIN + NUM_MAG_SHORT - 1)	/* =  71 */
#define REGION_MIDDLE_BEGIN		(REGION_SHORT_END + 1)						/* =  72 */
#define REGION_MIDDLE_END		(REGION_MIDDLE_BEGIN + NUM_MAG_MIDDLE - 1)	/* =  86 */
#define REGION_LONG_BEGIN		(REGION_MIDDLE_END + 1)						/* =  87 */
#define REGION_LONG_END			(REGION_LONG_BEGIN + NUM_MAG_LONG - 1)		/* = 255 */

static void mag_setname(void);

static char tbl_magname[NUM_MAG][256];	/* @̖Oi[e[u */
static int flag_setmagname = 0;			/* tbl_magnameɃf[^i[Ă邩ǂ */

const char *mag_getname(int index)
{
	/* `FbN */
	if (index < 0 || NUM_MAG <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setmagname)
		mag_setname();

	return tbl_magname[index];
}

static void mag_setname(void)
{
	int i;
	long filepos;

	/* @̖Of[^ */
	for (i = 0; i < NUM_MAG; i++)
		*tbl_magname[i] = '\0';

	/* @̖Of[^ǂݍ */
	filepos = FILEPOS_MAGNAME;
	for (i = 0; i < NUM_MAG; i++)
	{
		int origin = 0;
		size_t len_read = 0;
		unsigned char buf[SIZEOF_MAGNAME_LONG];

		if (REGION_SHORT_BEGIN <= i && i <= REGION_SHORT_END)
		{
			origin = 1;
			len_read = SIZEOF_MAGNAME_SHORT;
		}
		else if (REGION_MIDDLE_BEGIN <= i && i <= REGION_MIDDLE_END)
		{
			origin = 0;
			len_read = SIZEOF_MAGNAME_MIDDLE;
		}
		else if (REGION_LONG_BEGIN <= i && i <= REGION_LONG_END)
		{
			origin = 0;
			len_read = SIZEOF_MAGNAME_LONG;
		}
		else
			assert(0);

		filepos += file_read(buf, filepos, len_read);
		dec_string(tbl_magname[i], &buf[origin], sizeof(tbl_magname[i]), len_read - origin, 0);
	}

	flag_setmagname = 1;
}

magdata_t *mag_readdata(void)
{
	int i;
	magdata_t *buf_magdata = NULL;

	/* @̊ef[^i[obt@mۂ */
	if ((buf_magdata = malloc(sizeof(magdata_t) * NUM_MAG)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* @̊ef[^ǂݍ */
	for (i = 0; i < NUM_MAG; i++)
	{
		unsigned char buf[SIZEOF_MAGDATA];

		file_read(buf, FILEPOS_MAGDATA + SIZEOF_MAGDATA * i, SIZEOF_MAGDATA);
		mag_setparam(&buf_magdata[i], buf, sizeof(buf));
	}

	return buf_magdata;
}

int mag_setparam(magdata_t *ptr_magdata, unsigned char *ptr_buf, size_t bufsize)
{
	/* `FbN */
	if (ptr_magdata == NULL || ptr_buf == NULL || bufsize < SIZEOF_MAGDATA)
		return 0;

	/* lnf[^ */
	ptr_magdata->magic_point = ptr_buf[3] & 0x7f;
	ptr_magdata->effect_type = ptr_buf[4] & 0x7f;

	/* tOnf[^ */
	ptr_magdata->flag_roulette    = ((ptr_buf[0] & 0x04) != 0x00);
	ptr_magdata->flag_throw       = ((ptr_buf[2] & 0x40) == 0x00);
	ptr_magdata->flag_drink       = ((ptr_buf[2] & 0x20) != 0x00);
	ptr_magdata->flag_mix         = ((ptr_buf[2] & 0x02) == 0x00);
	ptr_magdata->flag_reflect     = ((ptr_buf[3] & 0x80) == 0x00);
	ptr_magdata->direction        = ((ptr_buf[0] & 0x08) != 0x00) ? 1 : 0;
	ptr_magdata->range            = ((ptr_buf[0] & 0x40) != 0x00) ? 1 : 0;
	ptr_magdata->change_direction = ((ptr_buf[0] & 0x10) != 0x00) ? 1 : 0;
	ptr_magdata->change_range     = ((ptr_buf[0] & 0x80) != 0x00) ? 1 : 0;

	/* σf[^() */
	ptr_magdata->effect_power   = 0;
	ptr_magdata->effect_rate    = 0;
	ptr_magdata->hit_rate       = 0;
	ptr_magdata->effect_time    = 0;
	ptr_magdata->effect_level   = 0;
	ptr_magdata->fixed_value    = 0;
	ptr_magdata->elem_effect    = 0x00;
	ptr_magdata->state_effect   = 0x00;
	ptr_magdata->montype_effect = 0x00;

	switch (ptr_magdata->effect_type)
	{
	case 0x06:	/* U(gpEʒlE) */
	case 0x2d:	/* U(regŖł)(gpEʒlE) */
		ptr_magdata->effect_power = ptr_buf[6];
		ptr_magdata->elem_effect  = ptr_buf[7];
		break;

	case 0x07:	/* ԕωƂȂ䗦_[WU(Eʔ䗦Eԕω2(ꎞI)) */
		ptr_magdata->hit_rate     = ptr_buf[5];
		ptr_magdata->effect_rate  = ptr_buf[6];
		ptr_magdata->state_effect = bits_make32(0x00, ptr_buf[7], 0x00, 0x00);
		break;

	case 0x08:	/* h(ق)̍U(gpEʒlE) */
		ptr_magdata->effect_power = ptr_buf[6];
		ptr_magdata->elem_effect  = ptr_buf[7];
		break;

	case 0x09:	/* _[W{̍U(gpE{E) */
		ptr_magdata->fixed_value = ptr_buf[6];
		ptr_magdata->elem_effect = ptr_buf[7];
		break;

	case 0x0a:	/* U̍U(EʒlE) */
		ptr_magdata->hit_rate     = ptr_buf[5];
		ptr_magdata->effect_power = ptr_buf[6];
		ptr_magdata->elem_effect  = ptr_buf[7];
		break;

	case 0x0b:	/* ԕωƂȂU(EʒlEԕω1(iI)) */
		ptr_magdata->elem_effect  = ptr_buf[5];
		ptr_magdata->effect_power = ptr_buf[6];
		ptr_magdata->state_effect = bits_make32(ptr_buf[7], 0x00, 0x00, 0x00);
		break;

	case 0x0c:	/* ԕω(Xbv)ƂȂh(ق)̑U(EʒlE) */
		ptr_magdata->elem_effect  = ptr_buf[5];
		ptr_magdata->effect_power = ptr_buf[6];
		ptr_magdata->effect_time  = ptr_buf[7];
		break;

	case 0x0d:	/* qbg|CgzU(EʒlEgp) */
	case 0x0e:	/* }WbN|CgzU(EʒlEgp) */
		ptr_magdata->hit_rate     = ptr_buf[5];
		ptr_magdata->effect_power = ptr_buf[6];
		break;

	case 0x0f:	/* mU(EgpEgp) */
		ptr_magdata->hit_rate = ptr_buf[5];
		break;

	case 0x13:	/* ԕω(EԁEԕω2(ꎞI)) */
	case 0x3d:	/* ԕω𔺂mU(EԁEԕω2(ꎞI)) */
		ptr_magdata->hit_rate     = ptr_buf[5];
		ptr_magdata->effect_time  = ptr_buf[6];
		ptr_magdata->state_effect = bits_make32(0x00, ptr_buf[7], 0x00, 0x00);
		break;

	case 0x10:	/* (gpEʒlEgp) */
	case 0x11:	/* (ڕWP͕̂̎ʌ)(gpEʒlEgp) */
	case 0x1b:	/* qbg|CgzU(zSɔz)(gpEʒlEgp) */
		ptr_magdata->effect_power = ptr_buf[6];
		break;

	case 0x12:	/* ԕω(EsEԕω1(iI)) */
	case 0x4f:	/* ꂢ(EgpEԕω1(iI)) */
		ptr_magdata->hit_rate     = ptr_buf[5];
		ptr_magdata->state_effect = bits_make32(ptr_buf[7], 0x00, 0x00, 0x00);
		break;

	case 0x14:	/* ԕω(EԁEԕω3(⏕@)) */
		ptr_magdata->hit_rate     = ptr_buf[5];
		ptr_magdata->effect_time  = ptr_buf[6];
		ptr_magdata->state_effect = bits_make32(0x00, 0x00, ptr_buf[7], 0x00);
		break;

	case 0x15:	/* ԕω(])(EgpEԕω1(iI)) */
	case 0x17:	/* ԕω(snX^[ւ̓)(EgpEԕω1(iI)) */
		ptr_magdata->hit_rate     = ptr_buf[5];
		ptr_magdata->effect_time  = ptr_buf[6];
		ptr_magdata->state_effect = bits_make32(ptr_buf[7], 0x00, 0x00, 0x00);
		break;

	case 0x16:	/* ԕω(ԉƂȂ)(Eԕω3(⏕@)Eԕω3(⏕@)) */
		ptr_magdata->hit_rate     = ptr_buf[5];
/*		ptr_magdata->state_effect = bits_make32(0x00, 0x00, ptr_buf[6], 0x00); */
		ptr_magdata->state_effect = bits_make32(0x00, 0x00, ptr_buf[7], 0x00);
		break;

	case 0x18:	/* n@(snX^[ւ̓ʂȂ)(EgpEgp) */
		ptr_magdata->hit_rate = ptr_buf[5];
		break;

	case 0x19:	/* ԕω(ԕω1(iI)Eԕω2(ꎞI)Eԕω3(⏕@)) */
	case 0x27:	/* ԕω(ԕω1(iI)Eԕω2(ꎞI)Eԕω3(⏕@)) */
		ptr_magdata->state_effect = bits_make32(ptr_buf[5], ptr_buf[6], ptr_buf[7], 0x00);
		break;

	case 0x1a:	/* 퓬s\(EʒlEʔ䗦) */
		ptr_magdata->hit_rate     = ptr_buf[5];
		ptr_magdata->effect_power = ptr_buf[6];
		ptr_magdata->effect_rate  = ptr_buf[7];
		break;

	case 0x1c:	/* zE(zEE_) */
		ptr_magdata->elem_effect = ptr_buf[5];
/*		ptr_magdata->elem_effect = ptr_buf[6]; */
		break;

	case 0x1f:	/* ~[g(EgpEgp) */
	case 0x20:	/* e|EƂ񂸂Eނ肾(EgpEgp) */
		ptr_magdata->hit_rate = ptr_buf[5];
		break;

	case 0x23:	/* A[XEH[(gpEWEgp) */
	case 0x46:	/* ɂȂ(M̌WE_[W̌WEgp) */
		ptr_magdata->fixed_value = ptr_buf[6];
		break;

	case 0x24:	/* 񕜗ʌŒ̃qbg|Cg(ʒlE{Egp) */
	case 0x25:	/* 񕜗ʌŒ̃}WbN|Cg(ʒlE{Egp) */
		ptr_magdata->effect_power = ptr_buf[5];
		ptr_magdata->fixed_value  = ptr_buf[6];
		break;

	case 0x28:	/* Œ_[WU(E_[Wl(ʃoCg)E_[Wl(ʃoCg)) */
		ptr_magdata->hit_rate    = ptr_buf[5];
		ptr_magdata->fixed_value = (ptr_buf[7] << 8) | ptr_buf[6];
		break;

	case 0x29:	/* ԕω(E̐鍐̃JEgEԕω4(̑)) */
		ptr_magdata->hit_rate     = ptr_buf[5];
		ptr_magdata->fixed_value  = ptr_buf[6];
		ptr_magdata->state_effect = bits_make32(0x00, 0x00, 0x00, ptr_buf[7]);
		break;

	case 0x2a:	/* ԕω(Xbv)ƂȂ䗦_[WU(Eʔ䗦E) */
		ptr_magdata->elem_effect = ptr_buf[5];
		ptr_magdata->effect_rate = ptr_buf[6];
		ptr_magdata->effect_time = ptr_buf[7];
		break;

	case 0x2b:	/* ɑ΂ԕωƂȂ䗦_[WU(Eʔ䗦Eԕω1(iI)) */
		ptr_magdata->elem_effect  = ptr_buf[5];
		ptr_magdata->effect_rate  = ptr_buf[6];
		ptr_magdata->state_effect = bits_make32(ptr_buf[7], 0x00, 0x00, 0x00);
		break;

	case 0x2c:	/* ԕω(2ނ̂ǂ炩̌)(ԕω1(iI)EԁEԕω2(ꎞI)) */
		ptr_magdata->state_effect = bits_make32(ptr_buf[5], ptr_buf[7], 0x00, 0x00);
		ptr_magdata->effect_time  = ptr_buf[6];
		break;

	case 0x2e:	/* ɂ݂̂邬(EʒlEԕω1(iI)) */
		ptr_magdata->hit_rate     = ptr_buf[5];
		ptr_magdata->effect_power = ptr_buf[6];
		ptr_magdata->state_effect = bits_make32(ptr_buf[7], 0x00, 0x00, 0x00);
		break;

	case 0x3e:	/* ς(mʂɂȂmEԁEԕω4) */
		ptr_magdata->hit_rate     = ptr_buf[5];
		ptr_magdata->effect_time  = ptr_buf[6];
		ptr_magdata->state_effect = bits_make32(0x00, 0x00, 0x00, ptr_buf[7]);
		break;

	case 0x3f:	/* 퓬s\ɏԕω𔺂U(gpEʒlEgp) */
		ptr_magdata->effect_power = ptr_buf[6];
		break;

	case 0x40:	/* ύX(EύX̎ʁEgp) */
		ptr_magdata->hit_rate = ptr_buf[5];
		break;

	case 0x42:	/*  + ԕω(EʒlEԕω1(iI)) */
		ptr_magdata->hit_rate     = ptr_buf[5];
		ptr_magdata->effect_power = ptr_buf[6];
		ptr_magdata->state_effect = bits_make32(ptr_buf[7], 0x00, 0x00, 0x00);
		break;

	case 0x43:	/* ʂ(EgpEgp) */
	case 0x45:	/* Ȃ(EgpEgp) */
	case 0x49:	/* /X̃bh(EǉʔEǉʂ̔ԍ) */
		ptr_magdata->hit_rate = ptr_buf[5];
		break;

	case 0x47:	/* Ȃ߂(gpEԁEԕω3(⏕@)) */
		ptr_magdata->effect_time  = ptr_buf[6];
		ptr_magdata->state_effect = bits_make32(0x00, 0x00, ptr_buf[7], 0x00);
		break;

	case 0x4b:	/* xnԕω(xEgpEԕω1(iI)) */
		ptr_magdata->effect_level = ptr_buf[5];
		ptr_magdata->state_effect = bits_make32(ptr_buf[7], 0x00, 0x00, 0x00);
		break;

	case 0x4c:	/* xnԕω(xEԁEԕω2(ꎞI)) */
		ptr_magdata->effect_level = ptr_buf[5];
		ptr_magdata->effect_time  = ptr_buf[6];
		ptr_magdata->state_effect = bits_make32(0x00, ptr_buf[7], 0x00, 0x00);
		break;

	case 0x4d:	/* xn䗦_[WU(xEʔ䗦Egp) */
		ptr_magdata->effect_level = ptr_buf[5];
		ptr_magdata->effect_rate  = ptr_buf[6];
		break;

	case 0x4e:	/* xnU(xEʒlEgp) */
		ptr_magdata->effect_level = ptr_buf[5];
		ptr_magdata->effect_power = ptr_buf[6];
		break;

	case 0x50:	/* Sup`(EgpEԕω1(iI)) */
		ptr_magdata->hit_rate     = ptr_buf[5];
		ptr_magdata->state_effect = bits_make32(ptr_buf[7], 0x00, 0x00, 0x00);
		break;

	case 0x51:	/* Xe[^X㏸ʂACe(EΏۂƂȂXe[^XE㏸l) */
		ptr_magdata->hit_rate    = ptr_buf[5];
		ptr_magdata->fixed_value = ptr_buf[7];
		break;

	case 0x52:	/* ԕω(XbvƂȂ)(Eԕω3(⏕@)Eԕω3(⏕@)) */
		ptr_magdata->hit_rate     = ptr_buf[5];
		ptr_magdata->state_effect = bits_make32(0x00, 0x00, ptr_buf[7], 0x00);
		break;

	case 0x53:	/* }WbN|Cgɑ΂䗦_[WU(Eʔ䗦Egp) */
		ptr_magdata->hit_rate    = ptr_buf[5];
		ptr_magdata->effect_rate = ptr_buf[6];
		break;

	case 0x54:	/* qbg|Cǧ_[WƂU(EgpEgp) */
	case 0x6d:	/* qbg|Cǧ_[WƂU(qbg|Cgz)(EgpEgp) */
		ptr_magdata->effect_power = ptr_buf[5];
		break;

	case 0x55:	/* 䂤䂤(gpEgpEԕω1(iI)) */
		ptr_magdata->state_effect = bits_make32(ptr_buf[7], 0x00, 0x00, 0x00);
		break;

	case 0x57:	/* ԕω(2ޗ̌)(ԕω1(iI)EԁEԕω2(ꎞI)) */
		ptr_magdata->effect_time  = ptr_buf[6];
		ptr_magdata->state_effect = bits_make32(ptr_buf[5], ptr_buf[7], 0x00, 0x00);
		break;

	case 0x58:	/* }ChuXg(EʒlE) */
		ptr_magdata->hit_rate     = ptr_buf[5];
		ptr_magdata->effect_power = ptr_buf[6];
		ptr_magdata->effect_time  = ptr_buf[7];
		break;

	case 0x5c:	/* ق悤(EgpEԕω1(iI)) */
		ptr_magdata->hit_rate     = ptr_buf[5];
		ptr_magdata->state_effect = bits_make32(ptr_buf[7], 0x00, 0x00, 0x00);
		break;

	case 0x5d:	/* ԕω(]r)(EgpEgp) */
		ptr_magdata->hit_rate = ptr_buf[5];
		break;

	case 0x5e:	/* p[^ω(gpEgpEωΏۂƂȂ\͒l) */
		break;

	case 0x5f:	/* ̎ނ̓Gɂ̂ݗLȏԕω(Xbv)ƂȂh(ق)̍U(LX^[^CvEʒlE) */
		ptr_magdata->montype_effect = ptr_buf[5];
		ptr_magdata->effect_power   = ptr_buf[6];
		ptr_magdata->effect_time    = ptr_buf[7];
		break;

	case 0x6b:	/* X^[^CvtEԕωϐtEUp[Abv(X^[^CvEԕω1(iI)E) */
		ptr_magdata->montype_effect = ptr_buf[5];
		ptr_magdata->state_effect   = bits_make32(ptr_buf[6], 0x00, 0x00, 0x00);
		ptr_magdata->elem_effect    = ptr_buf[7];
		break;

	case 0x6c:	/* ̎ނ̓GɃNeBJqbgU(gpEʒlELX^[^Cv) */
		ptr_magdata->effect_power   = ptr_buf[6];
		ptr_magdata->montype_effect = ptr_buf[7];
		break;

	case 0x70:	/* [z[(gpEgpEԕω4(̑)) */
		ptr_magdata->hit_rate     = ptr_buf[5];
		ptr_magdata->state_effect = bits_make32(0x00, 0x00, 0x00, ptr_buf[7]);
		break;

	case 0x00:	/* Ȃ(EǉʔEǉʂ̔ԍ) */
	case 0x01:	/* s(gpEgpEgp) */
	case 0x02:	/* s(gpEgpEgp) */
	case 0x03:	/* @(ID?EtEtԕω1(iI)) */
	case 0x04:	/* @(ID?EtEtԕω2(ꎞI)) */
	case 0x05:	/* @(ID?EtE?) */
	case 0x1d:	/* CuEׂE݂Ԃ(̎ʁEgpEgp) */
	case 0x1e:	/* Xs[h(gpEgpEgp) */
	case 0x21:	/* ^[(gpEgpEgp) */
	case 0x22:	/* NCbN(gpEgpEgp) */
	case 0x26:	/* qbg|CgE}WbN|Cg(񕜂p[^̎ʁEgpEgp) */
	case 0x30:	/* fE(NeBJqbgEgpEgp) */
	case 0x44:	/* Ƃ(gpEgpEgp) */
	case 0x48:	/* Ƃ炦(gpEgpEgp) */
	case 0x4a:	/* ǂ(sEsEs) */
	case 0x59:	/* 傶̂(gpEgpEgp) */
	case 0x5a:	/* zCgEBh(gpEgpEgp) */
	case 0x61:	/* Xg[J[(gpEgpEgp) */
	case 0x62:	/* ruX(gpEgpEgp) */
	case 0x63:	/* OhNX(gpEgpEgp) */
	case 0x65:	/* Pbg(sEsEs) */
	case 0x66:	/* ^[QbeBO(gpEgpEgp) */
	case 0x68:	/* イ傤(gpEgpEgp) */
	case 0x69:	/* (gpEgpEgp) */
	case 0x6a:	/* s(gpEgpEgp) */
	case 0x6f:	/* (gpEgpEgp) */
	case 0x71:	/* x_E(gpEgpEgp) */
	case 0x7f:	/* Ȃ(gpEgpEgp) */
	default:
		break;
	}

	return 1;
}

int mag_writeparam(const char *ptr_filename, const magdata_t *ptr_magdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_magdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* @̐lnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* @̐lnf[^o͂ */
	fprintf(fp, "No.      O                 "
				"MP ʒl   䗦     x Œl\n");
	for (i = 0; i < NUM_MAG; i++)
	{
		fprintf(fp, (ptr_magdata[i].effect_rate != 0) ?
					"%03d(%02X): %-18s :  %5d, %5d,%3d/16, %5d, %5d, %5d, %5d\n" :
					"%03d(%02X): %-18s :  %5d, %5d, %5d, %5d, %5d, %5d, %5d\n",
				i, i, mag_getname(i),
				ptr_magdata[i].magic_point,
				ptr_magdata[i].effect_power,
				ptr_magdata[i].effect_rate,
				ptr_magdata[i].hit_rate,
				ptr_magdata[i].effect_time,
				ptr_magdata[i].effect_level,
				ptr_magdata[i].fixed_value);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mag_writeelem(const char *ptr_filename, const magdata_t *ptr_magdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_magdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* @̑nf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* @̑nf[^o͂ */
	fprintf(fp, "No.      O                 "
				"\n");
	for (i = 0; i < NUM_MAG; i++)
	{
		fprintf(fp, "%03d(%02X): %-18s : ", i, i, mag_getname(i));

		/*  */
		str_printelem(fp, ptr_magdata[i].elem_effect, ELEM_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mag_writestate(const char *ptr_filename, const magdata_t *ptr_magdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_magdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* @̏ԕωnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* @̏ԕωnf[^o͂ */
	fprintf(fp, "No.      O                 "
				"ԕω1        2                3                4\n");
	for (i = 0; i < NUM_MAG; i++)
	{
		fprintf(fp, "%03d(%02X): %-18s : ", i, i, mag_getname(i));

		/* ԕω */
		str_printstate(fp, ptr_magdata[i].state_effect, STATE_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mag_writeflag(const char *ptr_filename, const magdata_t *ptr_magdata)
{
	int i;
	FILE *fp = NULL;
	const char *str_direction[2][2] =
	{
		{"",	"/G"},
		{"G",		"G/"},
	};
	const char *str_range[2][2] =
	{
		{"P",	"P/S"},
		{"S",	"S/P"},
	};
	const char *str_roulette = "[bg";

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_magdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* @̃tOnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* @̃tOnf[^o͂ */
	fprintf(fp, "No.      O                 "
				"  "
				"ʔ͈           "
				"L푰\n");
	for (i = 0; i < NUM_MAG; i++)
	{
		/* etO */
		fprintf(fp, "%03d(%02X): %-18s :  %2s   ", i, i, mag_getname(i),
				str_getflag(ptr_magdata[i].flag_reflect));

		/* ʔ͈ */
		if (ptr_magdata[i].flag_roulette)
			fprintf(fp, "%-17s", str_roulette);
		else
		{
			fprintf(fp, "%-7s %-9s",
					str_direction[ptr_magdata[i].direction][ptr_magdata[i].change_direction],
					str_range[ptr_magdata[i].range][ptr_magdata[i].change_range]);
		}

		/* LX^[^Cv */
		if (ptr_magdata[i].montype_effect != 0x00)
		{
			fprintf(fp, "  ");
			str_printmontype(fp, ptr_magdata[i].montype_effect, MONTYPE_ALL);
		}

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}
