/*
 *	ff5misc.c
 *		Copyright (C)2002 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "ff5file.h"
#include "ff5decode.h"
#include "ff5item.h"
#include "ff5magic.h"
#include "ff5misc.h"

#define FILEPOS_TRDATA		0x113210
#define NUM_TREASURE		252
#define SIZEOF_TRDATA		4

treasure_t *tr_readdata(void)
{
	int i, j;
	treasure_t *buf_trdata = NULL;

	/* 󔠂̊ef[^i[obt@mۂ */
	if ((buf_trdata = malloc(sizeof(treasure_t) * NUM_TREASURE)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* 󔠂̊ef[^ǂݍ */
	for (i = 0; i < NUM_TREASURE; i++)
	{
		unsigned char buf[SIZEOF_TRDATA];

		file_read(buf, FILEPOS_TRDATA + SIZEOF_TRDATA * i, SIZEOF_TRDATA);

		/* 󔠂̒g̎ނ𔻕ʂ */
		buf_trdata[i].flag_monster = ((buf[2] & 0x80) != 0);
		if (buf_trdata[i].flag_monster)
		{
			if ((buf[2] & 0x40) != 0)
				buf_trdata[i].type = TREASURE_TYPE_MAGIC;
			else
				buf_trdata[i].type = TREASURE_TYPE_ITEM;
		}
		else
		{
			if ((buf[2] & 0x40) != 0)
				buf_trdata[i].type = TREASURE_TYPE_ITEM;
			else if ((buf[2] & 0x20) != 0)
				buf_trdata[i].type = TREASURE_TYPE_MAGIC;
			else
				buf_trdata[i].type = TREASURE_TYPE_GIL;
		}

		/* 󔠂̒g̃f[^ǂݍ */
		switch (buf_trdata[i].type)
		{
		case TREASURE_TYPE_ITEM:
			buf_trdata[i].param.item_no = buf[3];
			break;

		case TREASURE_TYPE_MAGIC:
			buf_trdata[i].param.magic_no = buf[3];
			break;

		case TREASURE_TYPE_GIL:
			{
				long gil = buf[3];

				for (j = 0; j < (buf[2] & 0x1f); j++)
					gil *= 10;

				buf_trdata[i].param.gil = gil;
			}
			break;

		default:
			assert(0);
		}
	}

	return buf_trdata;
}

int tr_writedata(const char *ptr_filename, const treasure_t *ptr_trdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_trdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* 󔠂̊ef[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* 󔠂̊ef[^o͂ */
	fprintf(fp, "No.                     "
				" G\n");
	for (i = 0; i < NUM_TREASURE; i++)
	{
		fprintf(fp, "%03d(%02X): ", i, i);

		switch (ptr_trdata[i].type)
		{
		case TREASURE_TYPE_ITEM:
			fprintf(fp, "%-16s", item_getname(ptr_trdata[i].param.item_no));
			break;

		case TREASURE_TYPE_MAGIC:
			fprintf(fp, "%-16s", mag_getname(ptr_trdata[i].param.magic_no));
			break;

		case TREASURE_TYPE_GIL:
			{
				char str_gil[256] = "";

				sprintf(str_gil, "%dM", ptr_trdata[i].param.gil);
				fprintf(fp, "%-16s", str_gil);
			}
			break;

		default:
			assert(0);
		}

		fprintf(fp, " :  %2s\n", str_getflag(ptr_trdata[i].flag_monster));
	}

	fclose(fp);
	fp = NULL;

	return 1;
}
