/*
 *	ff5monster.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include "errmes.h"
#include "ff5file.h"
#include "ff5decode.h"
#include "ff5monster.h"

#define FILEPOS_MONNAME			(0x00a200 + 0x0fba00)	/* = 0x105c00 */
#define FILEPOS_MONDATA			(0x004600 + 0x0fba00)	/* = 0x100000 */
#define FILEPOS_MONITEM			(0x009600 + 0x0fba00)	/* = 0x105000 */
#define FILEPOS_MONCTRL			(0x009c00 + 0x0fba00)	/* = 0x105600 */
#define FILEPOS_MONFREE			(0x00cc00 + 0x0fba00)	/* = 0x108600 */
#define FILEPOS_MONTECH			(0x00df00 + 0x0fba00)	/* = 0x109900 */
#define FILEPOS_MONACT_ADDR		(0x00e200 + 0x0fba00)	/* = 0x109c00 */
#define FILEPOS_MONACT			(0x00e500 + 0x0fba00)	/* = 0x109f00 */
#define FILEPOS_MONTECHNAME		(0x00cd00 + 0x0fba00)	/* = 0x108700 */
#define FILEPOS_MONMES_ADDR		(0x013600 + 0x0fba00)	/* = 0x10f000 */
#define FILEPOS_MONMES			(0x0137d4 + 0x0fba00)	/* = 0x10f1d4 */
#define FILEPOS_MONPARTY		(0x007600 + 0x0fba00)	/* = 0x103000 */
#define FILEPOS_MONENC			(0x00ae00 + 0x0fba00)	/* = 0x106800 */
#define FILEPOS_MONMAP			(0x00c000 + 0x0fba00)	/* = 0x107a00 */
#define NUM_MON					384
#define NUM_MONTECH				64
#define NUM_MONMES				234
#define NUM_MONPARTY			512
#define NUM_MONENC				256
#define NUM_MONMAP				3
#define NUM_MONMAPAREA			8
#define SIZEOF_MONNAME			8
#define SIZEOF_MONDATA			32
#define SIZEOF_MONITEM			4
#define SIZEOF_MONCTRL			4
#define SIZEOF_MONFREE			1
#define SIZEOF_MONTECH			2
#define SIZEOF_MONTECHNAME		8
#define SIZEOF_MONPARTY			16
#define SIZEOF_MONENC			8
#define SIZEOF_MONMAP			8
#define SIZEOF_OFFSETADDR		2

static void mon_setname(void);
static void mon_settechname(void);
static void mon_setmes(void);
static const char *mon_gettechname(int index);
static const char *mon_getmes(int index);
static void mon_printact(FILE *stream, int index, mondata_t mondata);

static char tbl_monname[NUM_MON][256];			/* X^[̖Oi[e[u */
static char tbl_montechname[NUM_MONTECH][256];	/* uƂ킴v̖Oi[e[u */
static char tbl_monmes[NUM_MONMES][256];		/* X^[̃bZ[Wi[e[u */
static int flag_setmonname     = 0;				/* tbl_monnameɃf[^i[Ă邩ǂ */
static int flag_setmontechname = 0;				/* tbl_montechnameɃf[^i[Ă邩ǂ */
static int flag_setmonmes      = 0;				/* tbl_monmesɃf[^i[Ă邩ǂ */

const char *mon_getname(int index)
{
	/* `FbN */
	if (index < 0 || NUM_MON <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setmonname)
		mon_setname();

	return tbl_monname[index];
}

static const char *mon_gettechname(int index)
{
	/* `FbN */
	if (index < 0 || NUM_MONTECH <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setmontechname)
		mon_settechname();

	return tbl_montechname[index];
}

static const char *mon_getmes(int index)
{
	/* `FbN */
	if (index < 0 || NUM_MONMES <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setmonmes)
		mon_setmes();

	return tbl_monmes[index];
}

static void mon_setname(void)
{
	int i;

	/* X^[̖Of[^ */
	for (i = 0; i < NUM_MON; i++)
		*tbl_monname[i] = '\0';

	/* X^[̖Of[^ǂݍ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONNAME];

		file_read(buf, FILEPOS_MONNAME + SIZEOF_MONNAME * i, SIZEOF_MONNAME);
		dec_string(tbl_monname[i], buf, sizeof(tbl_monname[i]), SIZEOF_MONNAME, 0);
	}

	flag_setmonname = 1;
}

static void mon_settechname(void)
{
	int i;

	/* X^[́uƂ킴v̖Of[^ */
	for (i = 0; i < NUM_MONTECH; i++)
		*tbl_montechname[i] = '\0';

	/* X^[́uƂ킴v̖Of[^ǂݍ */
	for (i = 0; i < NUM_MONTECH; i++)
	{
		unsigned char buf[SIZEOF_MONTECHNAME];

		file_read(buf, FILEPOS_MONTECHNAME + SIZEOF_MONTECHNAME * i, SIZEOF_MONTECHNAME);
		dec_string(tbl_montechname[i], buf, sizeof(tbl_montechname[i]), SIZEOF_MONTECHNAME, 0);
	}

	flag_setmontechname = 1;
}

static void mon_setmes(void)
{
	int i;
	long filepos;
	int tbl_offset[NUM_MONMES + 1];

	/* X^[̃bZ[Wf[^ */
	for (i = 0; i < NUM_MONMES; i++)
		*tbl_monmes[i] = '\0';

	/* X^[̃bZ[Wf[^i[ĂAhX(ւ̃ItZbg)ǂݍ */
	for (i = 0; i < NUM_MONMES; i++)
	{
		unsigned char buf[SIZEOF_OFFSETADDR];

		file_read(buf, FILEPOS_MONMES_ADDR + SIZEOF_OFFSETADDR * i, SIZEOF_OFFSETADDR);
		tbl_offset[i] = (buf[1] << 8) | buf[0];
	}
	tbl_offset[NUM_MONMES] = tbl_offset[NUM_MONMES - 1] + 0x0a;	/*  */

	/* X^[̃bZ[Wf[^ǂݍ */
	filepos = FILEPOS_MONMES;
	for (i = 0; i < NUM_MONMES; i++)
	{
		size_t bufsize;
		unsigned char *buf = NULL;

		if (tbl_offset[i + 1] - tbl_offset[i] < 0)
			break;

		bufsize = tbl_offset[i + 1] - tbl_offset[i];
		if ((buf = malloc(bufsize)) == NULL)
			err_printf(STR_FATAL_ERROR_IN_MALLOC);

		filepos += file_read(buf, filepos, bufsize);
		if (bufsize > 1)
			dec_string(tbl_monmes[i], buf, sizeof(tbl_monmes[i]), bufsize - 1, 1);

		free(buf);
		buf = NULL;
	}

	flag_setmonmes = 1;
}

mondata_t *mon_readdata(void)
{
	int i, j;
	mondata_t *buf_mondata = NULL;

	/* X^[̊ef[^i[obt@mۂ */
	if ((buf_mondata = malloc(sizeof(mondata_t) * NUM_MON)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* X^[̏ACef[^ǂݍ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONITEM];

		file_read(buf, FILEPOS_MONITEM + SIZEOF_MONITEM * i, SIZEOF_MONITEM);
		for (j = 0; j < SIZEOF_MONITEM; j++)
			buf_mondata[i].item_no[j] = buf[j];
	}

	/* X^[́uvsf[^ǂݍ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONCTRL];

		file_read(buf, FILEPOS_MONCTRL + SIZEOF_MONCTRL * i, SIZEOF_MONCTRL);
		for (j = 0; j < SIZEOF_MONCTRL; j++)
			buf_mondata[i].control_no[j] = buf[j];
	}

	/* X^[́u͂Ȃvsf[^ǂݍ */
	for (i = 0; i < NUM_MON; i++)
	{
		if (i >= 256)
			buf_mondata[i].free_no = 0xff;
		else
		{
			unsigned char buf[SIZEOF_MONFREE];

			file_read(buf, FILEPOS_MONFREE + SIZEOF_MONFREE * i, SIZEOF_MONFREE);
			buf_mondata[i].free_no = buf[0];
		}
	}

	/* X^[́uƂ킴vsf[^ǂݍ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONTECH];

		file_read(buf, FILEPOS_MONTECH + SIZEOF_MONTECH * i, SIZEOF_MONTECH);

		buf_mondata[i].flag_tech_attack_up    = ((buf[0] & 0x01) != 0x00);
		buf_mondata[i].flag_tech_defence_down = ((buf[0] & 0x80) != 0x00);
		buf_mondata[i].flag_tech_become_old   = ((buf[0] & 0x40) != 0x00);
		buf_mondata[i].flag_tech_poison       = ((buf[0] & 0x20) != 0x00);
		buf_mondata[i].flag_tech_darkness     = ((buf[0] & 0x10) != 0x00);
		buf_mondata[i].flag_tech_paralyze     = ((buf[0] & 0x08) != 0x00);
		buf_mondata[i].flag_tech_confuse      = ((buf[0] & 0x04) != 0x00);
		buf_mondata[i].flag_tech_slip         = ((buf[0] & 0x02) != 0x00);

		buf_mondata[i].tech_no = buf[1];
	}

	/* X^[̊ef[^ǂݍ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONDATA];

		file_read(buf, FILEPOS_MONDATA + SIZEOF_MONDATA * i, SIZEOF_MONDATA);

		/* lnf[^ */
		buf_mondata[i].level               = buf[31];
		buf_mondata[i].hit_point           = (buf[ 9] << 8) | buf[ 8];
		buf_mondata[i].magic_point         = (buf[11] << 8) | buf[10];
		buf_mondata[i].experience_point    = (buf[13] << 8) | buf[12];
		buf_mondata[i].gil                 = (buf[15] << 8) | buf[14];
		buf_mondata[i].speed               = buf[ 0];
		buf_mondata[i].attack_power        = buf[ 1];
		buf_mondata[i].defense_power       = buf[ 4];
		buf_mondata[i].evade_rate          = buf[ 3];
		buf_mondata[i].magic_attack_power  = buf[ 5];
		buf_mondata[i].magic_defense_power = buf[ 6];
		buf_mondata[i].magic_evade_rate    = buf[ 7];
		buf_mondata[i].weapon_no           = buf[16];
		buf_mondata[i].multiplier          = buf[ 2];
		buf_mondata[i].id                  = buf[30];

		/* nf[^ */
		buf_mondata[i].elem_none   = buf[17];
		buf_mondata[i].elem_absorb = buf[21];
		buf_mondata[i].elem_double = buf[23];

		/* ԕωnf[^ */
		buf_mondata[i].state_defense = bits_make32(buf[18], buf[19], buf[20], 0x00);
		buf_mondata[i].state_apriori = bits_make32(buf[26], buf[27], buf[28], buf[29]);
		buf_mondata[i].montype       = buf[24];

		/* tOnf[^ */
		buf_mondata[i].flag_capture = ((buf[25] & 0x08) == 0x00);
		buf_mondata[i].flag_control = ((buf[25] & 0x10) == 0x00);
	}

	return buf_mondata;
}

monact_t *mon_readact(void)
{
	int i, j;
	long filepos;
	unsigned char *buf = NULL;
	monact_t *buf_monact = NULL;
	int tbl_offset[NUM_MON + 1];

	/* X^[̍sp^[f[^i[obt@mۂ */
	if ((buf_monact = malloc(sizeof(monact_t) * NUM_MON)) == NULL)
		return NULL;
	for (i = 0; i < NUM_MON; i++)
	{
		for (j = 0; j < sizeof(buf_monact[i].ptn) / sizeof(buf_monact[i].ptn[0]); j++)
		{
			buf_monact[i].ptn[j].condition_size = 0;
			buf_monact[i].ptn[j].action_size = 0;
			buf_monact[i].ptn[j].condition = NULL;
			buf_monact[i].ptn[j].action = NULL;
		}
	}

	/* X^[̍sp^[f[^i[ĂAhX(ւ̃ItZbg)ǂݍ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_OFFSETADDR];

		file_read(buf, FILEPOS_MONACT_ADDR + SIZEOF_OFFSETADDR * i, SIZEOF_OFFSETADDR);
		tbl_offset[i] = (buf[1] << 8) | buf[0];
	}
	tbl_offset[NUM_MON] = tbl_offset[NUM_MON - 1] + 0x93;	/*  */

	/* X^[̍sp^[f[^ǂݍ */
	filepos = FILEPOS_MONACT;
	for (i = 0; i < NUM_MON; i++)
	{
		int ptn_num = 0;
		int flag_condition = 1;
		size_t bufsize;
		unsigned char *ptr_buf = NULL;
		unsigned char *ptr_mark = NULL;

		if (tbl_offset[i + 1] - tbl_offset[i] <= 0)
			break;
		bufsize = tbl_offset[i + 1] - tbl_offset[i];
		if ((buf = malloc(bufsize)) == NULL)
			err_printf(STR_FATAL_ERROR_IN_MALLOC);
		filepos += file_read(buf, filepos, bufsize);

		ptr_buf = ptr_mark = buf;
		while (ptr_buf - buf < (ptrdiff_t)bufsize)
		{
			if (*ptr_buf == 0xfe || *ptr_buf == 0xff)
			{
				const size_t len_read = ptr_buf - ptr_mark;

				if (ptr_buf - ptr_mark <= 1 && (*ptr_mark == 0xfe || *ptr_mark == 0xff))
					break;

				if (flag_condition)
				{
					if ((buf_monact[i].ptn[ptn_num].condition = malloc(len_read)) == NULL)
						err_printf(STR_FATAL_ERROR_IN_MALLOC);
					memmove(buf_monact[i].ptn[ptn_num].condition, ptr_mark, len_read);
					buf_monact[i].ptn[ptn_num].condition_size = len_read;
				}
				else
				{
					if ((buf_monact[i].ptn[ptn_num].action = malloc(len_read)) == NULL)
						err_printf(STR_FATAL_ERROR_IN_MALLOC);
					memmove(buf_monact[i].ptn[ptn_num].action, ptr_mark, len_read);
					buf_monact[i].ptn[ptn_num].action_size = len_read;
					ptn_num++;
				}

				flag_condition = !flag_condition;
				ptr_mark = ptr_buf + 1;
			}

			ptr_buf++;
		}

		free(buf);
		buf = NULL;
	}

	return buf_monact;
}

monparty_t *mon_readparty(void)
{
	int i, j;
	monparty_t *buf_monparty = NULL;

	/* X^[̃p[eBҐf[^i[obt@mۂ */
	if ((buf_monparty = malloc(sizeof(monparty_t) * NUM_MONPARTY)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* X^[̃p[eBҐ̊ef[^ǂݍ */
	for (i = 0; i < NUM_MONPARTY; i++)
	{
		unsigned char buf[SIZEOF_MONPARTY];

		file_read(buf, FILEPOS_MONPARTY + SIZEOF_MONPARTY * i, SIZEOF_MONPARTY);

		buf_monparty[i].ability_point = buf[2];
		buf_monparty[i].escape_rate   = buf[1] & 0x7f;
		buf_monparty[i].flag_escape   = ((buf[1] & 0x80) == 0x00);
		buf_monparty[i].bits_exist    = buf[3];

		for (j = 0; j < 8; j++)
		{
			buf_monparty[i].monster_no[j] = buf[j + 4];
			if ((buf[15] & 0x20) != 0x00 && buf_monparty[i].monster_no[j] + 0x100 < NUM_MON)
				buf_monparty[i].monster_no[j] += 0x100;
		}
	}

	return buf_monparty;
}

monenc_t *mon_readenc(void)
{
	int i, j;
	monenc_t *buf_monenc = NULL;

	/* X^[e[uf[^i[obt@mۂ */
	if ((buf_monenc = malloc(sizeof(monenc_t) * NUM_MONENC)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* X^[e[uf[^ǂݍ */
	for (i = 0; i < NUM_MONENC; i++)
	{
		unsigned char buf[SIZEOF_MONENC];

		file_read(buf, FILEPOS_MONENC + SIZEOF_MONENC * i, SIZEOF_MONENC);
		for (j = 0; j < 4; j++)
			buf_monenc[i].party_no[j] = (buf[j * 2 + 1] << 8) | buf[j * 2];
	}

	return buf_monenc;
}

monmap_t *mon_readmap(void)
{
	int i, j, x, y;
	long filepos;
	monmap_t *buf_monmap = NULL;

	/* X^[}bvf[^i[obt@mۂ */
	if ((buf_monmap = malloc(sizeof(monmap_t) * NUM_MONMAP)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* X^[}bvf[^ǂݍ */
	filepos = FILEPOS_MONMAP;
	for (i = 0; i < NUM_MONMAP; i++)
	{
		for (y = 0; y < NUM_MONMAPAREA; y++)
		{
			for (x = 0; x < NUM_MONMAPAREA; x++)
			{
				unsigned char buf[SIZEOF_MONMAP];

				filepos += file_read(buf, filepos, SIZEOF_MONMAP);
				for (j = 0; j < 4; j++)
					buf_monmap[i].area[x][y].enc_no[j] = (buf[j * 2 + 1] << 8) | buf[j * 2];
			}
		}
	}

	return buf_monmap;
}

int mon_writeparam(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̐lnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̐lnf[^o͂ */
	fprintf(fp, "No.       O                "
				" LV     HP     MP    EXP   M "
				"     U   h    U h  {   ID\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s :  %3d, %5ld, %5ld, %5ld, %5ld, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d\n", i, i,
				mon_getname(i),
				ptr_mondata[i].level,
				ptr_mondata[i].hit_point,
				ptr_mondata[i].magic_point,
				ptr_mondata[i].experience_point,
				ptr_mondata[i].gil,
				ptr_mondata[i].speed,
				ptr_mondata[i].attack_power,
				ptr_mondata[i].defense_power,
				ptr_mondata[i].evade_rate,
				ptr_mondata[i].magic_attack_power,
				ptr_mondata[i].magic_defense_power,
				ptr_mondata[i].magic_evade_rate,
				ptr_mondata[i].multiplier,
				ptr_mondata[i].id);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeelem(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̑nf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̑nf[^o͂ */
	fprintf(fp, "No.       O               "
				"          "
				"z          "
				"_\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s : ", i, i, mon_getname(i));

		/*  */
		str_printelem(fp, ptr_mondata[i].elem_none, ELEM_ALL);
		fprintf(fp, "  ");

		/* z */
		str_printelem(fp, ptr_mondata[i].elem_absorb, ELEM_ALL);
		fprintf(fp, "  ");

		/* _ */
		str_printelem(fp, ptr_mondata[i].elem_double, ELEM_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writestate(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̏ԕωnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̏ԕωnf[^o͂ */
	fprintf(fp, "No.       O               "
				"ԕωϐ1    2                3                 "
				"ԕω1    2                3                4\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s : ", i, i, mon_getname(i));

		/* ԕωւ̑ϐ */
		str_printstate(fp, ptr_mondata[i].state_defense, STATE_ALL & ~STATE_OTHER);
		fprintf(fp, "  ");

		/* ŏKpԕω */
		str_printstate(fp, ptr_mondata[i].state_apriori, STATE_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeflag(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̃tOnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̃tOnf[^o͂ */
	fprintf(fp, "No.       O               "
				"ߊl  푰\n");
	for (i = 0; i < NUM_MON; i++)
	{
		/* etO */
		fprintf(fp, "%03d(%03X): %-16s :  %2s   %2s", i, i, mon_getname(i),
				str_getflag(ptr_mondata[i].flag_capture),
				str_getflag(ptr_mondata[i].flag_control));

		/* X^[^Cv */
		if ((ptr_mondata[i].montype & MONTYPE_ALL) != 0x00)
		{
			fprintf(fp, "  ");
			str_printmontype(fp, ptr_mondata[i].montype, MONTYPE_ALL);
		}

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeitem(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_nothing = "------------------";

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̏ACef[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̏ACef[^o͂ */
	fprintf(fp, "No.       O               "
				"(m)        "
				"(m)        "
				"Ƃ              "
				"Ƃ(100%%)\n");
	for (i = 0; i < NUM_MON; i++)
	{
		const char *ptr_name[4];

		for (j = 0; j < 4; j++)
			ptr_name[j] = (ptr_mondata[i].item_no[j] != 0x00) ? item_getname(ptr_mondata[i].item_no[j]) : str_nothing;

		fprintf(fp, "%03d(%03X): %-16s : ", i, i, mon_getname(i));
		for (j = 0; j < 4; j++)
		{
			if (j != 0)
				fprintf(fp, ", ");
			fprintf(fp, "%-18s", ptr_name[j]);
		}
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writefree(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;
	const char *str_nothing = "----------------";

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[́u͂Ȃvsf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[́u͂Ȃvsf[^o͂ */
	fprintf(fp, "No.       O               ߊl  ͂Ȃ\n");
	for (i = 0; i < NUM_MON; i++)
	{
		const char *ptr_name = (ptr_mondata[i].free_no != 0xff) ? mag_getname(ptr_mondata[i].free_no) : str_nothing;

		fprintf(fp, "%03d(%03X): %-16s :  %2s   %-16s\n", i, i,
				mon_getname(i),
				str_getflag(ptr_mondata[i].flag_capture),
				ptr_name);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writectrl(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_nothing = "----------------";

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[́uvsf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[́uvsf[^o͂ */
	fprintf(fp, "No.       O               "
				"  "
				"1         "
				"2         "
				"3         "
				"4\n");
	for (i = 0; i < NUM_MON; i++)
	{
		const char *ptr_name[4];

		for (j = 0; j < 4; j++)
		{
			if (ptr_mondata[i].control_no[j] == 0x81)		/* 0x81 = "Ƃ킴" */
				ptr_name[j] = mon_gettechname(ptr_mondata[i].tech_no);
			else if (ptr_mondata[i].control_no[j] == 0xff)	/* 0xff = Ȃ */
				ptr_name[j] = str_nothing;
			else
				ptr_name[j] = mag_getname(ptr_mondata[i].control_no[j]);
		}

		fprintf(fp, "%03d(%03X): %-16s :  %2s   ", i, i,
				mon_getname(i),
				str_getflag(ptr_mondata[i].flag_control));
		for (j = 0; j < 4; j++)
		{
			if (j != 0)
				fprintf(fp, ", ");
			fprintf(fp, "%-16s", ptr_name[j]);
		}
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writetech(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[́uƂ킴vf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[́uƂ킴vf[^o͂ */
	fprintf(fp, "No.       "
				"O               "
				"Ƃ킴        "
				"U㏸ h䖳 V ғ È   Xbv\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s : %-16s,    %2s       %2s     %2s   %2s   %2s   %2s   %2s   %2s\n", i, i,
				mon_getname(i),
				mon_gettechname(ptr_mondata[i].tech_no),
				str_getflag(ptr_mondata[i].flag_tech_attack_up),
				str_getflag(ptr_mondata[i].flag_tech_defence_down),
				str_getflag(ptr_mondata[i].flag_tech_become_old),
				str_getflag(ptr_mondata[i].flag_tech_poison),
				str_getflag(ptr_mondata[i].flag_tech_darkness),
				str_getflag(ptr_mondata[i].flag_tech_paralyze),
				str_getflag(ptr_mondata[i].flag_tech_confuse),
				str_getflag(ptr_mondata[i].flag_tech_slip));
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writemisc(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̂̑̃f[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̂̑̃f[^o͂ */
	fprintf(fp, "No.       "
				"O               "
				"ʏU̕\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s : %-18s\n", i, i,
				mon_getname(i),
				item_getname((ptr_mondata[i].weapon_no != 0x00) ? ptr_mondata[i].weapon_no : 1));
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeact(const char *ptr_filename, const monact_t *ptr_monact, const mondata_t *ptr_mondata)
{
	int i, j, k;
	FILE *fp = NULL;
	const char *str_target[] =
	{
		"obc", "i", "Kt", "t@X",
		"N", "X^[1", "X^[2", "X^[3",
		"X^[4", "X^[5", "X^[6", "X^[7",
		"X^[8", "X^[g", "ȊÕX^[S", "X^[S",
		"ȊÕX^[̒N", "11h", "ÕLN^S", "̃LN^S",
		"14h", "15h", "16h", "17h",
		"LN^̒N", "jLN^̒N", "1Ah", "퓬s\̃LN^̒N",
		"tNԂ̃X^[S", "tNԂ̃X^[̒N", "1Eh", "1Fh",
		"20h", "21h", "tNԂ̃LN^̒N", "LN^S",
		"⏕@ĂLN^̒N", "퓬s\̃X^[̒N", "26h", "27h",
		"28h", "29h", "2Ah", "2Bh",
		"2Ch", "ŌɍsLN^/G̒N", "2Eh", "2Fh",
		"30h", "31h", "32h", "33h",
		"34h", "35h", "36h", "37h",
		"38h", "39h", "3Ah", "3Bh",
		"3Ch", "3Dh", "3Eh", "3Fh",
	};

	assert(sizeof(str_target) / sizeof(str_target[0]) == 0x40);

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monact == NULL || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̍sp^[f[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̍sp^[f[^o͂ */
	for (i = 0; i < NUM_MON; i++)
	{
		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[No.%03d(%03X): %s]\n", i, i, mon_getname(i));

		for (j = 0; j < 16 && ptr_monact[i].ptn[j].action_size != 0 && ptr_monact[i].ptn[j].action != NULL; j++)
		{
			int turn = 0;
			int turn_old = 0;
			int index = 0;
			int letter_num = 0;
			const size_t bufsize = ptr_monact[i].ptn[j].action_size + 256;
			unsigned char *ptr_condition = NULL;
			unsigned char *ptr_action = NULL;
			unsigned char *buf_action = NULL;
			int *buf_turn = NULL;

			/* ͗p̈ꎞobt@mۂ */
			if ((buf_action = malloc(bufsize)) == NULL)
				err_printf(STR_FATAL_ERROR_IN_MALLOC);
			if ((buf_turn = malloc(sizeof(int) * bufsize)) == NULL)
				err_printf(STR_FATAL_ERROR_IN_MALLOC);
			memset(buf_action, 0x00, bufsize);
			memmove(buf_action, ptr_monact[i].ptn[j].action, ptr_monact[i].ptn[j].action_size);
			for (k = 0; k < (int)bufsize; k++)
				buf_turn[k] = 0;

			/* Kp */
			fprintf(fp, ": ");
			ptr_condition = ptr_monact[i].ptn[j].condition;
			for (k = 0; k < (int)ptr_monact[i].ptn[j].condition_size / 4; k++)
			{
				if (k != 0)
					fprintf(fp, " & ");

				if (ptr_condition[0] == 0x00)
				{
					/* ʏ펞(00 00 00 00) */
					fprintf(fp, "[ʏ펞]");
				}
				else if (ptr_condition[0] == 0x01)
				{
					/* Ԃ̏Ԃƈv(01 xx yy zz) */
					/*   xx: Ώ */
					/*   yy: ԕωrbg̉oCgڂgp邩(0`3) */
					/*   zz: ԕωrbg */
					int index;
					char str_state[256] = "";
					char *ptr = str_state;
					state_t state = 0x00;

					state = bits_make32((ptr_condition[2] == 0x00) ? ptr_condition[3] : 0x00,
										(ptr_condition[2] == 0x01) ? ptr_condition[3] : 0x00,
										(ptr_condition[2] == 0x02) ? ptr_condition[3] : 0x00,
										(ptr_condition[2] == 0x03) ? ptr_condition[3] : 0x00);

					for (index = 0; index < NUM_STATE; index++)
						ptr += sprintf(ptr, "%s", str_getstate(state, index));

					if (ptr_condition[1] < 0x40)
						fprintf(fp, "[%s̏Ԃ \"%s\" ł]", str_target[ptr_condition[1]], str_state);
				}
				else if (ptr_condition[0] == 0x02)
				{
					/* qbg|Cglȉ(02 xx yy zz) */
					/*   xx: Ώ */
					/*   yy: qbg|Cg(ʃoCg) */
					/*   zz: qbg|Cg(ʃoCg) */
					if (ptr_condition[1] < 0x40)
						fprintf(fp, "[%sHP%ldȉł]", str_target[ptr_condition[1]], (long)((ptr_condition[3] << 8) | ptr_condition[2]));
				}
				else if (ptr_condition[0] == 0x03)
				{
					/* ϐ̒lƈv(03 00 yy zz) */
					/*   yy: ϐ0 */
					/*   zz: ϐ1 */
					fprintf(fp, "[ϐ \"%d, %d\" ł]", ptr_condition[2], ptr_condition[3]);
				}
				else if (ptr_condition[0] == 0x04)
				{
					/* ̃X^[݂Ȃ(04 00 yy 00) */
					/*   yy: X^[ */
					if (ptr_condition[2] == 0x00)
						fprintf(fp, "[ȊÕX^[Ȃ]");
					else
						fprintf(fp, "[X^[%dȂ]", ptr_condition[2]);
				}
				else if (ptr_condition[0] == 0x05)
				{
					/* ʒuɃX^[݂(05 00 yy zz) */
					/*   zz: ʒu(rbga) */
					int index;
					char str_no[256] = "";
					char *ptr = str_no;

					for (index = 0; index < 8; index++)
					{
						if ((ptr_condition[3] & (1U << (7 - index))) != 0x00)
							ptr += sprintf(ptr, "%s%d", (ptr > str_no) ? ", " : "", index + 1);
					}
					fprintf(fp, "[%sԖڂ̃X^[oĂ]", str_no);
				}
				else if (ptr_condition[0] == 0x06 || ptr_condition[0] == 0x07)
				{
					/* sɑ΂JE^[(06 xx yy zz) */
					/* sɑ΂JE^[(07 xx yy zz) */
					/*   xx: tO(0: ɓĂ͂܂ꍇ, 1: ɓĂ͂܂Ȃꍇ) */
					/*   yy: s */
					/*   zz: rbg */
					fprintf(fp, "[");

					switch (ptr_condition[2])
					{
					case 0x04:
						fprintf(fp, "U");
						break;
					case 0x07:
						fprintf(fp, "Ȃ炩̍s");
						break;
					case 0x2b:
						fprintf(fp, "@U");
						break;
					case 0x0a:
					case 0x0d:
					default:
						fprintf(fp, "%02Xh", ptr_condition[2]);
					}

					if (ptr_condition[3] != 0x00)
					{
						int index;
						char str_element[256] = "";
						char *ptr = str_element;
						const elem_t element = ptr_condition[3];

						for (index = 0; index < NUM_ELEMENT; index++)
							ptr += sprintf(ptr, "%s", str_getelem(element, index));

						fprintf(fp, "(%s)", str_element);
					}

					fprintf(fp, "%s󂯂]", (ptr_condition[1] != 0x00) ? "ȊO" : "");
				}
				else if (ptr_condition[0] == 0x08)
				{
					/* ̖@ɑ΂JE^[(08 00 yy 00) */
					/*   yy: @̔ԍ */
					fprintf(fp, "[@ \"%s\" 󂯂]", mag_getname(ptr_condition[2]));
				}
				else if (ptr_condition[0] == 0x09)
				{
					/* ACeJE^[(09 00 yy 00) */
					/*   yy: ACe̔ԍ */
					fprintf(fp, "[ACe \"%s\" 󂯂]", item_getname(ptr_condition[2]));
				}
				else if (ptr_condition[0] == 0x0e)
				{
					/* _[Wɑ΂JE^[(0E 00 00 00) */
					fprintf(fp, "[_[W󂯂]");
				}
				else if (ptr_condition[0] == 0x0f)
				{
					/* t@CiA^bN(0F 00 00 00) */
					fprintf(fp, "[t@CiA^bN]");
				}
				else if (ptr_condition[0] == 0x11)
				{
					/* @΂JE^[(11 00 00 00) */
					fprintf(fp, "[@󂯂]");
				}
				else
				{
					/* ̑ */
					fprintf(fp, "[%02X %02X %02X %02X]", ptr_condition[0], ptr_condition[1], ptr_condition[2], ptr_condition[3]);
				}

				ptr_condition += 4;
			}

			fputc('\n', fp);

			/* ^[ */
			turn = 1;
			ptr_action = buf_action;
			while ((index = ptr_action - buf_action) < (ptrdiff_t)ptr_monact[i].ptn[j].action_size)
			{
				if (*ptr_action != 0xfd)
				{
					buf_turn[index] = turn;
					ptr_action++;
					turn++;
				}
				else
				{
					buf_turn[index] = turn;
					ptr_action++;

					if (*ptr_action < 0xf0)
					{
						for (k = 0; k < 3; k++)
						{
							buf_turn[ptr_action - buf_action] = turn;
							ptr_action++;
						}
						turn++;
					}
					else if (*ptr_action == 0xf7)
					{
						int skip_byte = *(ptr_action + 1);

						for (k = 0; k < skip_byte + 3; k++)
						{
							buf_turn[ptr_action - buf_action] = turn;
							ptr_action++;
						}
						turn++;
					}
					else
					{
						for (k = 0; k < 3; k++)
						{
							buf_turn[ptr_action - buf_action] = turn;
							ptr_action++;
						}
					}
				}
			}

			/* sp^[ */
			ptr_action = buf_action;
			while ((index = ptr_action - buf_action) < (ptrdiff_t)ptr_monact[i].ptn[j].action_size)
			{
				/* ^[̕\ */
				turn = buf_turn[index];
				if (turn != turn_old)
					letter_num = fprintf(fp, "%d^[: ", buf_turn[index]);
				else
					fprintf(fp, "%*s", letter_num, "");
				turn_old = turn;

				/* sp^[f[^̉ */
				if (*ptr_action != 0xfd)
				{
					mon_printact(fp, *ptr_action, ptr_mondata[i]);
					ptr_action++;
				}
				else
				{
					ptr_action++;

					if (*ptr_action < 0xf0)				/* _s */
					{
						for (k = 0; k < 3; k++)
						{
							if (k != 0)
								fprintf(fp, ", ");
							mon_printact(fp, *ptr_action, ptr_mondata[i]);
							ptr_action++;
						}
					}
					else
					{
						unsigned char triplet[3];

						for (k = 0; k < 3; k++)
							triplet[k] = *ptr_action++;

						if (triplet[0] == 0xf2)
						{
							/* oԐݒ(f2 xx yy) */
							fprintf(fp, "[oԐݒ: %02Xh, %02Xh]", triplet[1], triplet[2]);
						}
						else if (triplet[0] == 0xf3)
						{
							/* UڕWݒ(f3 xx yy) */
							/*   xx: Ώ */
							fprintf(fp, "[UڕWݒ: %s]", str_target[triplet[1]]);
						}
						else if (triplet[0] == 0xf4)
						{
							/* ϐݒ(f4 xx yy) */
							/*   yy: ϐ0 */
							/*   zz: ϐ1 */
							fprintf(fp, "[ϐݒ: %d, %d]", triplet[1], triplet[2]);
						}
						else if (triplet[0] == 0xf6)
						{
							/* bZ[W\(f6 xx yy) */
							/*   xx: bZ[W̔ԍ */
							fprintf(fp, "[bZ[W\: ");
							fputs(mon_getmes(triplet[2]), fp);
							fprintf(fp, "]");
						}
						else if (triplet[0] == 0xf7)
						{
							/* 1^[ŕ̍s(f7 xx yy) */
							fprintf(fp, "[s]");
						}
						else if (triplet[0] == 0xf8)
						{
							/* o(f8 xx yy) */
							/*   xx: s */
							/*   yy: s */
							fprintf(fp, "[o: %02Xh, %02Xh]", triplet[1], triplet[2]);
						}
						else if (triplet[0] == 0xf9)
						{
							/* CxgtOݒ(f9 xx yy) */
							fprintf(fp, "[CxgtOݒ: %02Xh, %02Xh]", triplet[1], triplet[2]);
						}
						else
						{
							fprintf(fp, "[%02X %02X %02X]", triplet[0], triplet[1], triplet[2]);
						}
					}
				}

				fputc('\n', fp);
			}

			free(buf_action);
			buf_action = NULL;
			free(buf_turn);
			buf_turn = NULL;
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writemes(const char *ptr_filename)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0')
	{
		assert(0);
		return 0;
	}

	/* X^[̃bZ[Wf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̃bZ[Wf[^o͂ */
	fprintf(fp, "No.      bZ[W\n");
	for (i = 0; i < NUM_MONMES; i++)
	{
		fprintf(fp, "%03d(%02X): ", i, i);
		fputs(mon_getmes(i), fp);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeparty(const char *ptr_filename, const monparty_t *ptr_monparty)
{
	int i, j;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monparty == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̃p[eBҐf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̃p[eBҐf[^o͂ */
	fprintf(fp, "No.       "
				" ABP   "
				"  "
				"o       "
				"X^[1         "
				"X^[2         "
				"X^[3         "
				"X^[4         "
				"X^[5         "
				"X^[6         "
				"X^[7         "
				"X^[8\n");
	for (i = 0; i < NUM_MONPARTY; i++)
	{
		fprintf(fp, "%03d(%03X):  %3d, %3d,  %2s   ", i, i,
				ptr_monparty[i].ability_point,
				ptr_monparty[i].escape_rate,
				str_getflag(ptr_monparty[i].flag_escape));
		for (j = 0; j < 8; j++)
		{
			fprintf(fp, "%-2s",
					str_getflag((ptr_monparty[i].bits_exist & (1U << (7 - j))) != 0x00));
		}
		fprintf(fp, "  ");
		for (j = 0; j < 8; j++)
		{
			int flag_absent = ((ptr_monparty[i].bits_exist & (1U << (7 - j))) == 0x00);

			if (ptr_monparty[i].monster_no[j] == 0xff)
				break;

			if (j != 0)
				fprintf(fp, ", ");
			fprintf(fp, "%s%-16s%s",
					(flag_absent) ? "(" : " ",
					mon_getname(ptr_monparty[i].monster_no[j]),
					(flag_absent) ? ")" : " ");
		}
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeenc(const char *ptr_filename, const monenc_t *ptr_monenc, const monparty_t *ptr_monparty)
{
	int i, j, k;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monenc == NULL || ptr_monparty == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[e[uf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[e[uo͂ */
	for (i = 0; i < NUM_MONENC; i++)
	{
		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[No.%03d(%02X)]\n", i, i);
		for (j = 0; j < 4; j++)
		{
			const monparty_t tmp_party = ptr_monparty[ptr_monenc[i].party_no[j]];

			fprintf(fp, "%02d: %03d(%03X): ", j + 1, ptr_monenc[i].party_no[j], ptr_monenc[i].party_no[j]);
			for (k = 0; k < 8; k++)
			{
				const int flag_absent = ((tmp_party.bits_exist & (1U << (7 - k))) == 0x00);

				if (tmp_party.monster_no[k] == 0xff)
					break;

				if (k != 0)
					fprintf(fp, ", ");
				fprintf(fp, "%s%-16s%s",
						(flag_absent) ? "(" : " ",
						mon_getname(tmp_party.monster_no[k]),
						(flag_absent) ? ")" : " ");
			}
			fputc('\n', fp);
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writemap(const char *ptr_filename, const monmap_t *ptr_monmap)
{
	int i, j, x, y;
	FILE *fp = NULL;
	const char *str_name[NUM_MONMAP][4] =
	{
		{"1E()",	"1E(X)",		"1E()",	"1E(C)"},
		{"2E()",	"2E(X)",		"2E()",		"2E(C)"},
		{"3E()",	"3E(X)",		"3E()",	"3E(C)"},
	};

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monmap == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[}bvf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[}bvo͂ */
	for (i = 0; i < NUM_MONMAP; i++)
	{
		for (j = 0; j < 4; j++)
		{
			if (i != 0 || j != 0)
				fputc('\n', fp);

			fprintf(fp, "[No.%1d-%1d: %s]\n", i, j, str_name[i][j]);
			for (y = 0; y < NUM_MONMAPAREA; y++)
			{
				for (x = 0; x < NUM_MONMAPAREA; x++)
				{
					if (x != 0)
						fprintf(fp, " ");

					fprintf(fp, "%03d", ptr_monmap[i].area[x][y].enc_no[j]);
				}
				fputc('\n', fp);
			}
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

static void mon_printact(FILE *stream, int index, mondata_t mondata)
{
	/* `FbN */
	if (stream == NULL || index < 0 || 0xff <= index)
	{
		assert(0);
		return;
	}

	switch (index)
	{
	case 0x81:		/* Ƃ킴 */
		fprintf(stream, "%s", mon_gettechname(mondata.tech_no));
		break;
	case 0xaa:		/*  */
		fprintf(stream, "(Ȃ)");
		break;
	case 0xb1:		/* ܂ */
		fprintf(stream, "()");
		break;
	case 0xde:		/* 䂤΂ */
		fprintf(stream, "(U");
		if (mondata.weapon_no != 0)
			fprintf(stream, ": %s", item_getname(mondata.weapon_no));
		fprintf(stream, ")");
		break;
	case 0xec:
		fprintf(stream, "(񂹂)");
		break;
	case 0xee:
		fprintf(stream, "(X^[oԕύX)");
		break;
	case 0xed:		/* s */
	default:
		fprintf(stream, "%s", mag_getname(index));
		break;
	}
}

void mon_freeact(monact_t *ptr_monact)
{
	int i, j;

	/* `FbN */
	if (ptr_monact == NULL)
	{
		assert(0);
		return;
	}

	for (i = 0; i < NUM_MON; i++)
	{
		for (j = 0; j < sizeof(ptr_monact[i].ptn) / sizeof(ptr_monact[i].ptn[0]); j++)
		{
			if (ptr_monact[i].ptn[j].condition != NULL)
			{
				free(ptr_monact[i].ptn[j].condition);
				ptr_monact[i].ptn[j].condition = NULL;
			}
			if (ptr_monact[i].ptn[j].action != NULL)
			{
				free(ptr_monact[i].ptn[j].action);
				ptr_monact[i].ptn[j].action = NULL;
			}
		}
	}
}
