/*
 *	ff5special.c
 *		Copyright (C)2002 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "ff5file.h"
#include "ff5decode.h"
#include "ff5special.h"

#define FILEPOS_SPNAME		0x116700
#define FILEPOS_SPDATA		0x116ab1
#define NUM_SP_ANIMAL		9
#define NUM_SP_PHARMACY		64
#define NUM_SP_LANDSCAPE	32
#define NUM_SP				(NUM_SP_ANIMAL + NUM_SP_PHARMACY + NUM_SP_LANDSCAPE)	/* = 105 */
#define SIZEOF_SPNAME		9
#define SIZEOF_SPDATA		8

static void sp_setname(void);

static char tbl_spname[NUM_SP][256];	/* U̖Oi[e[u */
static int flag_setspname = 0;			/* tbl_spnameɃf[^i[Ă邩ǂ */

const char *sp_getname(int index)
{
	/* `FbN */
	if (index < 0 || NUM_SP <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setspname)
		sp_setname();

	return tbl_spname[index];
}

static void sp_setname(void)
{
	int i;

	/* U̖Of[^ */
	for (i = 0; i < NUM_SP; i++)
		*tbl_spname[i] = '\0';

	/* U̖Of[^ǂݍ */
	for (i = 0; i < NUM_SP; i++)
	{
		unsigned char buf[SIZEOF_SPNAME];

		file_read(buf, FILEPOS_SPNAME + SIZEOF_SPNAME * i, SIZEOF_SPNAME);
		dec_string(tbl_spname[i], buf, sizeof(tbl_spname[i]), SIZEOF_SPNAME, 0);
	}

	flag_setspname = 1;
}

spdata_t *sp_readdata(void)
{
	int i;
	spdata_t *buf_spdata = NULL;

	/* Ůef[^i[obt@mۂ */
	if ((buf_spdata = malloc(sizeof(spdata_t) * NUM_SP)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* Ůef[^ǂݍ */
	for (i = 0; i < NUM_SP; i++)
	{
		unsigned char buf[SIZEOF_SPDATA];

		file_read(buf, FILEPOS_SPDATA + SIZEOF_SPDATA * i, SIZEOF_SPDATA);
		mag_setparam(&buf_spdata[i], buf, sizeof(buf));
	}

	return buf_spdata;
}

int sp_writeparam(const char *ptr_filename, const spdata_t *ptr_spdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_spdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* U̐lnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* U̐lnf[^o͂ */
	fprintf(fp, "No.      O                 "
				"ʒl   䗦     Œl\n");
	for (i = 0; i < NUM_SP; i++)
	{
		fprintf(fp, (ptr_spdata[i].effect_rate != 0) ?
					"%03d(%02X): %-18s :  %5d,%3d/16, %5d, %5d, %5d\n" :
					"%03d(%02X): %-18s :  %5d, %5d, %5d, %5d, %5d\n",
				i, i, sp_getname(i),
				ptr_spdata[i].effect_power,
				ptr_spdata[i].effect_rate,
				ptr_spdata[i].hit_rate,
				ptr_spdata[i].effect_time,
				ptr_spdata[i].fixed_value);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int sp_writeelem(const char *ptr_filename, const spdata_t *ptr_spdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_spdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* Ȗnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* Ȗnf[^o͂ */
	fprintf(fp, "No.      O                 "
				"\n");
	for (i = 0; i < NUM_SP; i++)
	{
		fprintf(fp, "%03d(%02X): %-18s : ", i, i, sp_getname(i));

		/*  */
		str_printelem(fp, ptr_spdata[i].elem_effect, ELEM_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int sp_writestate(const char *ptr_filename, const spdata_t *ptr_spdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_spdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* Ȕԕωnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* Ȕԕωnf[^o͂ */
	fprintf(fp, "No.      O                 "
				"ԕω1        2                3                4\n");
	for (i = 0; i < NUM_SP; i++)
	{
		fprintf(fp, "%03d(%02X): %-18s : ", i, i, sp_getname(i));

		/* ԕω */
		str_printstate(fp, ptr_spdata[i].state_effect, STATE_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int sp_writeflag(const char *ptr_filename, const spdata_t *ptr_spdata)
{
	int i;
	FILE *fp = NULL;
	const char *str_direction[2][2] =
	{
		{"",	"/G"},
		{"G",		"G/"},
	};
	const char *str_range[2][2] =
	{
		{"P",	"P/S"},
		{"S",	"S/P"},
	};
	const char *str_roulette = "[bg";

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_spdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ŨtOnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ŨtOnf[^o͂ */
	fprintf(fp, "No.      O                 "
				"  "
				"ʔ͈           "
				"L푰\n");
	for (i = 0; i < NUM_SP; i++)
	{
		/* etO */
		fprintf(fp, "%03d(%02X): %-18s :  %2s   ", i, i, sp_getname(i),
				str_getflag(ptr_spdata[i].flag_reflect));

		/* ʔ͈ */
		if (ptr_spdata[i].flag_roulette)
			fprintf(fp, "%-17s", str_roulette);
		else
		{
			fprintf(fp, "%-7s %-9s",
					str_direction[ptr_spdata[i].direction][ptr_spdata[i].change_direction],
					str_range[ptr_spdata[i].range][ptr_spdata[i].change_range]);
		}

		/* LX^[^Cv */
		if (ptr_spdata[i].montype_effect != 0x00)
		{
			fprintf(fp, "  ");
			str_printmontype(fp, ptr_spdata[i].montype_effect, MONTYPE_ALL);
		}

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}
