/*
 *	ff6analyzer.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "errmes.h"
#include "utility.h"
#include "ff6file.h"
#include "ff6monster.h"
#include "ff6character.h"
#include "ff6item.h"
#include "ff6magic.h"

#define FILENAME_MON_PARAM		"ff6monparam.txt"
#define FILENAME_MON_ELEM		"ff6monelem.txt"
#define FILENAME_MON_STATE		"ff6monstate.txt"
#define FILENAME_MON_FLAG		"ff6monflag.txt"
#define FILENAME_MON_ITEM		"ff6monitem.txt"
#define FILENAME_MON_RAGNA		"ff6monragna.txt"
#define FILENAME_MON_CTRL		"ff6monctrl.txt"
#define FILENAME_MON_SKETCH		"ff6monsketch.txt"
#define FILENAME_MON_RAGE		"ff6monrage.txt"
#define FILENAME_MON_MISC		"ff6monmisc.txt"
#define FILENAME_MON_ACT		"ff6monact.txt"
#define FILENAME_MON_MES		"ff6monmes.txt"
#define FILENAME_MON_EVENT		"ff6monevent.txt"
#define FILENAME_MON_PARTY		"ff6monparty.txt"
#define FILENAME_MON_ENC		"ff6monenc.txt"
#define FILENAME_MON_MAP		"ff6monmap.txt"
#define FILENAME_CHR_PARAM		"ff6chrparam.txt"
#define FILENAME_CHR_EQUIP		"ff6chrequip.txt"
#define FILENAME_CHR_LVUP		"ff6chrlvup.txt"
#define FILENAME_ITEM_PARAM		"ff6itemparam.txt"
#define FILENAME_ITEM_ELEM		"ff6itemelem.txt"
#define FILENAME_ITEM_STATE		"ff6itemstate.txt"
#define FILENAME_ITEM_FLAG		"ff6itemflag.txt"
#define FILENAME_ITEM_EFFECT	"ff6itemeffect.txt"
#define FILENAME_ITEM_COL		"ff6itemcol.txt"
#define FILENAME_MAG_PARAM		"ff6magparam.txt"
#define FILENAME_MAG_ELEM		"ff6magelem.txt"
#define FILENAME_MAG_STATE		"ff6magstate.txt"
#define FILENAME_MAG_FLAG		"ff6magflag.txt"

#define STR_TOO_FEW_ARGUMENTS		"Too few arguments"
#define STR_INVALID_OPTION			"Invalid option"
#define STR_INVALID_ROMIMAGE_FILE	"Invalid ROM-Image file"
#define STR_INVALID_CDROM_FILES		"Invalid CD-ROM files"
#define STR_CANNOT_CREATE_FILE		"Cannot create file"

/* 쌠\L */
static const char *str_copyright = "ff6analyzer - Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>";

int main(int argc, char *argv[])
{
	int i;
	int ret = EXIT_SUCCESS;
	int mode = FILEMODE_SFC;
	int flag_invalid = 0;
	int flag_usage   = 0;
	int index_srcpath = -1;
	int index_dstpath = -1;
	char str_srcpath[1024] = "";
	char str_dstpath[1024] = "";
	char str_dstfilename[1024] = "";
	mondata_t  *buf_mondata  = NULL;
	monact_t   *buf_monact   = NULL;
	monparty_t *buf_monparty = NULL;
	monenc_t   *buf_monenc   = NULL;
	monmap_t   *buf_monmap   = NULL;
	chrdata_t  *buf_chrdata  = NULL;
	chrlvup_t  *buf_chrlvup  = NULL;
	itemdata_t *buf_itemdata = NULL;
	magdata_t  *buf_magdata  = NULL;

	errno = 0;
	err_setprogname(argv[0]);

	/* R}hC͂ */
	for (i = 1; i < argc; i++)
	{
		if (argv[i] == NULL)
			continue;

		if (*argv[i] == '-')
		{
			const char *ptr = argv[i];

			while (*++ptr != '\0')
			{
				switch (tolower(*ptr))
				{
				case 'f':	/* X[p[t@~RŃ[h */
					mode = FILEMODE_SFC;
					break;
				case 'p':	/* vCXe[VŃ[h */
					mode = FILEMODE_PS;
					break;
				case 'h':	/* gp@̕\ */
					flag_usage = 1;
					break;
				default:
					flag_invalid = 1;
					break;
				}
			}
		}
		else
		{
			if (index_srcpath < 0)
				index_srcpath = i;
			else if (index_dstpath < 0)
				index_dstpath = i;
		}
	}

	/* /o̓pXݒ肷 */
	if (index_srcpath >= 0)
	{
		if (mode == FILEMODE_PS)
			transpath(str_srcpath, argv[index_srcpath], sizeof(str_srcpath));
		else
			transfilename(str_srcpath, argv[index_srcpath], sizeof(str_srcpath));
	}
	if (index_dstpath >= 0)
		transpath(str_dstpath, argv[index_dstpath], sizeof(str_dstpath));

	/* gp@\ */
	if (flag_usage)
	{
		puts(str_copyright);
		printf("%s: [option] [src-pathname] [dst-pathname]\n", err_getprogname());
		printf("\t-f\tanalyze in \"SUPER FAMICOM Ver.\" mode\n");
		printf("\t-p\tanalyze in \"PlayStation Ver.\" mode\n");
		printf("\t-h\tdisplay this help and exit\n");
		printf("\n");

		exit(EXIT_SUCCESS);
	}

	/* R}hC`FbN */
	if (flag_invalid)
		err_printf(STR_INVALID_OPTION);
	if (mode == FILEMODE_SFC && *str_srcpath == '\0')
		err_printf(STR_TOO_FEW_ARGUMENTS);

	/* t@CI[v */
	if (!file_open(str_srcpath, mode))
	{
		if (mode == FILEMODE_PS)
			err_printf(STR_INVALID_CDROM_FILES " - %s*", str_srcpath);
		else
			err_printf(STR_INVALID_ROMIMAGE_FILE " - %s", str_srcpath);
	}

	/* ef[^́E͂ */
	buf_mondata  = mon_readdata();
	buf_monact   = mon_readact();
	buf_monparty = mon_readparty();
	buf_monenc   = mon_readenc();
	buf_monmap   = mon_readmap();
	buf_chrdata  = chr_readdata();
	buf_chrlvup  = chr_readlvup();
	buf_itemdata = item_readdata();
	buf_magdata  = mag_readdata();

	/* ef[^o͂ */
	if (buf_mondata  == NULL ||
		buf_monact   == NULL ||
		buf_monparty == NULL ||
		buf_monenc   == NULL ||
		buf_monmap   == NULL ||
		buf_chrdata  == NULL ||
		buf_chrlvup  == NULL ||
		buf_itemdata == NULL ||
		buf_magdata  == NULL)
	{
		if (mode == FILEMODE_PS)
			wrn_printf(STR_INVALID_CDROM_FILES " - %s*", str_srcpath);
		else
			wrn_printf(STR_INVALID_ROMIMAGE_FILE " - %s", str_srcpath);

		ret = EXIT_FAILURE;
	}
	else
	{
		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_PARAM);
		if (!mon_writeparam(str_dstfilename, buf_mondata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_ELEM);
		if (!mon_writeelem(str_dstfilename, buf_mondata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_STATE);
		if (!mon_writestate(str_dstfilename, buf_mondata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_FLAG);
		if (!mon_writeflag(str_dstfilename, buf_mondata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_ITEM);
		if (!mon_writeitem(str_dstfilename, buf_mondata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_RAGNA);
		if (!mon_writeragna(str_dstfilename, buf_mondata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_CTRL);
		if (!mon_writectrl(str_dstfilename, buf_mondata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_SKETCH);
		if (!mon_writesketch(str_dstfilename, buf_mondata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_RAGE);
		if (!mon_writerage(str_dstfilename, buf_mondata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_MISC);
		if (!mon_writemisc(str_dstfilename, buf_mondata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_ACT);
		if (!mon_writeact(str_dstfilename, buf_monact, buf_mondata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_MES);
		if (!mon_writemes(str_dstfilename))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_EVENT);
		if (!mon_writeevent(str_dstfilename))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_PARTY);
		if (!mon_writeparty(str_dstfilename, buf_monparty))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_ENC);
		if (!mon_writeenc(str_dstfilename, buf_monenc, buf_monparty))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MON_MAP);
		if (!mon_writemap(str_dstfilename, buf_monmap))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_CHR_PARAM);
		if (!chr_writeparam(str_dstfilename, buf_chrdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_CHR_EQUIP);
		if (!chr_writeequip(str_dstfilename, buf_chrdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_CHR_LVUP);
		if (!chr_writelvup(str_dstfilename, buf_chrlvup))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_ITEM_PARAM);
		if (!item_writeparam(str_dstfilename, buf_itemdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_ITEM_ELEM);
		if (!item_writeelem(str_dstfilename, buf_itemdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_ITEM_STATE);
		if (!item_writestate(str_dstfilename, buf_itemdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_ITEM_FLAG);
		if (!item_writeflag(str_dstfilename, buf_itemdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_ITEM_EFFECT);
		if (!item_writeeffect(str_dstfilename, buf_itemdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_ITEM_COL);
		if (!item_writecol(str_dstfilename, buf_itemdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MAG_PARAM);
		if (!mag_writeparam(str_dstfilename, buf_magdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MAG_ELEM);
		if (!mag_writeelem(str_dstfilename, buf_magdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MAG_STATE);
		if (!mag_writestate(str_dstfilename, buf_magdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}

		sprintf(str_dstfilename, "%s%s", str_dstpath, FILENAME_MAG_FLAG);
		if (!mag_writeflag(str_dstfilename, buf_magdata))
		{
			wrn_printf(STR_CANNOT_CREATE_FILE " - %s:", str_dstfilename);
			ret = EXIT_FAILURE;
		}
	}

	/* t@CN[Y */
	file_close();

	/* obt@ */
	mon_freeact(buf_monact);
	free(buf_mondata);
	free(buf_monact);
	free(buf_monparty);
	free(buf_monenc);
	free(buf_monmap);
	free(buf_chrdata);
	free(buf_chrlvup);
	free(buf_itemdata);
	free(buf_magdata);
	buf_mondata  = NULL;
	buf_monact   = NULL;
	buf_monparty = NULL;
	buf_monenc   = NULL;
	buf_monmap   = NULL;
	buf_chrdata  = NULL;
	buf_chrlvup  = NULL;
	buf_itemdata = NULL;
	buf_magdata  = NULL;

	return ret;
}
