/*
 *	ff6character.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "ff6file.h"
#include "ff6decode.h"
#include "ff6item.h"
#include "ff6character.h"

#define FILEPOS_CHREXP			(0x039ff4 + 0x29e22c)	/* = 0x2d8220 */
#define FILEPOS_CHRHP			(0x039844 + 0x235c5c)	/* = 0x26f4a0 */
#define FILEPOS_CHRMP			(0x0398a6 + 0x235c5c)	/* = 0x26f502 */
#define FILEPOS_CHRNAME			(0x02bc28 + 0x01bc98)	/* = 0x0478c0 */
#define FILEPOS_CHRJOBNAME		(0x003b40 + 0x0f0000)	/* = 0x0f3b40 */
#define FILEPOS_CHRDATA			(0x039a74 + 0x29e22c)	/* = 0x2d7ca0 */
#define FILEPOS_CHRCMDNAME		(0x02ce48 + 0x2a14b8)	/* = 0x2ce300 */
#define NUM_CHR					64
#define NUM_CHRCMDNAME			32
#define SIZEOF_CHREXP			2
#define SIZEOF_CHRHP			1
#define SIZEOF_CHRMP			1
#define SIZEOF_CHRNAME			6
#define SIZEOF_CHRJOBNAME		7
#define SIZEOF_CHRDATA			22
#define SIZEOF_CHRCMDNAME		6

static void chr_setname(void);
static void chr_setjobname(void);
static void chr_setcmdname(void);

static char tbl_chrname[NUM_CHR][256];			/* LN^̖Oi[e[u */
static char tbl_chrjobname[NUM_CHR][256];		/* Wu̖Oi[e[u */
static char tbl_cmdname[NUM_CHRCMDNAME][256];	/* 퓬R}h̖Oi[e[u */
static int flag_setchrname    = 0;				/* tbl_chrnameɃf[^i[Ă邩ǂ */
static int flag_setchrjobname = 0;				/* tbl_chrjobnameɃf[^i[Ă邩ǂ */
static int flag_setcmdname    = 0;				/* tbl_cmdnameɃf[^i[Ă邩ǂ */

const char *chr_getname(int index)
{
	/* `FbN */
	if (index < 0 || NUM_CHR <= index)
	{
		assert(0);
		return "";
	}

	if (!flag_setchrname)
		chr_setname();

	return tbl_chrname[index];
}

const char *chr_getjobname(int index)
{
	/* `FbN */
	if (index < 0 || NUM_CHR <= index)
	{
		assert(0);
		return "";
	}

	if (!flag_setchrjobname)
		chr_setjobname();

	return tbl_chrjobname[index];
}

const char *chr_getcmdname(int index)
{
	/* `FbN */
	if (index < 0 || NUM_CHRCMDNAME <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setcmdname)
		chr_setcmdname();

	return tbl_cmdname[index];
}

static void chr_setname(void)
{
	int i;

	/* LN^̖Of[^ */
	for (i = 0; i < NUM_CHR; i++)
		*tbl_chrname[i] = '\0';

	/* LN^̖Of[^ǂݍ */
	for (i = 0; i < NUM_CHR; i++)
	{
		unsigned char buf[SIZEOF_CHRNAME];

		file_read(buf, FILEPOS_CHRNAME + SIZEOF_CHRNAME * i, SIZEOF_CHRNAME);
		dec_string(tbl_chrname[i], buf, sizeof(tbl_chrname[i]), SIZEOF_CHRNAME, 0);
	}

	flag_setchrname = 1;
}

static void chr_setjobname(void)
{
	int i;

	/* LN^̃Wuf[^ǂݍ */
	for (i = 0; i < NUM_CHR; i++)
	{
		unsigned char buf[SIZEOF_CHRJOBNAME];

		file_read(buf, FILEPOS_CHRJOBNAME + SIZEOF_CHRJOBNAME * i, SIZEOF_CHRJOBNAME);
		dec_string(tbl_chrjobname[i], buf, sizeof(tbl_chrjobname[i]), SIZEOF_CHRJOBNAME, 0);
	}

	flag_setchrjobname = 1;
}

static void chr_setcmdname(void)
{
	int i;

	/* 퓬R}h̃f[^ǂݍ */
	for (i = 0; i < NUM_CHRCMDNAME; i++)
	{
		unsigned char buf[SIZEOF_CHRCMDNAME];

		file_read(buf, FILEPOS_CHRCMDNAME + SIZEOF_CHRCMDNAME * i, SIZEOF_CHRCMDNAME);
		dec_string(tbl_cmdname[i], buf, sizeof(tbl_cmdname[i]), SIZEOF_CHRCMDNAME, 0);
	}

	flag_setcmdname = 1;
}

chrdata_t *chr_readdata(void)
{
	int i, j;
	chrdata_t *buf_chrdata = NULL;
	chrlvup_t tbl_lvup[NUM_CHRLV];
	const int tbl_level_cng[4] = {0, 2, 5, -3};

	/* LN^f[^i[obt@mۂ */
	if ((buf_chrdata = malloc(sizeof(chrdata_t) * NUM_CHR)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* xƂ̊{qbg|Cgǂݍ */
	tbl_lvup[0].hit_point = 0;
	for (i = 1; i < NUM_CHRLV; i++)
	{
		unsigned char buf[SIZEOF_CHRHP];

		file_read(buf, FILEPOS_CHRHP + SIZEOF_CHRHP * (i - 1), SIZEOF_CHRHP);
		tbl_lvup[i].hit_point = tbl_lvup[i - 1].hit_point + buf[0];
	}

	/* xƂ̊{}WbN|Cgǂݍ */
	tbl_lvup[0].magic_point = 0;
	for (i = 1; i < NUM_CHRLV; i++)
	{
		unsigned char buf[SIZEOF_CHRMP];

		file_read(buf, FILEPOS_CHRMP + SIZEOF_CHRMP * (i - 1), SIZEOF_CHRMP);
		tbl_lvup[i].magic_point = tbl_lvup[i - 1].magic_point + buf[0];
	}

	/* LN^f[^ǂݍ */
	for (i = 0; i < NUM_CHR; i++)
	{
		unsigned char buf[SIZEOF_CHRDATA];

		file_read(buf, FILEPOS_CHRDATA + SIZEOF_CHRDATA * i, SIZEOF_CHRDATA);

		/* lnf[^ */
		buf_chrdata[i].strength            = buf[ 6];
		buf_chrdata[i].speed               = buf[ 7];
		buf_chrdata[i].stamina             = buf[ 8];
		buf_chrdata[i].magic_power         = buf[ 9];
		buf_chrdata[i].attack_power        = buf[10];
		buf_chrdata[i].defense_power       = buf[11];
		buf_chrdata[i].evade_rate          = buf[13];
		buf_chrdata[i].magic_defense_power = buf[12];
		buf_chrdata[i].magic_evade_rate    = buf[14];
		buf_chrdata[i].level_cng           = tbl_level_cng[(buf[21] & 0x0c) >> 2];

		buf_chrdata[i].level[0].hit_point   = buf[ 0];
		buf_chrdata[i].level[0].magic_point = buf[ 1];

		/* tOnf[^ */
		buf_chrdata[i].flag_level = ((buf[21] & 0x10) == 0);

		/* ACef[^ */
		for (j = 0; j < 6; j++)
			buf_chrdata[i].equipment_no[j] = buf[j + 15];

		/* R}hf[^ */
		for (j = 0; j < 4; j++)
			buf_chrdata[i].cmd_no[j] = buf[j + 2];

		/* xƂ̃qbg|CgE}WbN|Cg */
		for (j = 1; j < NUM_CHRLV; j++)
		{
			buf_chrdata[i].level[j].hit_point   = tbl_lvup[j].hit_point   + buf_chrdata[i].level[0].hit_point;
			buf_chrdata[i].level[j].magic_point = tbl_lvup[j].magic_point + buf_chrdata[i].level[0].magic_point;
		}
	}

	return buf_chrdata;
}

chrlvup_t *chr_readlvup(void)
{
	int i;
	chrlvup_t *buf_chrlvup = NULL;

	/* LN^f[^i[obt@mۂ */
	if ((buf_chrlvup = malloc(sizeof(chrdata_t) * NUM_CHRLV)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* xƂ̌olf[^ǂݍ */
	buf_chrlvup[0].experience_point = 0;
	for (i = 1; i < NUM_CHRLV; i++)
	{
		unsigned char buf[SIZEOF_CHREXP];

		file_read(buf, FILEPOS_CHREXP + SIZEOF_CHREXP * (i - 1), SIZEOF_CHREXP);
		buf_chrlvup[i].experience_point = buf_chrlvup[i - 1].experience_point + ((buf[1] << 8) | buf[0]) * 8;
	}

	/* xƂ̊{qbg|Cgǂݍ */
	buf_chrlvup[0].hit_point = 0;
	for (i = 1; i < NUM_CHRLV; i++)
	{
		unsigned char buf[SIZEOF_CHRHP];

		file_read(buf, FILEPOS_CHRHP + SIZEOF_CHRHP * (i - 1), SIZEOF_CHRHP);
		buf_chrlvup[i].hit_point = buf_chrlvup[i - 1].hit_point + buf[0];
	}

	/* xƂ̊{}WbN|Cgǂݍ */
	buf_chrlvup[0].magic_point = 0;
	for (i = 1; i < NUM_CHRLV; i++)
	{
		unsigned char buf[SIZEOF_CHRMP];

		file_read(buf, FILEPOS_CHRMP + SIZEOF_CHRMP * (i - 1), SIZEOF_CHRMP);
		buf_chrlvup[i].magic_point = buf_chrlvup[i - 1].magic_point + buf[0];
	}

	return buf_chrlvup;
}

int chr_writeparam(const char *ptr_filename, const chrdata_t *ptr_chrdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* LN^f[^(ln)o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* LN^f[^(ln)o͂ */
	fprintf(fp, "No.      O           "
				"Wu          "
				"HP MP "
				"      ̗  "
				"  U   h    h  "
				"LVC LVe\n");
	for (i = 0; i < NUM_CHR; i++)
	{
		fprintf(fp, "%03d(%02X): %-12s : %-14s,  %5d, %5d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %+5d,  %2s\n", i, i,
				chr_getname(i),
				chr_getjobname(i),
				ptr_chrdata[i].level[0].hit_point,
				ptr_chrdata[i].level[0].magic_point,
				ptr_chrdata[i].strength,
				ptr_chrdata[i].speed,
				ptr_chrdata[i].stamina,
				ptr_chrdata[i].magic_power,
				ptr_chrdata[i].attack_power,
				ptr_chrdata[i].defense_power,
				ptr_chrdata[i].evade_rate,
				ptr_chrdata[i].magic_defense_power,
				ptr_chrdata[i].magic_evade_rate,
				ptr_chrdata[i].level_cng,
				str_getflag(ptr_chrdata[i].flag_level));
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int chr_writeequip(const char *ptr_filename, const chrdata_t *ptr_chrdata)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_nothing = "----------------";

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* LN^̏ACef[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* LN^̏ACef[^o͂ */
	fprintf(fp, "No.      O           "
				"E              "
				"              "
				"                "
				"                "
				"ANZT[1     "
				"ANZT[2\n");
	for (i = 0; i < NUM_CHR; i++)
	{
		const char *ptr_name[6];

		for (j = 0; j < 6; j++)
			ptr_name[j] = (ptr_chrdata[i].equipment_no[j] != 0xff) ? item_getname(ptr_chrdata[i].equipment_no[j]) : str_nothing;

		fprintf(fp, "%03d(%02X): %-12s : ", i, i, chr_getname(i));
		for (j = 0; j < 6; j++)
		{
			if (j != 0)
				fprintf(fp, ", ");
			fprintf(fp, "%-16s", ptr_name[j]);
		}
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int chr_writelvup(const char *ptr_filename, const chrlvup_t *ptr_chrlvup)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrlvup == NULL)
	{
		assert(0);
		return 0;
	}

	/* LN^f[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* LN^f[^o͂ */
	fprintf(fp, "LV       EXP őHP ㏸HP őMP ㏸MP\n");
	for (i = 0; i < NUM_CHRLV; i++)
	{
		fprintf(fp, "%2d: %8ld, %5d, %5d, %5d, %5d\n", i + 1,
				ptr_chrlvup[i].experience_point,
				ptr_chrlvup[i].hit_point,
				(i == 0) ? 0 : ptr_chrlvup[i].hit_point - ptr_chrlvup[i - 1].hit_point,
				ptr_chrlvup[i].magic_point,
				(i == 0) ? 0 : ptr_chrlvup[i].magic_point - ptr_chrlvup[i - 1].magic_point);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}
