/*
 *	ff6define.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "ff6define.h"

#define elem_maskbit(elem_no)	(1U << (elem_no))
#define state_maskbit(state_no)	(1U << (state_no))

#define STR_NOTHING		"E"
#define STR_FLAG_TRUE	""
#define STR_FLAG_FALSE	"~"

static const char *str_element[/* NUM_ELEMENT */] =
{
	"",
	"X",
	"",
	"",
	"",
	"",
	"n",
	"",
};

static const char *str_state[/* NUM_STATE */] =
{
	"",	/* È */
	"]",	/* ]r[ */
	"",	/* ғ */
	"?1",	/* s1(A[}[?) */
	"",	/*  */
	"J",	/* Jbp */
	"",	/* Ή */
	"",	/* 퓬s\ */

	"",	/* ̐鍐 */
	"m",	/* m */
	"",	/* g */
	"",	/*  */
	"",	/* o[TN */
	"",	/*  */
	"X",	/* Xbv */
	"",	/*  */

	"",	/* ł */
	"",	/* WFl */
	"x",	/* XE */
	"",	/* wCXg */
	"~",	/* Xgbv */
	"",	/* VF */
	"",	/* veX */
	"",	/* tN */

	STR_NOTHING,	/* gp */
	"",			/* XÂ */
	"",			/* CY */
	STR_NOTHING,	/* gp */
	STR_NOTHING,	/* gp */
	"E",			/* E */
	"?2",			/* s2(C^[Zv^[?) */
	"",			/* reg */
};

const char *str_getflag(int flag)
{
	return (flag != 0x00) ? STR_FLAG_TRUE : STR_FLAG_FALSE;
}

const char *str_getelem(elem_t elem, int index)
{
	assert(sizeof(str_element) / sizeof(str_element[0]) == NUM_ELEMENT);

	/* `FbN */
	if (index < 0 || NUM_ELEMENT <= index)
	{
		assert(0);
		return "";
	}

	return ((elem & elem_maskbit(index)) != 0x00) ? str_element[index] : "";
}

void str_printelem(FILE *stream, elem_t elem, elem_t elem_mask)
{
	int i;

	assert(sizeof(str_element) / sizeof(str_element[0]) == NUM_ELEMENT);

	/* `FbN */
	if (stream == NULL)
	{
		assert(0);
		return;
	}

	for (i = 0; i < NUM_ELEMENT; i++)
	{
		if ((elem_mask & elem_maskbit(i)) != 0x00)
			fprintf(stream, "%-2s", ((elem & elem_maskbit(i)) != 0x00) ? str_element[i] : STR_NOTHING);
	}
}

const char *str_getstate(state_t state, int index)
{
	assert(sizeof(str_state) / sizeof(str_state[0]) == NUM_STATE);

	/* `FbN */
	if (index < 0 || NUM_STATE <= index)
	{
		assert(0);
		return "";
	}

	return ((state & state_maskbit(index)) != 0x00) ? str_state[index] : "";
}

void str_printstate(FILE *stream, state_t state, state_t state_mask)
{
	int i, count = 0;

	assert(sizeof(str_state) / sizeof(str_state[0]) == NUM_STATE);

	/* `FbN */
	if (stream == NULL)
	{
		assert(0);
		return;
	}

	for (i = 0; i < NUM_STATE; i++)
	{
		if ((i % 8) == 0)
		{
			if (count > 0 && state_mask > state_maskbit(i))
				fprintf(stream, " ");
			count = 0;
		}

		if ((state_mask & state_maskbit(i)) != 0x00)
		{
			fprintf(stream, "%-2s", ((state & state_maskbit(i)) != 0x00) ? str_state[i] : STR_NOTHING);
			count++;
		}
	}
}
