/*
 *	ff6file.c
 *		Copyright (C)2001, 2002 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdlib.h>
#include "errmes.h"
#include "filesfc.h"
#include "ff6fileps.h"
#include "ff6file.h"

#define STR_TITLE_SFC	"FINAL FANTASY 6      "
			/* 0x46, 0x49, 0x4E, 0x41, 0x4C, 0x20, 0x46, 0x41, 0x4E, 0x54, 0x41, 0x53, 0x59, 0x20, 0x36, */
			/* 0x20, 0x20, 0x20, 0x20, 0x20, 0x20 */

static int file_mode = FILEMODE_UNKNOWN;	/* t@CANZX̃[h */
static size_t (*ptr_func_file_read)(void *, long, size_t) = NULL;	/* t@C[hp֐̃|C^ */

int file_open(const char *ptr_path, int mode)
{
	/* `FbN */
	if (ptr_path == NULL || (mode != FILEMODE_SFC && mode != FILEMODE_PS))
	{
		assert(0);
		return 0;
	}

	file_mode = FILEMODE_UNKNOWN;

	if (mode == FILEMODE_SFC)
	{
		if (!file_open_sfc(ptr_path, STR_TITLE_SFC))
			return 0;

		file_mode = FILEMODE_SFC;
		ptr_func_file_read = file_read_sfc;
	}
	else if (mode == FILEMODE_PS)
	{
		if (!file_open_ps(ptr_path))
			return 0;

		file_mode = FILEMODE_PS;
		ptr_func_file_read = file_read_ps;
	}
	else
		assert(0);

	return 1;
}

int file_close(void)
{
	int ret = 0;

	if (file_mode == FILEMODE_SFC)
		ret = file_close_sfc();
	else if (file_mode == FILEMODE_PS)
		ret = file_close_ps();
	else
		assert(0);

	file_mode = FILEMODE_UNKNOWN;
	ptr_func_file_read = NULL;

	return ret;
}

size_t file_read(void *ptr_buf, long filepos, size_t size)
{
	/* ϐ`FbN */
	if (ptr_func_file_read == NULL)
	{
		assert(0);
		return 0;
	}

	return ptr_func_file_read(ptr_buf, filepos, size);
}
