/*
 *	ff6fileps.c
 *		Copyright (C)2001, 2002 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "ff6fileps.h"

#define STR_INVALID_CDROM_FILES		"Invalid CD-ROM files"

struct filelayout_tag
{
	long filepos_sfc;
	long filepos_ps;
	const char *ptr_filename_ps;
};

static size_t unitread(void *ptr_buf, long filepos, size_t size);

static FILE *fp_ps = NULL;						/* t@C|C^(vCXe[Vŗp) */
static char str_dirname_ps[1024] = "";			/* CD-ROM̃[gpX */
static const char *ptr_cur_filename_ps = NULL;	/* ݃I[ṽt@C */

int file_open_ps(const char *ptr_dirname)
{
	/* `FbN */
	if (ptr_dirname == NULL)
	{
		assert(0);
		return 0;
	}

	strncpy(str_dirname_ps, ptr_dirname, sizeof(str_dirname_ps));
	str_dirname_ps[sizeof(str_dirname_ps) - 1] = '\0';

	return 1;
}

int file_close_ps(void)
{
	int ret = 0;

	if (fp_ps != NULL)
		fclose(fp_ps);

	/* f[^ */
	fp_ps = NULL;
	*str_dirname_ps = '\0';
	ptr_cur_filename_ps = NULL;

	return ret;
}

size_t file_read_ps(void *ptr_buf, long filepos, size_t size)
{
	unsigned char *ptr_buf_now = ptr_buf;
	long filepos_now = filepos;
	size_t len_rest = size;

	/* `FbN */
	if (ptr_buf == NULL || filepos < 0)
	{
		assert(0);
		return 0;
	}

	/* w肳ꂽTCYǂݏI܂ŌJԂ */
	while (len_rest > 0)
	{
		size_t ret;

		if ((ret = unitread(ptr_buf_now, filepos_now, len_rest)) == 0)
			break;

		ptr_buf_now += ret;
		filepos_now += ret;
		len_rest    -= ret;
	}

	return size - len_rest;
}

static size_t unitread(void *ptr_buf, long filepos, size_t size)
{
	int i;
	int index;
	size_t ret;
	size_t len_read;
	long filepos_ps;
	static const struct filelayout_tag tbl_filelayout[] =
	{
		{0x000000, 0x000000, NULL},
		{0x00e100, 0x061660, "nar/nar_bin.bin"},
		{0x00e120, 0x061760, "nar/nar_bin.bin"},
		{0x00f4a0, 0x085990, "nar/nar_bin.bin"},
		{0x00fd00, 0x062b00, "nar/nar_bin.bin"},
		{0x00fe00, 0x079780, "nar/nar_bin.bin"},
		{0x00fe40, 0x062c00, "nar/nar_bin.bin"},
		{0x00ff00, 0x000000, NULL},
		{0x02e100, 0x044aa6, "btl/ff6_btl.bin"},
		{0x02e100, 0x044aa6, "btl/ff6_btl.bin"},
		{0x02e24a, 0x03603c, "btl/ff6_btl.bin"},
		{0x02e2ac, 0x000000, NULL},
		{0x02fe6d, 0x071b50, "nar/nar_bin.bin"},
		{0x040000, 0x063700, "oga/oga_bin.bin"},
		{0x046ac0, 0x008f00, "har/menumenu.bin"},
		{0x048fc0, 0x012adb, "har/menumenu.bin"},
		{0x050000, 0x000000, NULL},
		{0x0a0000, 0x000000, "oga/oga_bin.bin"},
		{0x0ce600, 0x06b710, "nar/nar_bin.bin"},
		{0x0d0000, 0x000000, "nar/nar_bin3.bin"},
		{0x0eb000, 0x06a1c0, "oga/oga_bin.bin"},
		{0x0eb500, 0x000000, NULL},
		{0x0efb00, 0x00b400, "har/menumenu.bin"},
		{0x0f0000, 0x000000, "btl/ff6_btl.bin"},
		{0x0f4800, 0x070ad0, "nar/nar_bin.bin"},
		{0x0f5880, 0x01b380, "nar/nar_bin3.bin"},
		{0x0f5900, 0x014800, "btl/ff6_btl.bin"},
		{0x100000, 0x004800, "btl/ff6_btl.bin"},
		{0x110000, 0x026800, "btl/ff6_btl2.bin"},
		{0x120000, 0x053800, "btl/ff6_btl2.bin"},
		{0x127000, 0x019edb, "har/menumenu.bin"},
		{0x130000, 0x036800, "btl/ff6_btl2.bin"},
		{0x14d000, 0x028c28, "btl/ff6_btl.bin"},
		{0x150000, 0x000000, NULL},
		{0x17e4e0, 0x015820, "nar/nar_bin2.bin"},
		{0x17e520, 0x000000, NULL},
		{0x185000, 0x00cabb, "har/menumenu.bin"},
		{0x187000, 0x02eb78, "btl/ff6_btl.bin"},
		{0x18c9a0, 0x02c848, "btl/ff6_btl.bin"},
		{0x18d000, 0x034518, "btl/ff6_btl.bin"},
		{0x18daf2, 0x03500c, "btl/ff6_btl.bin"},
		{0x18dd00, 0x06ba40, "oga/oga_bin.bin"},
		{0x18e800, 0x000000, NULL},
		{0x18f000, 0x000000, "har/titledat.bin"},
		{0x194e96, 0x000000, NULL},
		{0x1982f9, 0x092a39, "nar/nar_bin2.bin"},
		{0x19832c, 0x000000, NULL},
		{0x19a800, 0x081400, "nar/nar_bin.bin"},
		{0x19cd90, 0x080ee0, "nar/nar_bin.bin"},
		{0x19d1b0, 0x062500, "nar/nar_bin2.bin"},
		{0x1fba00, 0x081300, "nar/nar_bin.bin"},
		{0x1fda00, 0x0797c0, "nar/nar_bin.bin"},
		{0x1fdb00, 0x000000, "nar/nar_bin2.bin"},
		{0x260000, 0x087df0, "nar/nar_bin.bin"},
		{0x268400, 0x01b000, "nar/nar_bin3.bin"},
		{0x268780, 0x0798c0, "nar/nar_bin.bin"},
		{0x26cda0, 0x08fdf0, "nar/nar_bin.bin"},
		{0x26f200, 0x06ff10, "nar/nar_bin.bin"},
		{0x26f440, 0x000000, NULL},
		{0x26f490, 0x00bf4b, "har/menumenu.bin"},
		{0x270000, 0x000000, "btl/ff6_btl2.bin"},
		{0x296300, 0x000000, NULL},
		{0x297000, 0x000000, "har/colos.bin"},
		{0x2c0000, 0x000000, NULL},
		{0x2ce300, 0x02ce48, "btl/ff6_btl.bin"},
		{0x2ce3c0, 0x039a34, "btl/ff6_btl.bin"},
		{0x2ce400, 0x000000, NULL},
		{0x2d0000, 0x000000, "har/menumenu.bin"},
		{0x2d8f00, 0x076200, "nar/nar_bin.bin"},
		{0x2dfe00, 0x02e878, "btl/ff6_btl.bin"},
		{0x2e0000, 0x000000, NULL},
		{0x2eaf90, 0x06c5c0, "oga/oga_bin.bin"},
		{0x300000, 0x000000, NULL},
	};

	/* `FbN */
	if (ptr_buf == NULL || filepos < 0)
	{
		assert(0);
		return 0;
	}

	/* ǂ̃t@Ĉǂ̈ʒuɏ񂪊i[Ă邩 */
	index = 0;
	for (i = 0; i < sizeof(tbl_filelayout) / sizeof(struct filelayout_tag) - 1; i++)
	{
		if (filepos < tbl_filelayout[i + 1].filepos_sfc)
		{
			index = i;
			break;
		}
	}

	if (tbl_filelayout[index].ptr_filename_ps == NULL)
		err_printf(STR_INVALID_CDROM_FILES);

	/* t@CقȂĂꍇ̓I[v */
	if (ptr_cur_filename_ps == NULL ||
		(ptr_cur_filename_ps != tbl_filelayout[index].ptr_filename_ps && strcmp(ptr_cur_filename_ps, tbl_filelayout[index].ptr_filename_ps) != 0))
	{
		char path[1024] = "";

		if (fp_ps != NULL)
			fclose(fp_ps);

		strcpy(path, str_dirname_ps);
		strcat(path, tbl_filelayout[index].ptr_filename_ps);
		if ((fp_ps = fopen(path, "rb")) == NULL)
			err_printf(STR_INVALID_CDROM_FILES " - %s", path);

		ptr_cur_filename_ps = tbl_filelayout[index].ptr_filename_ps;
	}

	/* V[Nʒu␳ */
	filepos_ps = ftell(fp_ps);
	if (filepos - tbl_filelayout[index].filepos_sfc != filepos_ps - tbl_filelayout[index].filepos_ps)
	{
		if (fseek(fp_ps, tbl_filelayout[index].filepos_ps + (filepos - tbl_filelayout[index].filepos_sfc), SEEK_SET))
			err_printf(STR_FATAL_ERROR_IN_FSEEK);
		filepos_ps = ftell(fp_ps);
	}

	/* f[^ǂݍ */
	len_read = (size < (size_t)(tbl_filelayout[index + 1].filepos_sfc - filepos)) ?
				size : (size_t)(tbl_filelayout[index + 1].filepos_sfc - filepos);
	if ((ret = fread(ptr_buf, 1, len_read, fp_ps)) < len_read)
		err_printf(STR_FATAL_ERROR_IN_FREAD);

	return ret;
}
