/*
 *	ff6item.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "ff6file.h"
#include "ff6decode.h"
#include "ff6monster.h"
#include "ff6item.h"

#define FILEPOS_ITEMNAME		(0x02cf78 + 0x242788)	/* = 0x26f700 */
#define FILEPOS_ITEMDATA		(0x037a34 + 0x14d5cc)	/* = 0x185000 */
#define FILEPOS_ITEMCOL			(0x013a00 + 0x0fb800)	/* = 0x10f200 */
#define NUM_ITEM				256
#define SIZEOF_ITEMNAME			9
#define SIZEOF_ITEMDATA			30
#define SIZEOF_ITEMCOL			4

static void item_setname(void);

static char tbl_itemname[NUM_ITEM][256];	/* ACe̖Oi[e[u */
static int flag_setitemname = 0;			/* tbl_itemnameɃf[^i[Ă邩ǂ */

const char *item_getname(int index)
{
	/* `FbN */
	if (index < 0 || NUM_ITEM <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setitemname)
		item_setname();

	return tbl_itemname[index];
}

static void item_setname(void)
{
	int i;

	/* ACe̖Of[^ */
	for (i = 0; i < NUM_ITEM; i++)
		*tbl_itemname[i] = '\0';

	/* ACe̖Of[^ǂݍ */
	for (i = 0; i < NUM_ITEM; i++)
	{
		int origin = 0;
		unsigned char buf[SIZEOF_ITEMNAME];

		file_read(buf, FILEPOS_ITEMNAME + SIZEOF_ITEMNAME * i, SIZEOF_ITEMNAME);
		origin = ((0xe3 <= buf[0] && buf[0] <= 0xf4) || buf[0] == 0xff) ? 1 : 0;
		dec_string(tbl_itemname[i], &buf[origin], sizeof(tbl_itemname[i]), SIZEOF_ITEMNAME - origin, 0);
	}

	flag_setitemname = 1;
}

itemdata_t *item_readdata(void)
{
	int i;
	itemdata_t *buf_itemdata = NULL;

	/* ACe̊ef[^i[obt@mۂ */
	if ((buf_itemdata = malloc(sizeof(itemdata_t) * NUM_ITEM)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* ACe̊ef[^ǂݍ */
	for (i = 0; i < NUM_ITEM; i++)
	{
		unsigned char buf[SIZEOF_ITEMDATA];

		file_read(buf, FILEPOS_ITEMDATA + SIZEOF_ITEMDATA * i, SIZEOF_ITEMDATA);

		/* lnf[^ */
		buf_itemdata[i].price                = (buf[29] << 8) | buf[28];
		buf_itemdata[i].attack_defense_power = buf[20];
		buf_itemdata[i].magic_defense_power  = buf[21];
		buf_itemdata[i].evade_rate           = ((buf[26] & 0x0f) >> 0) * 10;
		buf_itemdata[i].magic_evade_rate     = ((buf[26] & 0xf0) >> 4) * 10;

		buf_itemdata[i].increase_strength    = (((buf[16] & 0x08) == 0x00) ? 1 : -1) * ((buf[16] & 0x07) >> 0);
		buf_itemdata[i].increase_speed       = (((buf[16] & 0x80) == 0x00) ? 1 : -1) * ((buf[16] & 0x70) >> 4);
		buf_itemdata[i].increase_stamina     = (((buf[17] & 0x08) == 0x00) ? 1 : -1) * ((buf[17] & 0x07) >> 0);
		buf_itemdata[i].increase_magic_power = (((buf[17] & 0x80) == 0x00) ? 1 : -1) * ((buf[17] & 0x70) >> 4);

		buf_itemdata[i].effect_no        = buf[18] & 0x3f;
		buf_itemdata[i].learn_magic_no   = buf[ 4];
		buf_itemdata[i].learn_magic_rate = buf[ 3];

		/* nf[^ */
		buf_itemdata[i].elem_half   = buf[15];
		buf_itemdata[i].elem_none   = buf[22];
		buf_itemdata[i].elem_absorb = buf[23];
		buf_itemdata[i].elem_double = buf[24];

		/* ԕωnf[^ */
		buf_itemdata[i].state_defense = bits_make32(buf[ 6], buf[ 7], 0x00, 0x00);
		buf_itemdata[i].state_apriori = bits_make32(0x00, buf[25], buf[ 8], 0x00);

		/* tOnf[^ */
		buf_itemdata[i].bits_equip_chr         = (buf[ 2] << 8) | buf[ 1];
		buf_itemdata[i].flag_throw             = ((buf[ 0] & 0x10) != 0x00);
		buf_itemdata[i].flag_using_effect      = ((buf[18] & 0x80) != 0x00);
		buf_itemdata[i].flag_additional_effect = ((buf[18] & 0x40) != 0x00);
		buf_itemdata[i].flag_runic             = ((buf[19] & 0x80) != 0x00);
		buf_itemdata[i].flag_2handed           = ((buf[19] & 0x40) != 0x00);
		buf_itemdata[i].flag_backrow           = ((buf[19] & 0x20) != 0x00);
		buf_itemdata[i].flag_swordtech         = ((buf[19] & 0x02) != 0x00);
		buf_itemdata[i].flag_range_all         = (buf[14] == 0x04);	/* 0x04ʈ */
		buf_itemdata[i].direction              = ((buf[14] & 0x40) != 0x00) ? 1 : 0;
		buf_itemdata[i].range                  = ((buf[14] & 0x08) != 0x00) ? 1 : 0;
		buf_itemdata[i].change_direction       = ((buf[14] & 0x02) == 0x00 && (buf[14] & 0x01) != 0x00) ? 1 : 0;
		buf_itemdata[i].change_range           = ((buf[14] & 0x20) != 0x00 && (buf[14] & 0x01) != 0x00) ? 1 : 0;
	}

	/* RVAACef[^ǂݍ */
	for (i = 0; i < NUM_ITEM; i++)
	{
		unsigned char buf[SIZEOF_ITEMCOL];

		file_read(buf, FILEPOS_ITEMCOL + SIZEOF_ITEMCOL * i, SIZEOF_ITEMCOL);

		buf_itemdata[i].col_monster_no = buf[0];
		buf_itemdata[i].col_item_no    = buf[2];
	}

	return buf_itemdata;
}

int item_writeparam(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ACe̐lnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ACe̐lnf[^o͂ */
	fprintf(fp, "No.      O               "
				"  i "
				" U/h  h  "
				"      ̗ \n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :  %5ld, %5d, %5d, %5d, %5d, %3d, %3d, %3d, %3d\n", i, i,
				item_getname(i),
				ptr_itemdata[i].price,
				ptr_itemdata[i].attack_defense_power,
				ptr_itemdata[i].evade_rate,
				ptr_itemdata[i].magic_defense_power,
				ptr_itemdata[i].magic_evade_rate,
				ptr_itemdata[i].increase_strength,
				ptr_itemdata[i].increase_speed,
				ptr_itemdata[i].increase_stamina,
				ptr_itemdata[i].increase_magic_power);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writeelem(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ACȇnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ACȇnf[^o͂ */
	fprintf(fp, "No.      O               "
				"/U     "
				"          "
				"z          "
				"_\n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s : ", i, i, item_getname(i));

		/*  */
		str_printelem(fp, ptr_itemdata[i].elem_half, ELEM_ALL);
		fprintf(fp, "  ");

		/*  */
		str_printelem(fp, ptr_itemdata[i].elem_none, ELEM_ALL);
		fprintf(fp, "  ");

		/* z */
		str_printelem(fp, ptr_itemdata[i].elem_absorb, ELEM_ALL);
		fprintf(fp, "  ");

		/* _ */
		str_printelem(fp, ptr_itemdata[i].elem_double, ELEM_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writestate(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ACẽtOnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ACẽtOnf[^o͂ */
	fprintf(fp, "No.      O               "
				"ԕωϐ1    2                 "
				"ԕω2    3\n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s : ", i, i, item_getname(i));

		/* ԕωւ̑ϐ */
		str_printstate(fp, ptr_itemdata[i].state_defense, STATE_ALL & ~STATE_ASSISTMAG & ~STATE_OTHER);
		fprintf(fp, "  ");

		/* ŏKpԕω */
		str_printstate(fp, ptr_itemdata[i].state_apriori, STATE_ALL & ~STATE_PERMANENT & ~STATE_OTHER);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writeflag(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_nothing = "..";
	const char *str_direction[2][2] =
	{
		{"",	"/G"},
		{"G",		"G/"},
	};
	const char *str_range[2][2] =
	{
		{"P",	"P/S"},
		{"S",	"S/P"},
	};
	const char *str_range_all = "GS";
	const char *str_chrmain[] =
	{
		"Ti",	/* eBi */
		"Lo",	/* bN */
		"Ca",	/* JCG */
		"Sh",	/* VhE */
		"Ed",	/* GhK[ */
		"Ma",	/* }bV */
		"Ce",	/* ZX */
		"St",	/* XgSX */
		"Re",	/*  */
		"Se",	/* Zbc@[ */
		"Mo",	/* O */
		"Ga",	/* KE */
		"Go",	/* SS */
		"Um",	/* E[} */
		"?1",	/* s1 */
		"?2",	/* s2 */
	};

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ACẽtOnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ACẽtOnf[^o͂ */
	fprintf(fp, "No.      O               "
				"    KE  "
				"\LN^                               "
				"ʔ͈\n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :", i, i, item_getname(i));

		/* etO */
		fprintf(fp, "  %2s   %2s   %2s   %2s   %2s  ",
				str_getflag(ptr_itemdata[i].flag_throw),
				str_getflag(ptr_itemdata[i].flag_backrow),
				str_getflag(ptr_itemdata[i].flag_2handed),
				str_getflag(ptr_itemdata[i].flag_runic),
				str_getflag(ptr_itemdata[i].flag_swordtech));

		/* \LN^ */
		for (j = 0; j < 16; j++)
			fprintf(fp, " %-2s", ((ptr_itemdata[i].bits_equip_chr & (1U << (j))) != 0x00) ? str_chrmain[j] : str_nothing);
		fprintf(fp, "  ");

		/* ʔ͈ */
		if (ptr_itemdata[i].flag_range_all)
			fprintf(fp, "%-17s", str_range_all);
		else
		{
			fprintf(fp, "%-7s %-9s",
					str_direction[ptr_itemdata[i].direction][ptr_itemdata[i].change_direction],
					str_range[ptr_itemdata[i].range][ptr_itemdata[i].change_range]);
		}
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writeeffect(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;
	const char *str_nothing = "----------------";

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ACěʃf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ACěʃf[^o͂ */
	fprintf(fp, "No.      O               "
				"gp ǉ  "
				"          "
				"C@          "
				"C\n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		const char *ptr_name_effect = str_nothing;
		const char *ptr_name_learn  = str_nothing;

		ptr_name_effect = (ptr_itemdata[i].flag_using_effect || ptr_itemdata[i].flag_additional_effect) ?
							mag_getname(ptr_itemdata[i].effect_no) : str_nothing;
		ptr_name_learn  = (ptr_itemdata[i].learn_magic_rate > 0) ?
							mag_getname(ptr_itemdata[i].learn_magic_no) : str_nothing;

		fprintf(fp, "%03d(%02X): %-16s :  %2s   %2s   %-16s, %-16s,  %5d\n", i, i, item_getname(i),
				str_getflag(ptr_itemdata[i].flag_using_effect),
				str_getflag(ptr_itemdata[i].flag_additional_effect),
				ptr_name_effect,
				ptr_name_learn,
				ptr_itemdata[i].learn_magic_rate);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writecol(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ACěʃf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ACěʃf[^o͂ */
	fprintf(fp, "No.      O               "
				"ACe      ΐ탂X^[\n");
	for (i = 0; i < NUM_ITEM; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s : %-16s, %-16s\n", i, i, item_getname(i),
				item_getname(ptr_itemdata[i].col_item_no),
				mon_getname(ptr_itemdata[i].col_monster_no));
	}

	fclose(fp);
	fp = NULL;

	return 1;
}
