/*
 *	ff6magic.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "ff6file.h"
#include "ff6decode.h"
#include "ff6magic.h"

#define FILEPOS_MAGNAME			(0x02bda8 + 0x2ab658)	/* = 0x2d7400 */
#define FILEPOS_MAGDATA			(0x036b54 + 0x00ff6c)	/* = 0x046ac0 */
#define NUM_MAG_SHORT			54
#define NUM_MAG_MIDDLE			27
#define NUM_MAG_LONG			175
#define NUM_MAG					(NUM_MAG_SHORT + NUM_MAG_MIDDLE + NUM_MAG_LONG)	/* = 256 */
#define SIZEOF_MAGNAME_SHORT	5
#define SIZEOF_MAGNAME_MIDDLE	8
#define SIZEOF_MAGNAME_LONG		8
#define SIZEOF_MAGDATA			14
#define REGION_SHORT_BEGIN		0
#define REGION_SHORT_END		(REGION_SHORT_BEGIN + NUM_MAG_SHORT - 1)	/* =  53 */
#define REGION_MIDDLE_BEGIN		(REGION_SHORT_END + 1)						/* =  54 */
#define REGION_MIDDLE_END		(REGION_MIDDLE_BEGIN + NUM_MAG_MIDDLE - 1)	/* =  80 */
#define REGION_LONG_BEGIN		(REGION_MIDDLE_END + 1)						/* =  81 */
#define REGION_LONG_END			(REGION_LONG_BEGIN + NUM_MAG_LONG - 1)		/* = 255 */

static void mag_setname(void);

static char tbl_magname[NUM_MAG][256];	/* @̖Oi[e[u */
static int flag_setmagname = 0;			/* tbl_magnameɃf[^i[Ă邩ǂ */

const char *mag_getname(int index)
{
	/* `FbN */
	if (index < 0 || NUM_MAG <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setmagname)
		mag_setname();

	return tbl_magname[index];
}

static void mag_setname(void)
{
	int i;
	long filepos;

	/* @̖Of[^ */
	for (i = 0; i < NUM_MAG; i++)
		*tbl_magname[i] = '\0';

	/* @̖Of[^ǂݍ */
	filepos = FILEPOS_MAGNAME;
	for (i = 0; i < NUM_MAG; i++)
	{
		int origin = 0;
		size_t len_read = 0;
		int flag_multibyte = 0;
		unsigned char buf[SIZEOF_MAGNAME_LONG];

		if (REGION_SHORT_BEGIN <= i && i <= REGION_SHORT_END)
		{
			origin = 1;
			len_read = SIZEOF_MAGNAME_SHORT;
			flag_multibyte = 0;
		}
		else if (REGION_MIDDLE_BEGIN <= i && i <= REGION_MIDDLE_END)
		{
			origin = 0;
			len_read = SIZEOF_MAGNAME_MIDDLE;
			flag_multibyte = 0;
		}
		else if (REGION_LONG_BEGIN <= i && i <= REGION_LONG_END)
		{
			origin = 0;
			len_read = SIZEOF_MAGNAME_LONG;
			flag_multibyte = 1;
		}
		else
			assert(0);

		filepos += file_read(buf, filepos, len_read);
		dec_string(tbl_magname[i], &buf[origin], sizeof(tbl_magname[i]), len_read - origin, flag_multibyte);
	}

	flag_setmagname = 1;
}

magdata_t *mag_readdata(void)
{
	int i;
	magdata_t *buf_magdata = NULL;

	/* @̊ef[^i[obt@mۂ */
	if ((buf_magdata = malloc(sizeof(magdata_t) * NUM_MAG)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* @̊ef[^ǂݍ */
	for (i = 0; i < NUM_MAG; i++)
	{
		unsigned char buf[SIZEOF_MAGDATA];

		file_read(buf, FILEPOS_MAGDATA + SIZEOF_MAGDATA * i, SIZEOF_MAGDATA);

		/* lnf[^ */
		buf_magdata[i].magic_point  = buf[5];

		/* Eԕωnf[^ */
		buf_magdata[i].elem_effect  = buf[1];
		buf_magdata[i].state_effect = bits_make32(buf[10], buf[11], buf[12], buf[13]);

		/* tOnf[^ */
		buf_magdata[i].flag_runic            = ((buf[3] & 0x08) != 0x00);
		buf_magdata[i].flag_reflect          = ((buf[3] & 0x02) == 0x00);
		buf_magdata[i].flag_cure             = ((buf[4] & 0x01) != 0x00);
		buf_magdata[i].flag_absorb           = ((buf[4] & 0x02) != 0x00);
		buf_magdata[i].flag_ignore_defense   = ((buf[2] & 0x20) != 0x00);
		buf_magdata[i].flag_physical         = ((buf[2] & 0x01) != 0x00);
		buf_magdata[i].flag_state_reset      = ((buf[4] & 0x04) != 0x00);
		buf_magdata[i].flag_state_reverse    = ((buf[4] & 0x08) != 0x00);
		buf_magdata[i].flag_ratio            = ((buf[4] & 0x80) != 0x00);
		buf_magdata[i].flag_level            = ((buf[4] & 0x40) != 0x00);
		buf_magdata[i].flag_range_all        = (buf[0] == 0x04);	/* 0x04ʈ */
		buf_magdata[i].flag_roulette         = ((buf[0] & 0x80) != 0x00);
		buf_magdata[i].direction             = ((buf[0] & 0x40) != 0x00) ? 1 : 0;
		buf_magdata[i].range                 = ((buf[0] & 0x08) != 0x00) ? 1 : 0;
		buf_magdata[i].change_direction      = ((buf[0] & 0x02) == 0x00 && (buf[0] & 0x01) != 0x00) ? 1 : 0;
		buf_magdata[i].change_range          = ((buf[0] & 0x20) != 0x00 && (buf[0] & 0x01) != 0x00) ? 1 : 0;

		/* buf[6]Ebuf[8]̉ */
		buf_magdata[i].effect_power = 0;
		buf_magdata[i].effect_rate  = 0;
		buf_magdata[i].effect_level = 0;
		buf_magdata[i].hit_rate     = 0;

		if (buf_magdata[i].flag_ratio)
			buf_magdata[i].effect_rate  = buf[6];
		else
			buf_magdata[i].effect_power = buf[6];

		if (buf_magdata[i].flag_level)
			buf_magdata[i].effect_level = buf[8];
		else
			buf_magdata[i].hit_rate     = buf[8];

	}

	return buf_magdata;
}

int mag_writeparam(const char *ptr_filename, const magdata_t *ptr_magdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_magdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* @̐lnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* @̐lnf[^o͂ */
	fprintf(fp, "No.      O               "
				"MP ʒl   䗦   Hit%% x\n");
	for (i = 0; i < NUM_MAG; i++)
	{
		fprintf(fp, (ptr_magdata[i].effect_rate != 0) ?
					"%03d(%02X): %-16s :  %5d, %5d,%3d/16, %5d, %5d\n" :
					"%03d(%02X): %-16s :  %5d, %5d, %5d, %5d, %5d\n",
				i, i, mag_getname(i),
				ptr_magdata[i].magic_point,
				ptr_magdata[i].effect_power,
				ptr_magdata[i].effect_rate,
				ptr_magdata[i].hit_rate,
				ptr_magdata[i].effect_level);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mag_writeelem(const char *ptr_filename, const magdata_t *ptr_magdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_magdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* @̑nf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* @̑nf[^o͂ */
	fprintf(fp, "No.      O               "
				"\n");
	for (i = 0; i < NUM_MAG; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s : ", i, i, mag_getname(i));

		/*  */
		str_printelem(fp, ptr_magdata[i].elem_effect, ELEM_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mag_writestate(const char *ptr_filename, const magdata_t *ptr_magdata)
{
	int i;
	FILE *fp = NULL;
	const char *str_state[2][2] =
	{
		{"ω",	""},
		{"]",	"]"},
	};

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_magdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* @̏ԕωnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* @̏ԕωnf[^o͂ */
	fprintf(fp, "No.      O               "
				"  "
				"ԕω1        2                3                4\n");
	for (i = 0; i < NUM_MAG; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s : %4s  ", i, i, mag_getname(i),
				str_state[(ptr_magdata[i].flag_state_reverse) ? 1 : 0][(ptr_magdata[i].flag_state_reset) ? 1 : 0]);

		/* ԕω */
		str_printstate(fp, ptr_magdata[i].state_effect, STATE_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mag_writeflag(const char *ptr_filename, const magdata_t *ptr_magdata)
{
	int i;
	FILE *fp = NULL;
	const char *str_type[2] = {"", "@"};
	const char *str_direction[2][2] =
	{
		{"",	"/G"},
		{"G",		"G/"},
	};
	const char *str_range[2][2] =
	{
		{"P",	"P/S"},
		{"S",	"S/P"},
	};
	const char *str_range_all = "GS";
	const char *str_roulette  = "[bg";

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_magdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* @̃tOnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* @̃tOnf[^o͂ */
	fprintf(fp, "No.      O               "
				"  h 䗦  "
				"  ʔ͈\n");
	for (i = 0; i < NUM_MAG; i++)
	{
		/* etO */
		fprintf(fp, "%03d(%02X): %-16s :  %2s   %2s   %2s   %2s   %4s  ", i, i, mag_getname(i),
				str_getflag(ptr_magdata[i].flag_runic),
				str_getflag(ptr_magdata[i].flag_reflect),
				str_getflag(ptr_magdata[i].flag_ignore_defense),
				str_getflag(ptr_magdata[i].flag_ratio),
				str_type[(ptr_magdata[i].flag_physical) ? 0 : 1]);

		/* ʔ͈ */
		if (ptr_magdata[i].flag_range_all)
			fprintf(fp, "%-17s", str_range_all);
		else if (ptr_magdata[i].flag_roulette)
			fprintf(fp, "%-17s", str_roulette);
		else
		{
			fprintf(fp, "%-7s %-9s",
					str_direction[ptr_magdata[i].direction][ptr_magdata[i].change_direction],
					str_range[ptr_magdata[i].range][ptr_magdata[i].change_range]);
		}

		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}
