/*
 *	ff6monster.c
 *		Copyright (C)2000-2002 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include "errmes.h"
#include "ff6file.h"
#include "ff6decode.h"
#include "ff6character.h"
#include "ff6monster.h"

#define FILEPOS_MONNAME			(0x01b700 + 0x0e1100)	/* = 0x0fc800 */
#define FILEPOS_MONDATA			(0x000000 + 0x0f0000)	/* = 0x0f0000 */
#define FILEPOS_MONITEM			(0x003000 + 0x0f0000)	/* = 0x0f3000 */
#define FILEPOS_MONRAGNA		(0x0379b4 + 0x01058c)	/* = 0x047f40 */
#define FILEPOS_MONCTRL			(0x003d00 + 0x0f0000)	/* = 0x0f3d00 */
#define FILEPOS_MONSKETCH		(0x004300 + 0x0f0000)	/* = 0x0f4300 */
#define FILEPOS_MONRAGE			(0x004600 + 0x0f0000)	/* = 0x0f4600 */
#define FILEPOS_MONACT_ADDR		(0x017300 + 0x0e1100)	/* = 0x0f8400 */
#define FILEPOS_MONACT			(0x017600 + 0x0e1100)	/* = 0x0f8700 */
#define FILEPOS_MONTECHNAME		(0x01c300 + 0x0e1100)	/* = 0x0fd400 */
#define FILEPOS_MONMES_ADDR		(0x01cf00 + 0x0e1100)	/* = 0x0fe000 */
#define FILEPOS_MONMES			(0x01d300 + 0x0e1100)	/* = 0x0fe400 */
#define FILEPOS_MONPARTY		(0x015100 + 0x0e1100)	/* = 0x0f6200 */
#define FILEPOS_MONLEARNGAIN	(0x014600 + 0x0fb800)	/* = 0x10fe00 */
#define FILEPOS_MONPARTY2		(0x014800 + 0x0e1100)	/* = 0x0f5900 */
#define FILEPOS_EVENTMES_ADDR	(0x014400 + 0x0fb800)	/* = 0x10fc00 */
#define FILEPOS_EVENTMES		(0x011a00 + 0x0fb800)	/* = 0x10d200 */
#define FILEPOS_EVENTDATA_ADDR	(0x00e000 + 0x0fb800)	/* = 0x109800 */
#define FILEPOS_EVENTDATA		(0x00e042 + 0x0fb800)	/* = 0x109842 */
#define FILEPOS_MONENCRND		(0x06a7c0 + 0x08a040)	/* = 0x0f4800 */
#define FILEPOS_MONENCFIXED		(0x06afc0 + 0x08a040)	/* = 0x0f5000 */
#define FILEPOS_MONMAP			(0x06b3c0 + 0x08a040)	/* = 0x0f5400 */
#define NUM_MON					384
#define NUM_MONTECH				NUM_MON
#define NUM_MONMES				142
#define NUM_MONRAGNA			26
#define NUM_MONPARTY			576
#define NUM_MONENCRND			256
#define NUM_MONENCFIXED			256
#define NUM_MONENC				(NUM_MONENCRND + NUM_MONENCFIXED)	/* = 512 */
#define NUM_MONMAP				2
#define NUM_MONMAPAREA			8
#define NUM_EVENTMES			256
#define NUM_EVENTDATA			33
#define SIZEOF_MONNAME			8
#define SIZEOF_MONDATA			32
#define SIZEOF_MONITEM			4
#define SIZEOF_MONRAGNA			4
#define SIZEOF_MONCTRL			4
#define SIZEOF_MONSKETCH		2
#define SIZEOF_MONRAGE			2
#define SIZEOF_MONTECHNAME		8
#define SIZEOF_MONLEARNGAIN		1
#define SIZEOF_MONPARTY			15
#define SIZEOF_MONPARTY2		4
#define SIZEOF_MONENCRND		8
#define SIZEOF_MONENCFIXED		4
#define SIZEOF_MONMAP			4
#define SIZEOF_OFFSETADDR		2

static const char *mon_gettechname(int index);
static void mon_setname(void);
static void mon_settechname(void);
static void mon_setmes(void);
static void mon_seteventdata(void);
static const char *mon_gettarget(int index);
static const char *mon_getmes(int index);
static const char *mon_geteventmes(int index);
static int mon_geteventdata(int index1, int index2);
static unsigned char *mon_printactcnd(FILE *stream, unsigned char *buf_cnd);
static void mon_printact(FILE *stream, int index, int monster_no);

static char tbl_monname[NUM_MON][256];			/* X^[̖Oi[e[u */
static char tbl_montechname[NUM_MONTECH][256];	/* uƂv̖Oi[e[u */
static char tbl_monmes[NUM_MONMES][256];		/* X^[̃bZ[Wi[e[u */
static char tbl_eventmes[NUM_EVENTMES][256];	/* Cxg̃bZ[Wi[e[u */
static int tbl_eventdata[NUM_EVENTDATA][256];	/* Cxgf[^i[e[u */
static int flag_setmonname      = 0;			/* tbl_monnameɃf[^i[Ă邩ǂ */
static int flag_setmontechname  = 0;			/* tbl_montechnameɃf[^i[Ă邩ǂ */
static int flag_setmoneventmes  = 0;			/* tbl_monmesEtbl_eventmesɃf[^i[Ă邩ǂ */
static int flag_setmoneventdata = 0;			/* tbl_eventdataɃf[^i[Ă邩ǂ */

static const int tbl_skip_byte[16] = {4, 2, 4, 3, 4, 4, 4, 2, 3, 4, 4, 3, 4, 1, 1, 1};

const char *mon_getname(int index)
{
	/* `FbN */
	if (index < 0 || NUM_MON <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setmonname)
		mon_setname();

	return tbl_monname[index];
}

static const char *mon_gettechname(int index)
{
	/* `FbN */
	if (index < 0 || NUM_MONTECH <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setmontechname)
		mon_settechname();

	return tbl_montechname[index];
}

static const char *mon_getmes(int index)
{
	/* `FbN */
	if (index < 0 || NUM_MONMES <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setmoneventmes)
		mon_setmes();

	return tbl_monmes[index];
}

static const char *mon_geteventmes(int index)
{
	/* `FbN */
	if (index < 0 || NUM_EVENTMES <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setmoneventmes)
		mon_setmes();

	return tbl_eventmes[index];
}

static int mon_geteventdata(int index1, int index2)
{
	/* `FbN */
	if (index1 < 0 || NUM_EVENTDATA <= index1 || index2 < 0 || 64 <= index2)
	{
		assert(0);
		return -1;
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setmoneventdata)
		mon_seteventdata();

	return tbl_eventdata[index1][index2];
}

static const char *mon_gettarget(int index)
{
	const char *ptr = NULL;
	static const char *str_target[] =
	{
		"X^[1", "X^[2", "X^[3", "X^[4",
		"X^[5", "X^[6", "X^[g", "ȊÕX^[S",
		"X^[S", "ȊÕX^[̒N", "X^[̒N", "퓬s\̃LN^S",
		"퓬s\̃LN^̒N", "퓬s\̃X^[S", "퓬s\̃X^[̒N", "3Fh",
		"40h", "41h", "42h", "LN^S",
		"LN^̒N", "ŌɎɍULN^̒N", "GS", "47h",
		"1Ԗڂ̃LN^", "2Ԗڂ̃LN^", "3Ԗڂ̃LN^", "4Ԗڂ̃LN^",
		"4Ch", "4Dh",
	};

	assert(sizeof(str_target) / sizeof(str_target[0]) == 0x4e - 0x30);

	if (index < 0x30)
		ptr = chr_getname(index);
	else if (index < 0x4e)
		ptr = str_target[index - 0x30];

	return ptr;
}

static void mon_setname(void)
{
	int i;

	/* X^[̖Of[^ */
	for (i = 0; i < NUM_MON; i++)
		*tbl_monname[i] = '\0';

	/* X^[̖Of[^ǂݍ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONNAME];

		file_read(buf, FILEPOS_MONNAME + SIZEOF_MONNAME * i, SIZEOF_MONNAME);
		dec_string(tbl_monname[i], buf, sizeof(tbl_monname[i]), SIZEOF_MONNAME, 0);
	}

	flag_setmonname = 1;
}

static void mon_settechname(void)
{
	int i;

	/* X^[́uƂv̖Of[^ */
	for (i = 0; i < NUM_MONTECH; i++)
		*tbl_montechname[i] = '\0';

	/* X^[́uƂv̖Of[^ǂݍ */
	for (i = 0; i < NUM_MONTECH; i++)
	{
		unsigned char buf[SIZEOF_MONTECHNAME];

		file_read(buf, FILEPOS_MONTECHNAME + SIZEOF_MONTECHNAME * i, SIZEOF_MONTECHNAME);
		dec_string(tbl_montechname[i], buf, sizeof(tbl_montechname[i]), SIZEOF_MONTECHNAME, 1);
	}

	flag_setmontechname = 1;
}

static void mon_setmes(void)
{
	int i;
	long filepos;
	int tbl_offset_m[NUM_MONMES + 1];
	int tbl_offset_e[NUM_EVENTMES + 1];

	/* bZ[Wf[^ */
	for (i = 0; i < NUM_MONMES; i++)
		*tbl_monmes[i] = '\0';
	for (i = 0; i < NUM_EVENTMES; i++)
		*tbl_eventmes[i] = '\0';

	/* X^[̃bZ[Wf[^i[ĂAhX(ւ̃ItZbg)ǂݍ */
	for (i = 0; i < NUM_MONMES; i++)
	{
		unsigned char buf[SIZEOF_OFFSETADDR];

		file_read(buf, FILEPOS_MONMES_ADDR + SIZEOF_OFFSETADDR * i, SIZEOF_OFFSETADDR);
		tbl_offset_m[i] = (buf[1] << 8) | buf[0];
	}
	tbl_offset_m[NUM_MONMES] = tbl_offset_m[NUM_MONMES - 1] + 0x0d;	/*  */

	/* X^[̃bZ[Wf[^ǂݍ */
	filepos = FILEPOS_MONMES;
	for (i = 0; i < NUM_MONMES; i++)
	{
		size_t bufsize;
		unsigned char *buf = NULL;

		if (tbl_offset_m[i + 1] - tbl_offset_m[i] < 0)
			break;

		bufsize = tbl_offset_m[i + 1] - tbl_offset_m[i];
		if ((buf = malloc(bufsize)) == NULL)
			err_printf(STR_FATAL_ERROR_IN_MALLOC);

		filepos += file_read(buf, filepos, bufsize);
		if (bufsize > 1)
			dec_string(tbl_monmes[i], buf, sizeof(tbl_monmes[i]), bufsize - 1, 1);

		free(buf);
		buf = NULL;
	}

	/* Cxg̃bZ[Wf[^i[ĂAhX(ւ̃ItZbg)ǂݍ */
	for (i = 0; i < NUM_EVENTMES; i++)
	{
		unsigned char buf[SIZEOF_OFFSETADDR];

		file_read(buf, FILEPOS_EVENTMES_ADDR + SIZEOF_OFFSETADDR * i, SIZEOF_OFFSETADDR);
		tbl_offset_e[i] = (buf[1] << 8) | buf[0];
	}
	tbl_offset_e[NUM_EVENTMES] = tbl_offset_e[NUM_EVENTMES - 1] + 0x68;	/*  */

	/* Cxg̃bZ[Wf[^ǂݍ */
	filepos = FILEPOS_EVENTMES;
	for (i = 0; i < NUM_EVENTMES; i++)
	{
		size_t bufsize;
		unsigned char *buf = NULL;

		if (tbl_offset_e[i + 1] - tbl_offset_e[i] < 0)
			break;

		bufsize = tbl_offset_e[i + 1] - tbl_offset_e[i];
		if ((buf = malloc(bufsize)) == NULL)
			err_printf(STR_FATAL_ERROR_IN_MALLOC);

		filepos += file_read(buf, filepos, bufsize);
		if (bufsize > 1)
			dec_string(tbl_eventmes[i], buf, sizeof(tbl_eventmes[i]), bufsize - 1, 1);

		free(buf);
		buf = NULL;
	}

	flag_setmoneventmes = 1;
}

static void mon_seteventdata(void)
{
	int i;
	const size_t bufsize = 0x3900 - SIZEOF_OFFSETADDR * NUM_EVENTDATA;
	unsigned char *buf = NULL;
	int tbl_offset[NUM_EVENTDATA + 1];

	/* Cxgf[^ */
	for (i = 0; i < NUM_EVENTDATA; i++)
		memset(tbl_eventdata[i], 0, sizeof(tbl_eventdata[i]));

	/* Cxgf[^i[ĂAhX(ւ̃ItZbg)ǂݍ */
	for (i = 0; i < NUM_EVENTDATA; i++)
	{
		unsigned char buf[SIZEOF_OFFSETADDR];

		file_read(buf, FILEPOS_EVENTDATA_ADDR + SIZEOF_OFFSETADDR * i, SIZEOF_OFFSETADDR);
		tbl_offset[i] = (buf[1] << 8) | buf[0];
	}
	tbl_offset[NUM_EVENTDATA] = tbl_offset[NUM_EVENTDATA - 1] + (0x00d100 - 0x00cf45);	/*  */

	/* Cxgf[^ǂݍ */
	if ((buf = malloc(bufsize)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);
	memset(buf, 0x00, bufsize);
	file_read(buf, FILEPOS_EVENTDATA, bufsize);

	for (i = 0; i < NUM_EVENTDATA; i++)
	{
		int index = 0;
		int flag_chunk = 0;
		unsigned char *ptr = buf + (ptrdiff_t)(tbl_offset[i] - tbl_offset[0]);

		while (*ptr != 0xff || flag_chunk)
		{
			switch (*ptr)
			{
			case 0x00:	/* bZ[W\(㕔EBhE) */
			case 0x01:	/* bZ[W\(EBhE) */
				if (!flag_chunk)
				{
					ptr++;
					tbl_eventdata[i][index] = *ptr;
					index++;
				}
				break;

			case 0x03:	/* s */
				ptr++;
				break;

			case 0x0d:	/* ÃCxgJn? */
				ptr++;
				/* FALLTHRU */
			case 0x0e:	/* ÃCxgJn? */
				flag_chunk = 1;
				break;

			case 0x0f:	/* ÃCxgI? */
				flag_chunk = 0;
				break;
			}

			ptr++;
		}

		for (; index < 64; index++)
			tbl_eventdata[i][index] = -1;
	}

	free(buf);
	buf = NULL;

	flag_setmoneventdata = 1;
}

mondata_t *mon_readdata(void)
{
	int i, j;
	mondata_t *buf_mondata = NULL;
	int tbl_ragna[NUM_MONRAGNA][SIZEOF_MONRAGNA];

	/* X^[̊ef[^i[obt@mۂ */
	if ((buf_mondata = malloc(sizeof(mondata_t) * NUM_MON)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* X^[̏ACef[^ǂݍ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONITEM];

		file_read(buf, FILEPOS_MONITEM + SIZEOF_MONITEM * i, SIZEOF_MONITEM);
		for (j = 0; j < SIZEOF_MONITEM; j++)
			buf_mondata[i].item_no[j] = buf[j];
	}

	/* X^[̃ACeωf[^ǂݍ */
	for (i = 0; i < NUM_MONRAGNA; i++)
	{
		unsigned char buf[SIZEOF_MONRAGNA];

		file_read(buf, FILEPOS_MONRAGNA + SIZEOF_MONRAGNA * i, SIZEOF_MONRAGNA);
		for (j = 0; j < SIZEOF_MONRAGNA; j++)
			tbl_ragna[i][j] = buf[j];
	}

	/* X^[́uvsf[^ǂݍ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONCTRL];

		file_read(buf, FILEPOS_MONCTRL + SIZEOF_MONCTRL * i, SIZEOF_MONCTRL);
		for (j = 0; j < SIZEOF_MONCTRL; j++)
			buf_mondata[i].control_no[j] = buf[j];
	}

	/* X^[́uXPb`vsf[^ǂݍ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONSKETCH];

		file_read(buf, FILEPOS_MONSKETCH + SIZEOF_MONSKETCH * i, SIZEOF_MONSKETCH);
		for (j = 0; j < SIZEOF_MONSKETCH; j++)
			buf_mondata[i].sketch_no[j] = buf[j];
	}

	/* X^[́u΂vsf[^ǂݍ */
	for (i = 0; i < NUM_MON; i++)
	{
		if (i >= 256)
		{
			for (j = 0; j < SIZEOF_MONRAGE; j++)
				buf_mondata[i].rage_no[j] = 0xff;
		}
		else
		{
			unsigned char buf[SIZEOF_MONRAGE];

			file_read(buf, FILEPOS_MONRAGE + SIZEOF_MONRAGE * i, SIZEOF_MONRAGE);
			for (j = 0; j < SIZEOF_MONRAGE; j++)
				buf_mondata[i].rage_no[j] = buf[j];
		}
	}

	/* X^[̊ef[^ǂݍ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_MONDATA];

		file_read(buf, FILEPOS_MONDATA + SIZEOF_MONDATA * i, SIZEOF_MONDATA);

		/* lnf[^ */
		buf_mondata[i].level               = buf[16];
		buf_mondata[i].hit_point           = (buf[ 9] << 8) | buf[ 8];
		buf_mondata[i].magic_point         = (buf[11] << 8) | buf[10];
		buf_mondata[i].experience_point    = (buf[13] << 8) | buf[12];
		buf_mondata[i].gil                 = (buf[15] << 8) | buf[14];
		buf_mondata[i].speed               = buf[ 0];
		buf_mondata[i].attack_power        = buf[ 1];
		buf_mondata[i].defense_power       = buf[ 5];
		buf_mondata[i].evade_rate          = buf[ 3];
		buf_mondata[i].magic_attack_power  = buf[ 7];
		buf_mondata[i].magic_defense_power = buf[ 6];
		buf_mondata[i].magic_evade_rate    = buf[ 4];
		buf_mondata[i].hit_rate            = buf[ 2];
		buf_mondata[i].item_metamorph      = (buf[17] & 0xe0) >> 5;
		buf_mondata[i].tech_no             = buf[31] & 0x3f;

		/* nf[^ */
		buf_mondata[i].elem_none   = buf[24];
		buf_mondata[i].elem_absorb = buf[23];
		buf_mondata[i].elem_double = buf[25];

		/* ԕωnf[^ */
		buf_mondata[i].state_defense = bits_make32(buf[20], buf[21], buf[22], 0x00);
		buf_mondata[i].state_apriori = bits_make32(buf[27], buf[28], buf[29], buf[30]);

		/* tOnf[^ */
		buf_mondata[i].flag_undead     = ((buf[18] & 0x80) != 0x00);
		buf_mondata[i].flag_human      = ((buf[18] & 0x10) != 0x00);
		buf_mondata[i].flag_erase_name = ((buf[18] & 0x04) != 0x00);
		buf_mondata[i].flag_mp0        = ((buf[18] & 0x01) != 0x00);
		buf_mondata[i].flag_escape     = ((buf[19] & 0x08) == 0x00);

		/* ACeωf[^ */
		for (j = 0; j < 4; j++)
			buf_mondata[i].ragna_item_no[j] = tbl_ragna[buf[17] & 0x1f][j];
	}

	return buf_mondata;
}

monact_t *mon_readact(void)
{
	int i, j;
	unsigned char *buf = NULL;
	monact_t *buf_monact = NULL;
	int tbl_offset[NUM_MON];
	const size_t bufsize = 16384;

	/* X^[̍sp^[f[^i[obt@mۂ */
	if ((buf_monact = malloc(sizeof(monact_t) * NUM_MON)) == NULL)
		return NULL;
	for (i = 0; i < NUM_MON; i++)
	{
		buf_monact[i].ptn_num = 0;
		for (j = 0; j < sizeof(buf_monact[i].ptn) / sizeof(buf_monact[i].ptn[0]); j++)
		{
			buf_monact[i].ptn[j].action_size = 0;
			buf_monact[i].ptn[j].action = NULL;
		}
	}

	/* X^[̍sp^[f[^i[ĂAhX(ւ̃ItZbg)ǂݍ */
	for (i = 0; i < NUM_MON; i++)
	{
		unsigned char buf[SIZEOF_OFFSETADDR];

		file_read(buf, FILEPOS_MONACT_ADDR + SIZEOF_OFFSETADDR * i, SIZEOF_OFFSETADDR);
		tbl_offset[i] = (buf[1] << 8) | buf[0];
	}

	/* X^[̍sp^[f[^ǂݍ */
	if ((buf = malloc(bufsize)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);
	memset(buf, 0x00, bufsize);
	file_read(buf, FILEPOS_MONACT, bufsize);

	for (i = 0; i < NUM_MON; i++)
	{
		int ptn_num = 0;
		int separater_num = 0;
		unsigned char *ptr_buf = buf + tbl_offset[i];
		unsigned char *ptr_mark = ptr_buf;

		while (separater_num < 2)
		{
			if (*ptr_buf < 0xf0)
				ptr_buf++;
			else
			{
				if (*ptr_buf == 0xfe || *ptr_buf == 0xff)
				{
					size_t len_read = ptr_buf - ptr_mark;

					if (len_read > 0)
					{
						if ((buf_monact[i].ptn[ptn_num].action = malloc(len_read)) == NULL)
							err_printf(STR_FATAL_ERROR_IN_MALLOC);
						memmove(buf_monact[i].ptn[ptn_num].action, ptr_mark, len_read);
						buf_monact[i].ptn[ptn_num].action_size = len_read;
						ptn_num++;
					}

					if (*ptr_buf == 0xff)
						separater_num++;

					ptr_mark = ptr_buf + 1;
				}

				ptr_buf += tbl_skip_byte[*ptr_buf - 0xf0];
			}
		}

		buf_monact[i].ptn_num = ptn_num;
	}

	free(buf);
	buf = NULL;

	return buf_monact;
}

monparty_t *mon_readparty(void)
{
	int i, j;
	monparty_t *buf_monparty = NULL;

	/* X^[̃p[eBҐf[^i[obt@mۂ */
	if ((buf_monparty = malloc(sizeof(monparty_t) * NUM_MONPARTY)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* X^[̃p[eBҐ̊ef[^ǂݍ */
	for (i = 0; i < NUM_MONPARTY; i++)
	{
		unsigned char buf[SIZEOF_MONPARTY];

		file_read(buf, FILEPOS_MONPARTY + SIZEOF_MONPARTY * i, SIZEOF_MONPARTY);
		buf_monparty[i].bits_exist = buf[1];

		for (j = 0; j < 6; j++)
		{
			buf_monparty[i].monster_no[j] = buf[j + 2];
			if ((buf[14] & (1U << j)) != 0x00)
				buf_monparty[i].monster_no[j] += 0x100;
		}
	}

	/* X^[p[eB̖@Clf[^ǂݍ */
	for (i = 0; i < NUM_MONPARTY; i++)
	{
		unsigned char buf[SIZEOF_MONLEARNGAIN];

		file_read(buf, FILEPOS_MONLEARNGAIN + SIZEOF_MONLEARNGAIN * i, SIZEOF_MONLEARNGAIN);
		buf_monparty[i].magic_learning_gain = buf[0] & 0x1f;
	}
	for (; i < NUM_MONPARTY; i++)
		buf_monparty[i].magic_learning_gain = 0;

	/* X^[p[eBbɏo邩肷(sm) */
	for (i = 0; i < NUM_MONPARTY; i++)
	{
		int flag_beastfield = 1;

		if ((buf_monparty[i].bits_exist & 0x3f) == 0x00)
			flag_beastfield = 0;
		else
		{
			for (j = 0; j < 6; j++)
			{
				if ((buf_monparty[i].bits_exist & (1U << j)) != 0x00 && buf_monparty[i].monster_no[j] >= 256)
					flag_beastfield = 0;
			}
		}

		buf_monparty[i].flag_beastfield = flag_beastfield;
	}
	for (; i < NUM_MONPARTY; i++)
		buf_monparty[i].flag_beastfield = 0;

	/* X^[p[eB̑`f[^ǂݍ */
	for (i = 0; i < NUM_MONPARTY; i++)
	{
		unsigned char buf[SIZEOF_MONPARTY2];

		file_read(buf, FILEPOS_MONPARTY2 + SIZEOF_MONPARTY2 * i, SIZEOF_MONPARTY2);

		buf_monparty[i].flag_normal     = ((buf[0] & 0x10) == 0x00);
		buf_monparty[i].flag_backattack = ((buf[0] & 0x20) == 0x00);
		buf_monparty[i].flag_pincer     = ((buf[0] & 0x40) == 0x00);
		buf_monparty[i].flag_sideattack = ((buf[0] & 0x80) == 0x00);
	}

	return buf_monparty;
}

monenc_t *mon_readenc(void)
{
	int i, j;
	monenc_t *buf_monenc = NULL;

	/* X^[e[uf[^i[obt@mۂ */
	if ((buf_monenc = malloc(sizeof(monenc_t) * NUM_MONENC)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);
	for (i = 0; i < NUM_MONENC; i++)
	{
		for (j = 0; j < 4; j++)
			buf_monenc[i].party_no[j] = 0;
	}

	/* ʏ탂X^[̃X^[e[uf[^ǂݍ */
	for (i = 0; i < NUM_MONENCRND; i++)
	{
		unsigned char buf[SIZEOF_MONENCRND];

		file_read(buf, FILEPOS_MONENCRND + SIZEOF_MONENCRND * i, SIZEOF_MONENCRND);
		for (j = 0; j < 4; j++)
			buf_monenc[i].party_no[j] = ((buf[j * 2 + 1] & 0x7f) << 8) | buf[j * 2];
	}

	/* Œ胂X^[̃X^[e[uf[^ǂݍ */
	for (i = NUM_MONENCRND; i < NUM_MONENC; i++)
	{
		unsigned char buf[SIZEOF_MONENCFIXED];

		file_read(buf, FILEPOS_MONENCFIXED + SIZEOF_MONENCFIXED * (i - NUM_MONENCRND), SIZEOF_MONENCFIXED);
		for (j = 0; j < 2; j++)
			buf_monenc[i].party_no[j] = ((buf[j * 2 + 1] & 0x7f) << 8) | buf[j * 2];
	}

	return buf_monenc;
}

monmap_t *mon_readmap(void)
{
	int i, j, x, y;
	long filepos;
	monmap_t *buf_monmap = NULL;

	/* X^[}bvf[^i[obt@mۂ */
	if ((buf_monmap = malloc(sizeof(monmap_t) * NUM_MONMAP)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* X^[}bvf[^ǂݍ */
	filepos = FILEPOS_MONMAP;
	for (i = 0; i < NUM_MONMAP; i++)
	{
		for (y = 0; y < NUM_MONMAPAREA; y++)
		{
			for (x = 0; x < NUM_MONMAPAREA; x++)
			{
				unsigned char buf[SIZEOF_MONMAP];

				filepos += file_read(buf, filepos, SIZEOF_MONMAP);
				for (j = 0; j < 4; j++)
					buf_monmap[i].area[x][y].enc_no[j] = buf[j];
			}
		}
	}

	return buf_monmap;
}

int mon_writeparam(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̐lnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̐lnf[^o͂ */
	fprintf(fp, "No.       O                "
				" LV     HP     MP    EXP   M "
				"     U   h    U h  Hit%%\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s :  %3d, %5ld, %5ld, %5ld, %5ld, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d\n", i, i,
				mon_getname(i),
				ptr_mondata[i].level,
				ptr_mondata[i].hit_point,
				ptr_mondata[i].magic_point,
				ptr_mondata[i].experience_point,
				ptr_mondata[i].gil,
				ptr_mondata[i].speed,
				ptr_mondata[i].attack_power,
				ptr_mondata[i].defense_power,
				ptr_mondata[i].evade_rate,
				ptr_mondata[i].magic_attack_power,
				ptr_mondata[i].magic_defense_power,
				ptr_mondata[i].magic_evade_rate,
				ptr_mondata[i].hit_rate);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeelem(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̑nf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̑nf[^o͂ */
	fprintf(fp, "No.       O               "
				"          "
				"z          "
				"_\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s : ", i, i, mon_getname(i));

		/*  */
		str_printelem(fp, ptr_mondata[i].elem_none, ELEM_ALL);
		fprintf(fp, "  ");

		/* z */
		str_printelem(fp, ptr_mondata[i].elem_absorb, ELEM_ALL);
		fprintf(fp, "  ");

		/* _ */
		str_printelem(fp, ptr_mondata[i].elem_double, ELEM_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writestate(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̏ԕωnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̏ԕωnf[^o͂ */
	fprintf(fp, "No.       O               "
				"ԕωϐ1    2                3                 "
				"ԕω1    2                3                4\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s : ", i, i, mon_getname(i));

		/* ԕωւ̑ϐ */
		str_printstate(fp, ptr_mondata[i].state_defense, STATE_ALL & ~STATE_OTHER);
		fprintf(fp, "  ");

		/* ŏKpԕω */
		str_printstate(fp, ptr_mondata[i].state_apriori, STATE_ALL);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeflag(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̃tOnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̃tOnf[^o͂ */
	fprintf(fp, "No.       O               "
				" MP s l\n");
	for (i = 0; i < NUM_MON; i++)
	{
		/* etO */
		fprintf(fp, "%03d(%03X): %-16s :  %2s   %2s   %2s   %2s\n", i, i, mon_getname(i),
				str_getflag(ptr_mondata[i].flag_escape),
				str_getflag(ptr_mondata[i].flag_mp0),
				str_getflag(ptr_mondata[i].flag_undead),
				str_getflag(ptr_mondata[i].flag_human));
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeitem(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_nothing = "----------------";

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̏ACef[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̏ACef[^o͂ */
	fprintf(fp, "No.       O               "
				"(m)      "
				"(m)      "
				"Ƃ(m)    "
				"Ƃ(m)\n");
	for (i = 0; i < NUM_MON; i++)
	{
		const char *ptr_name[4];

		for (j = 0; j < 4; j++)
			ptr_name[j] = (ptr_mondata[i].item_no[j] != 0xff) ? item_getname(ptr_mondata[i].item_no[j]) : str_nothing;

		fprintf(fp, "%03d(%03X): %-16s : ", i, i, mon_getname(i));
		for (j = 0; j < 4; j++)
		{
			if (j != 0)
				fprintf(fp, ", ");
			fprintf(fp, "%-16s", ptr_name[j]);
		}
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeragna(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_nothing = "----------------";

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̃ACeωf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̃ACeωf[^o͂ */
	fprintf(fp, "No.       O               "
				"x   "
				"ACe1         "
				"ACe2         "
				"ACe3         "
				"ACe4\n");
	for (i = 0; i < NUM_MON; i++)
	{
		const char *ptr_name[4];

		for (j = 0; j < 4; j++)
			ptr_name[j] = (ptr_mondata[i].ragna_item_no[j] != 0xff) ? item_getname(ptr_mondata[i].ragna_item_no[j]) : str_nothing;

		fprintf(fp, "%03d(%03X): %-16s :  %3d,  ", i, i,
			mon_getname(i),
			ptr_mondata[i].item_metamorph);
		for (j = 0; j < 4; j++)
		{
			if (j != 0)
				fprintf(fp, ", ");
			fprintf(fp, "%-16s", ptr_name[j]);
		}
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writectrl(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_nothing = "----------------";

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[́uvsf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[́uvsf[^o͂ */
	fprintf(fp, "No.       O               "
				"1         "
				"2         "
				"3         "
				"4\n");
	for (i = 0; i < NUM_MON; i++)
	{
		const char *ptr_name[4];

		for (j = 0; j < 4; j++)
		{
			if (ptr_mondata[i].control_no[j] == 0xef)		/* 0xef = "Ƃ" */
				ptr_name[j] = mon_gettechname(i);
			else if (ptr_mondata[i].control_no[j] == 0xff)	/* 0xff = gp */
				ptr_name[j] = str_nothing;
			else
				ptr_name[j] = mag_getname(ptr_mondata[i].control_no[j]);
		}

		fprintf(fp, "%03d(%03X): %-16s : ", i, i, mon_getname(i));
		for (j = 0; j < 4; j++)
		{
			if (j != 0)
				fprintf(fp, ", ");
			fprintf(fp, "%-16s", ptr_name[j]);
		}
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writerage(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_nothing = "----------------";

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[́u΂vsf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[́u΂vsf[^o͂ */
	fprintf(fp, "No.       O               "
				"΂1         "
				"΂2\n");
	for (i = 0; i < NUM_MON; i++)
	{
		const char *ptr_name[2];

		for (j = 0; j < 2; j++)
		{
			if (ptr_mondata[i].rage_no[j] == 0xef)			/* 0xef = "Ƃ" */
				ptr_name[j] = mon_gettechname(i);
			else if (ptr_mondata[i].rage_no[j] == 0xff)		/* 0xff = gp */
				ptr_name[j] = str_nothing;
			else
				ptr_name[j] = mag_getname(ptr_mondata[i].rage_no[j]);
		}

		fprintf(fp, "%03d(%03X): %-16s : ", i, i, mon_getname(i));
		for (j = 0; j < 2; j++)
		{
			if (j != 0)
				fprintf(fp, ", ");
			fprintf(fp, "%-16s", ptr_name[j]);
		}
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writesketch(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_nothing = "----------------";

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[́uXPb`vsf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[́uXPb`vsf[^o͂ */
	fprintf(fp, "No.       O               "
				"XPb`1         "
				"XPb`2\n");
	for (i = 0; i < NUM_MON; i++)
	{
		const char *ptr_name[2];

		for (j = 0; j < 2; j++)
		{
			if (ptr_mondata[i].sketch_no[j] == 0xef)		/* 0xef = "Ƃ" */
				ptr_name[j] = mon_gettechname(i);
			else if (ptr_mondata[i].sketch_no[j] == 0xff)	/* 0xff = gp */
				ptr_name[j] = str_nothing;
			else
				ptr_name[j] = mag_getname(ptr_mondata[i].sketch_no[j]);
		}

		fprintf(fp, "%03d(%03X): %-16s : ", i, i, mon_getname(i));
		for (j = 0; j < 2; j++)
		{
			if (j != 0)
				fprintf(fp, ", ");
			fprintf(fp, "%-16s", ptr_name[j]);
		}
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writemisc(const char *ptr_filename, const mondata_t *ptr_mondata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̂̑̃f[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̂̑̃f[^o͂ */
	fprintf(fp, "No.       "
				"O               "
				"Ƃ\n");
	for (i = 0; i < NUM_MON; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s : %-16s\n", i, i,
				mon_getname(i),
				mon_gettechname(i));
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeact(const char *ptr_filename, const monact_t *ptr_monact, const mondata_t *ptr_mondata)
{
	int i, j, k;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monact == NULL || ptr_mondata == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̍sp^[f[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̍sp^[f[^o͂ */
	for (i = 0; i < NUM_MON; i++)
	{
		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[No.%03d(%03X): %s]\n", i, i, mon_getname(i));

		for (j = 0; j < ptr_monact[i].ptn_num; j++)
		{
			int turn = 1;
			int turn_old = 0;
			int letter_num = 0;
			unsigned char *buf = NULL;
			unsigned char *ptr_buf = NULL;
			const size_t bufsize = ptr_monact[i].ptn[j].action_size + 16;

			/* ͗p̈ꎞobt@mۂ */
			if ((buf = malloc(bufsize)) == NULL)
				err_printf(STR_FATAL_ERROR_IN_MALLOC);
			memset(buf, 0x00, bufsize);
			memmove(buf, ptr_monact[i].ptn[j].action, ptr_monact[i].ptn[j].action_size);

			/* Kp */
			ptr_buf = buf;
			ptr_buf = mon_printactcnd(fp, ptr_buf);
			fputc('\n', fp);

			/* ̂ݑ݂čsp^[݂Ȃꍇ */
			if (ptr_buf - buf >= (ptrdiff_t)ptr_monact[i].ptn[j].action_size)
			{
				fprintf(fp, "%d^[: ", turn);
				mon_printact(fp, 0xfe, i);
				fputc('\n', fp);
			}

			/* sp^[ */
			while (ptr_buf - buf < (ptrdiff_t)ptr_monact[i].ptn[j].action_size)
			{
				/* ^[o */
				if (*ptr_buf == 0xfd)
				{
					turn++;
					ptr_buf++;
					continue;
				}

				/* ^[̕\ */
				if (turn != turn_old)
					letter_num = fprintf(fp, "%d^[: ", turn);
				else
					fprintf(fp, "%*s", letter_num, "");
				turn_old = turn;

				/* sp^[f[^̉ */
				if (*ptr_buf < 0xf0)
				{
					mon_printact(fp, *ptr_buf, i);
					ptr_buf++;
				}
				else if (*ptr_buf == 0xfc)
				{
					ptr_buf = mon_printactcnd(fp, ptr_buf);
				}
				else
				{
					if (ptr_buf[0] == 0xf0)
					{
						/* _s(f0 xx yy zz) */
						/*   xx: @̔ԍ1 */
						/*   yy: @̔ԍ2 */
						/*   zz: @̔ԍ3 */
						for (k = 0; k < 3; k++)
						{
							if (k != 0)
								fprintf(fp, ", ");
							mon_printact(fp, ptr_buf[k + 1], i);
						}
					}
					else if (ptr_buf[0] == 0xf1)
					{
						/* UڕWݒ(f1 xx) */
						/*   xx: Ώ */
						fprintf(fp, "[UڕWݒ: %s]", mon_gettarget(ptr_buf[1]));
					}
					else if (ptr_buf[0] == 0xf2)
					{
						/* X^[p[eBύX(f2 xx yy zz) */
						/*   xx: o? */
						/*   yy: X^[p[eB̔ԍ(8rbg) */
						/*   zz: X^[p[eB̔ԍ(7rbg) */
						int party_no = ((ptr_buf[3] & 0x7f) << 8) | ptr_buf[2];

						fprintf(fp, "[ʃX^[p[eBo: %d(%03Xh)%s]", party_no, party_no,
								((ptr_buf[3] & 0x80) == 0x00) ? " (HPێ)" : "");
					}
					else if (ptr_buf[0] == 0xf3)
					{
						/* bZ[W\(f3 xx yy) */
						/*   xx: bZ[W̔ԍ(8rbg) */
						/*   yy: bZ[W̔ԍ(8rbg) */
						fprintf(fp, "[bZ[W\: %s]", mon_getmes((ptr_buf[2] << 8) | ptr_buf[1]));
					}
					else if (ptr_buf[0] == 0xf4)
					{
						/* pR}h̃_s(f4 xx yy zz) */
						/*   xx: R}h̔ԍ1 */
						/*   yy: R}h̔ԍ2 */
						/*   zz: R}h̔ԍ3 */
						for (k = 0; k < 3; k++)
						{
							if (k != 0)
								fprintf(fp, ", ");

							if (ptr_buf[k + 1] != 0xfe)
								fprintf(fp, "%s", chr_getcmdname(ptr_buf[k + 1]));
							else
								mon_printact(fp, ptr_buf[k + 1], i);
						}
					}
					else if (ptr_buf[0] == 0xf5)
					{
						/* ʉo(f5 xx yy zz) */
						/*   xx: s */
						/*   yy: s */
						/*   zz: s */
						fprintf(fp, "[ʉo: %02Xh, %02Xh, %02Xh]", ptr_buf[1], ptr_buf[2], ptr_buf[3]);
					}
					else if (ptr_buf[0] == 0xf6)
					{
						/* ACegp(f6 xx yy zz) */
						/*   xx: tO(0: ACegp, 1: ACe𓊂) */
						/*   yy: ACe̔ԍ1 */
						/*   zz: ACe̔ԍ2 */
						fprintf(fp, "[ACe%s] ", (ptr_buf[1] == 0x00) ? "gp" : "");
						for (k = 0; k < 2; k++)
						{
							if (k != 0)
								fprintf(fp, ", ");
							fprintf(fp, "%s", item_getname(ptr_buf[k + 2]));
						}
					}
					else if (ptr_buf[0] == 0xf7)
					{
						/* Cxg(f7 xx) */
						/*   xx: Cxg̔ԍ */
						int index;

						fprintf(fp, "[Cxg]");
						for (k = 0; k < 64 && (index = mon_geteventdata(ptr_buf[1], k)) >= 0; k++)
						{
							fprintf(fp, "\n%*s[bZ[W\: ", letter_num, "");
							fputs(mon_geteventmes(index), fp);
							fprintf(fp, "]");
						}
					}
					else if (ptr_buf[0] == 0xf8)
					{
						/* ϐݒ(f8 xx yy) */
						/*   xx: [hԍ */
						fprintf(fp, "[ϐ%dݒ: %s%d]",
								ptr_buf[1] & 0x1f,
								((ptr_buf[2] & 0x80) == 0x00) ? "" : (((ptr_buf[2] & 0x40) == 0x00) ? "+" : "-"),
								ptr_buf[2] & 0x3f);
					}
					else if (ptr_buf[0] == 0xf9)
					{
						/* rbgϐݒ(f9 xx yy zz) */
						/*   xx: rbg̎(0: ], 1: Zbg, 2: NA) */
						/*   yy: rbgԍ */
						/*   zz: rbgp^[ */
						const char *str_operation[] = {"]", "Zbg", "NA"};

						if (ptr_buf[1] < 3)
						{
							fprintf(fp, "[rbgϐ%d%s: bit%d]",
									ptr_buf[2],
									str_operation[ptr_buf[1]],
									ptr_buf[3]);
						}
					}
					else if (ptr_buf[0] == 0xfa)
					{
						/* X^[o(fa xx yy zz) */
						/*   xx: s */
						/*   yy: s */
						/*   zz: s */
						fprintf(fp, "[X^[o: %02Xh, %02Xh, %02Xh]", ptr_buf[1], ptr_buf[2], ptr_buf[3]);
					}
					else if (ptr_buf[0] == 0xfb)
					{
						/* e(fb xx yy) */
						/*   xx: ʂ̎ */
						switch (ptr_buf[1])
						{
						case 0x00:
							fprintf(fp, "[X^[oԂ]");
							break;
						case 0x01:
						case 0x05:
							fprintf(fp, "[G%s]", (ptr_buf[1] == 0x01) ? "ݒ" : "");
							break;
						case 0x02:
						case 0x09:
							fprintf(fp, "[퓬I]");
							break;
						case 0x03:
							fprintf(fp, "[KEp[eBo[ɉ]");
							break;
						case 0x04:
							fprintf(fp, "[퓬JňoߎԂ]");
							break;
						case 0x06:
						case 0x07:
							fprintf(fp, "[%sڕWݒ%sɂ]",
									mon_gettarget(ptr_buf[2]),
									(ptr_buf[1] == 0x06) ? "\" : "s\");
							break;
						case 0x0b:
						case 0x0c:
							fprintf(fp, "[%s: \"%s\"]",
									(ptr_buf[1] == 0x0b) ? "ݒ" : "",
									str_getstate(~0UL, ptr_buf[2]));
							break;
						default:
							fprintf(fp, "[%02X %02X %02X]", ptr_buf[0], ptr_buf[1], ptr_buf[2]);
							break;
						}
					}
					else
					{
						fprintf(fp, "[");
						for (k = 0; k < tbl_skip_byte[*ptr_buf - 0xf0]; k++)
						{
							fprintf(fp, "%02X", ptr_buf[k]);
							if (k < tbl_skip_byte[*ptr_buf - 0xf0] - 1)
								fprintf(fp, " ");
						}
						fprintf(fp, "]");
					}

					ptr_buf += tbl_skip_byte[*ptr_buf - 0xf0];
				}

				fputc('\n', fp);
			}

			free(buf);
			buf = NULL;
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writemes(const char *ptr_filename)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0')
	{
		assert(0);
		return 0;
	}

	/* X^[̃bZ[Wf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̃bZ[Wf[^o͂ */
	fprintf(fp, "No.      bZ[W\n");
	for (i = 0; i < NUM_MONMES; i++)
	{
		fprintf(fp, "%03d(%02X): ", i, i);
		fputs(mon_getmes(i), fp);
		fputc('\n', fp);
	}

	/* Cxg̃bZ[Wf[^o͂ */
	fprintf(fp, "No.      bZ[W\n");
	for (i = 0; i < NUM_EVENTMES; i++)
	{
		fprintf(fp, "%03d(%02X): ", i, i);
		fputs(mon_geteventmes(i), fp);
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeevent(const char *ptr_filename)
{
	int i, j;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0')
	{
		assert(0);
		return 0;
	}

	/* Cxg̃bZ[Wf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* Cxg̃bZ[Wf[^o͂ */
	for (i = 0; i < NUM_EVENTDATA; i++)
	{
		int index;

		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[No.%03d(%02X)]\n", i, i);
		for (j = 0; j < 64 && (index = mon_geteventdata(i, j)) >= 0; j++)
		{
			fputs(mon_geteventmes(index), fp);
			fputc('\n', fp);
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeparty(const char *ptr_filename, const monparty_t *ptr_monparty)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_nothing = "----------------";

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monparty == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[̃p[eBҐf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[̃p[eBҐf[^o͂ */
	fprintf(fp, "No.       "
				"C  "
				"ʏ w    "
				"o   "
				"X^[1         "
				"X^[2         "
				"X^[3         "
				"X^[4         "
				"X^[5         "
				"X^[6\n");
	for (i = 0; i < NUM_MONPARTY; i++)
	{
		const char *ptr_name[6];

		for (j = 0; j < 6; j++)
			ptr_name[j] = (ptr_monparty[i].monster_no[j] < NUM_MON) ? mon_getname(ptr_monparty[i].monster_no[j]) : str_nothing;

		fprintf(fp, "%03d(%03X):  %3d,  %2s   %2s   %2s   %2s   ", i, i,
				ptr_monparty[i].magic_learning_gain,
				str_getflag(ptr_monparty[i].flag_normal),
				str_getflag(ptr_monparty[i].flag_backattack),
				str_getflag(ptr_monparty[i].flag_pincer),
				str_getflag(ptr_monparty[i].flag_sideattack));
		for (j = 0; j < 6; j++)
		{
			fprintf(fp, "%-2s",
					str_getflag((ptr_monparty[i].bits_exist & (1U << j)) != 0x00));
		}
		fprintf(fp, "  ");
		for (j = 0; j < 6; j++)
		{
			const int flag_absent = (ptr_monparty[i].monster_no[j] < NUM_MON && (ptr_monparty[i].bits_exist & (1U << j)) == 0x00);

			if (j != 0)
				fprintf(fp, ", ");
			fprintf(fp, "%s%-16s%s",
					(flag_absent) ? "(" : " ",
					ptr_name[j],
					(flag_absent) ? ")" : " ");
		}
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writeenc(const char *ptr_filename, const monenc_t *ptr_monenc, const monparty_t *ptr_monparty)
{
	int i, j, k;
	FILE *fp = NULL;
	const char *str_nothing = "----------------";

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monenc == NULL || ptr_monparty == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[e[uf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[e[uo͂ */
	for (i = 0; i < NUM_MONENC; i++)
	{
		int size = (i < NUM_MONENCRND)? 4 : 2;

		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[No.%03d(%03X)]\n", i, i);
		for (j = 0; j < size; j++)
		{
			const char *ptr_name[6];
			const monparty_t tmp_party = ptr_monparty[ptr_monenc[i].party_no[j]];

			for (k = 0; k < 6; k++)
				ptr_name[k] = (tmp_party.monster_no[k] < NUM_MON) ? mon_getname(tmp_party.monster_no[k]) : str_nothing;

			fprintf(fp, "%02d: %03d(%03X): ", j + 1, ptr_monenc[i].party_no[j], ptr_monenc[i].party_no[j]);
			for (k = 0; k < 6; k++)
			{
				const int flag_absent = (tmp_party.monster_no[k] < NUM_MON && (tmp_party.bits_exist & (1U << k)) == 0x00);

				if (k != 0)
					fprintf(fp, ", ");
				fprintf(fp, "%s%-16s%s",
						(flag_absent) ? "(" : " ",
						ptr_name[k],
						(flag_absent) ? ")" : " ");
			}
			fputc('\n', fp);
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int mon_writemap(const char *ptr_filename, const monmap_t *ptr_monmap)
{
	int i, j, x, y;
	FILE *fp = NULL;
	const char *str_name[NUM_MONMAP][4] =
	{
		{"EO(r)",	"EO(X)",		"EO()",		"EO()"},
		{"E(r)",	"E(X)",		"E()",		"E()"},
	};

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_monmap == NULL)
	{
		assert(0);
		return 0;
	}

	/* X^[}bvf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* X^[}bvo͂ */
	for (i = 0; i < NUM_MONMAP; i++)
	{
		for (j = 0; j < 4; j++)
		{
			if (i != 0 || j != 0)
				fputc('\n', fp);

			fprintf(fp, "[No.%1d-%1d: %s]\n", i, j, str_name[i][j]);
			for (y = 0; y < NUM_MONMAPAREA; y++)
			{
				for (x = 0; x < NUM_MONMAPAREA; x++)
				{
					if (x != 0)
						fprintf(fp, " ");

					fprintf(fp, "%03d", ptr_monmap[i].area[x][y].enc_no[j]);
				}
				fputc('\n', fp);
			}
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

static unsigned char *mon_printactcnd(FILE *stream, unsigned char *buf_cnd)
{
	int condition_num = 0;
	unsigned char *ptr_cnd = buf_cnd;

	/* `FbN */
	if (stream == NULL || buf_cnd == NULL)
	{
		assert(0);
		return buf_cnd;
	}

	while (*ptr_cnd == 0xfc)
	{
		if (condition_num == 0)
			fprintf(stream, ": ");
		else
			fprintf(stream, " & ");

		if (ptr_cnd[1] == 0x01)
		{
			/* ̃R}hɑ΂JE^[(fc 01 xx yy) */
			/*   xx: R}h̔ԍ1 */
			/*   yy: R}h̔ԍ2 */
			fprintf(stream, "[\"%s\" ", chr_getcmdname(ptr_cnd[2]));
			if (ptr_cnd[2] != ptr_cnd[3])
				fprintf(stream, " \"%s\" ", chr_getcmdname(ptr_cnd[3]));
			fprintf(stream, "󂯂]");
		}
		else if (ptr_cnd[1] == 0x02)
		{
			/* ̖@ɑ΂JE^[(fc 02 xx yy) */
			/*   xx: @̔ԍ1 */
			/*   yy: @̔ԍ2 */
			fprintf(stream, "[@ \"%s\" ", mag_getname(ptr_cnd[2]));
			if (ptr_cnd[2] != ptr_cnd[3])
				fprintf(stream, " \"%s\" ", mag_getname(ptr_cnd[3]));
			fprintf(stream, "󂯂]");
		}
		else if (ptr_cnd[1] == 0x03)
		{
			/* ̃ACeɑ΂JE^[(fc 03 xx yy) */
			/*   xx: ACe̔ԍ1 */
			/*   yy: ACe̔ԍ2 */
			fprintf(stream, "[ACe \"%s\" ", item_getname(ptr_cnd[2]));
			if (ptr_cnd[2] != ptr_cnd[3])
				fprintf(stream, " \"%s\" ", item_getname(ptr_cnd[3]));
			fprintf(stream, "󂯂]");
		}
		else if (ptr_cnd[1] == 0x04)
		{
			/* Uɑ΂JE^[(fc 04 xx 00) */
			/*   xx:  */
			int index;
			char str_element[256] = "";
			char *ptr = str_element;
			const elem_t element = ptr_cnd[2];

			for (index = 0; index < NUM_ELEMENT; index++)
				ptr += sprintf(ptr, "%s", str_getelem(element, index));

			fprintf(stream, "[%s̍U󂯂]", str_element);
		}
		else if (ptr_cnd[1] == 0x05)
		{
			/* _[W/񕜂ɑ΂JE^[(fc 05 00 00) */
			fprintf(stream, "[_[W/񕜂󂯂]");
		}
		else if (ptr_cnd[1] == 0x06)
		{
			/* qbg|Cglȉ(fc 06 xx yy) */
			/*   xx: Ώ */
			/*   yy: Eqbg|Cg128Ŋl */
			fprintf(stream, "[%sHP%ldȉł]",
					mon_gettarget(ptr_cnd[2]),
					(long)ptr_cnd[3] * 128);
		}
		else if (ptr_cnd[1] == 0x07)
		{
			/* }WbN|Cglȉ(fc 07 xx yy) */
			/*   xx: Ώ */
			/*   yy: E}WbN|Cg */
			fprintf(stream, "[%sMP%ldȉł]",
					mon_gettarget(ptr_cnd[2]),
					(long)ptr_cnd[3]);
		}
		else if (ptr_cnd[1] == 0x08 || ptr_cnd[1] == 0x09)
		{
			/* Ԃ̏Ԃƈv(fc 08 xx yy) */
			/* Ԃ̏Ԃł͂Ȃ(fc 09 xx yy) */
			/*   xx: Ώ */
			/*   yy: Ԃ̃CfbNX */
			fprintf(stream, "[%s̏Ԃ \"%s\" %s]",
					mon_gettarget(ptr_cnd[2]),
					str_getstate(~0UL, ptr_cnd[3]),
					(ptr_cnd[1] == 0x08) ? "" : "Ȃ");
		}
		else if (ptr_cnd[1] == 0x0b || ptr_cnd[1] == 0x16)
		{
			/* X^[oԂ莞Ԉȏł(fb 0d xx 00) */
			/* 퓬JňoߎԂ莞Ԉȏł(fb 16 xx 00) */
			/*   xx: Pʎ */
			fprintf(stream, "[%s%dPʎԈȏł]",
					(ptr_cnd[1] == 0x0b) ? "X^[o" : "퓬Jňoߎ",
					ptr_cnd[2]);
		}
		else if (ptr_cnd[1] == 0x0c || ptr_cnd[1] == 0x0d)
		{
			/* ϐ̒lł(fc 0c xx yy) */
			/* ϐ̒lȏł(fc 0d xx yy) */
			/*   xx: ǂ̓ϐlgp邩 */
			/*   yy: ϐ̒l */
			fprintf(stream, "[ϐ%d%d%sł]",
					ptr_cnd[2] & 0x1f,
					ptr_cnd[3],
					(ptr_cnd[1] == 0x0c) ? "" : "ȏ");
		}
		else if (ptr_cnd[1] == 0x0e || ptr_cnd[1] == 0x0f)
		{
			/* x̒lł(fc 0c xx yy) */
			/* x̒lȏł(fc 0d xx yy) */
			/*   xx: Ώ */
			/*   yy: x */
			fprintf(stream, "[%s̃x%d%sł]",
					mon_gettarget(ptr_cnd[2]),
					ptr_cnd[3],
					(ptr_cnd[1] == 0x0e) ? "" : "ȏ");
		}
		else if (ptr_cnd[1] == 0x10)
		{
			/* ȊÕX^[Ȃ(fc 10 00 00) */
			fprintf(stream, "[ȊÕX^[Ȃ]");
		}
		else if (ptr_cnd[1] == 0x11 || ptr_cnd[1] == 0x12 || ptr_cnd[1] == 0x19)
		{
			/* ʒuɃX^[݂(fc 11 xx 00) */
			/* ʒuɃX^[݂Ȃ(fc 12 xx 00) */
			/* gʒuɑ݂(fc 19 xx 00) */
			/*   xx: ʒu(rbga) */
			int index;
			char str_no[256] = "";
			char *ptr = str_no;

			for (index = 0; index < 8; index++)
			{
				if ((ptr_cnd[2] & (1U << index)) != 0x00)
					ptr += sprintf(ptr, "%s%d", (ptr > str_no) ? ", " : "", index + 1);
			}

			switch (ptr_cnd[1])
			{
			case 0x11:
				fprintf(stream, "[%sԖڂ̃X^[oĂ]", str_no);
				break;
			case 0x12:
				if (ptr_cnd[2] != 0x00)
					fprintf(stream, "[%sԖڂ̃X^[Ȃ]", str_no);
				else
					fprintf(stream, "[t@CiA^bN]");
				break;
			case 0x19:
				fprintf(stream, "[g%sԖڂ̈ʒuɂ]", str_no);
				break;
			}
		}
		else if (ptr_cnd[1] == 0x13)
		{
			/* /G̐l(fc 13 xx yy) */
			/*   xx: tO(0: ̐lyylȏア邩, 1: G̐yÿȉ) */
			/*   yy: /G̐l */
			if (ptr_cnd[2] == 0x00)
				fprintf(stream, "[LN^%dlȏ㑶݂]", ptr_cnd[3]);
			else
				fprintf(stream, "[X^[̐%d̈ȉł]", ptr_cnd[3]);
		}
		else if (ptr_cnd[1] == 0x14 || ptr_cnd[1] == 0x15)
		{
			fprintf(stream, "[rbgϐ%dbit%d%sĂ]",
					ptr_cnd[2],
					ptr_cnd[3],
					(ptr_cnd[1] == 0x14) ? "Zbg" : "NA");
		}
		else if (ptr_cnd[1] == 0x1a)
		{
			/* _̑ƈv(fc 1a xx yy) */
			/*   xx: Ώ */
			/*   yy:  */
			int index;

			fprintf(stream, "[%s̎_ \"",
					mon_gettarget(ptr_cnd[2]));
			for (index = 0; index < NUM_ELEMENT; index++)
				fprintf(stream, "%s", str_getelem(ptr_cnd[3], index));
			fprintf(stream, "\" ł]");
		}
		else if (ptr_cnd[1] == 0x1b)
		{
			/* X^[p[eBԍ̒lƈv(fc 1b xx yy) */
			/*   xx: X^[p[eBԍ(ʃoCg) */
			/*   yy: X^[p[eBԍ(ʃoCg) */
			int party_no = (ptr_cnd[3] << 8) | ptr_cnd[2];

			fprintf(stream, "[X^[p[eBԍ%d(%03Xh)ł]", party_no, party_no);
		}
		else
		{
			fprintf(stream, "[%02X %02X %02X %02X]", ptr_cnd[0], ptr_cnd[1], ptr_cnd[2], ptr_cnd[3]);
		}

		condition_num++;
		ptr_cnd += 4;
	}

	if (condition_num == 0)
		fprintf(stream, ": [ʏ펞]");

	return ptr_cnd;
}

static void mon_printact(FILE *stream, int index, int monster_no)
{
	/* `FbN */
	if (stream == NULL || index < 0 || index >= 0xff || monster_no < 0 || monster_no >= NUM_MON)
	{
		assert(0);
		return;
	}

	switch (index)
	{
	case 0xef:		/* Ƃ */
		fprintf(stream, "%s", mon_gettechname(monster_no));
		break;
	case 0xfe:
		fprintf(stream, "(Ȃ)");
		break;
	default:
		fprintf(stream, "%s", mag_getname(index));
		break;
	}
}

void mon_freeact(monact_t *ptr_monact)
{
	int i, j;

	/* `FbN */
	if (ptr_monact == NULL)
	{
		assert(0);
		return;
	}

	for (i = 0; i < NUM_MON; i++)
	{
		for (j = 0; j < ptr_monact[i].ptn_num; j++)
		{
			if (ptr_monact[i].ptn[j].action != NULL)
			{
				free(ptr_monact[i].ptn[j].action);
				ptr_monact[i].ptn[j].action = NULL;
			}
		}
	}
}
