/*
 *	filesfc.c
 *		Copyright (C)2001-2003 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include "errmes.h"
#include "filesfc.h"

#define FILEPOS_INFOBLOCK_LOROM	0x007fc0	/* CtH[VubN̊i[ʒu(LoROM̏ꍇ) */
#define FILEPOS_INFOBLOCK_HIROM	0x00ffc0	/* CtH[VubN̊i[ʒu(HiROM̏ꍇ) */
#define SIZEOF_INFOBLOCK		0x20		/* CtH[VubÑTCY */

static int ptnseek(FILE *stream, const unsigned char *ptr_matchptn, size_t ptnsize);

static FILE *fp_sfc = NULL;		/* t@C|C^(X[p[t@~Rŗp) */
static long offset_sfc = 0;		/* ROMC[W̐擪t@Cʒu */

int file_open_sfc(const char *ptr_filename, const char *ptr_title)
{
	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_title == NULL || *ptr_title == '\0')
	{
		assert(0);
		return 0;
	}

	/* ROMC[Wt@CJ */
	if ((fp_sfc = fopen(ptr_filename, "rb")) == NULL)
		return 0;

	/* CtH[VubNꏊTāAwb_̗L𔻕ʂ */
	offset_sfc = 0;
	if (!ptnseek(fp_sfc, (unsigned char *)ptr_title, strlen(ptr_title)))
	{
		unsigned char buf[SIZEOF_INFOBLOCK];
		long offset_infoblock = ftell(fp_sfc);

		/* CtH[VubNǂݍ */
		if (fread(buf, 1, sizeof(buf), fp_sfc) < sizeof(buf))
			err_printf(STR_FATAL_ERROR_IN_FREAD);

		/* CtH[VubN̏ƂROM^Cv𔻕ʂāAwb_̗L𔻕ʂ */
		offset_sfc = offset_infoblock - (((buf[21] & 0x0f) == 0) ? FILEPOS_INFOBLOCK_LOROM : FILEPOS_INFOBLOCK_HIROM);
	}

	/* t@C|C^t@C擪ʒuɖ߂ */
	if (fseek(fp_sfc, 0, SEEK_SET))
		err_printf(STR_FATAL_ERROR_IN_FSEEK);

	return 1;
}

int file_close_sfc(void)
{
	int ret = 0;

	if (fp_sfc != NULL)
		fclose(fp_sfc);

	/* f[^ */
	fp_sfc = NULL;
	offset_sfc = 0;

	return ret;
}

size_t file_read_sfc(void *ptr_buf, long position, size_t size)
{
	size_t ret = 0;

	/* `FbN */
	if (ptr_buf == NULL || position < 0)
	{
		assert(0);
		return 0;
	}

	if (fseek(fp_sfc, position + offset_sfc, SEEK_SET))
		err_printf(STR_FATAL_ERROR_IN_FSEEK);
	if ((ret = fread(ptr_buf, 1, size, fp_sfc)) < size)
		err_printf(STR_FATAL_ERROR_IN_FREAD);

	return ret;
}

static int ptnseek(FILE *stream, const unsigned char *ptr_matchptn, size_t ptnsize)
{
	int i;
	int flag_success = 0;
	long file_offset;
	long file_pos;
	long file_tail;
	unsigned char *buf = NULL;
	unsigned char *ptr_buf = NULL;
	int tbl_skip[256];
	const size_t bufsize = (ptnsize * 16 > 1024) ? ptnsize * 16 : 1024;

	/* `FbN */
	if (stream == NULL || ptr_matchptn == NULL || ptnsize == 0)
	{
		assert(0);
		return 0;
	}

	/* t@C̏I[̃ItZbg擾 */
	file_offset = file_pos = ftell(stream);
	if (fseek(stream, 0, SEEK_END))
		err_printf(STR_FATAL_ERROR_IN_FSEEK);
	file_tail = ftell(stream);
	if (fseek(stream, file_offset, SEEK_SET))
		err_printf(STR_FATAL_ERROR_IN_FSEEK);

	/* p̃XLbve[u쐬 */
	for (i = 0; i < sizeof(tbl_skip) / sizeof(tbl_skip[0]); i++)
		tbl_skip[i] = (int)ptnsize;
	for (i = 0; i < (int)ptnsize - 1; i++)
		tbl_skip[ptr_matchptn[i]] = (int)ptnsize - i - 1;

	/* ꎞobt@mۂ */
	if ((buf = malloc(bufsize)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);
	ptr_buf = buf;

	/* Jn */
	if (fread(buf, 1, ptnsize, stream) < ptnsize && ferror(stream))
		err_printf(STR_FATAL_ERROR_IN_FREAD);
	ptr_buf += ptnsize;

	while (file_pos < file_tail)
	{
		if (*(ptr_buf - 1) == ptr_matchptn[ptnsize - 1])
		{
			if (memcmp(ptr_buf - ptnsize, ptr_matchptn, ptnsize) == 0)
			{
				if (fseek(stream, -(int)ptnsize, SEEK_CUR))
					err_printf(STR_FATAL_ERROR_IN_FSEEK);
				flag_success = 1;
				break;
			}
		}

		if (ptr_buf - buf + tbl_skip[*(ptr_buf - 1)] >= (ptrdiff_t)bufsize - 8)
		{
			memmove(buf, ptr_buf - ptnsize, ptnsize);
			ptr_buf = &buf[ptnsize];
		}

		if (fread(ptr_buf, 1, tbl_skip[*(ptr_buf - 1)], stream) < (size_t)tbl_skip[*(ptr_buf - 1)] && ferror(stream))
			err_printf(STR_FATAL_ERROR_IN_FREAD);
		ptr_buf  += tbl_skip[*(ptr_buf - 1)];
		file_pos += tbl_skip[*(ptr_buf - 1)];
	}

	/* Ȃꍇ́At@C|C^̈ʒuɖ߂ */
	if (!flag_success)
	{
		if (fseek(stream, file_offset, SEEK_SET))
			err_printf(STR_FATAL_ERROR_IN_FSEEK);
	}

	return !flag_success;
}
